/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: AbstractCursesModel.java 256 2008-07-15 15:54:02Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package uk.co.ruelgnoj.orion.cc;

import java.util.ArrayList;
import java.util.List;

/**
 * A curses model foundation providing support for event firing.
 * @author Steve Millington
 * @version $Id: AbstractCursesModel.java 256 2008-07-15 15:54:02Z steve $
 */
public abstract class AbstractCursesModel implements CursesModel {
    private final List<CursesModelListener> listeners;
    private final CursesModelEvent theEvent;
    private boolean	batch;
    private boolean	eventsPending;

    /**
     * Constructor.
     */
    public AbstractCursesModel() {
	theEvent = new CursesModelEvent(this);
	listeners = new ArrayList<>();
	batch = false;
    }

    @Override
    public void addModelListener(CursesModelListener l) {
	listeners.add(l);
    }

    @Override
    public void removeModelListener(CursesModelListener l) {
	listeners.remove(l);
    }

    /**
     * Fire an event saying the model has changed.
     * <p>This only fires the event if batch mode is off, otherwise it
     * effectively defer the even generation until the event firing is turned
     * back on again.
     */
    protected void maybeFireEvent() {
	if (batch) {
	    eventsPending = true;
	} else {
	    fireEvent();
	    eventsPending = false;
	}
    }

    /**
     * Fire an event.
     */
    private void fireEvent() {
	for (CursesModelListener l : listeners) {
	    l.dataChanged(theEvent);
	}
    }

    /**
     * Set the batch mode. This determines whether events will be fired when
     * <code>maybeFireEvent</code> is called by the model implementation.
     * <p>This allows support for many writes into a modifiable model to
     * produce only a singe change event.
     * @param b True defers event firing, False resumes it.
     */
    public void setBatch(boolean b) {
	if (b == batch) {
	    return;
	}
	batch = b;
	if (!b) {
	    if (eventsPending) {
		fireEvent();
		eventsPending = false;
	    }
	}
    }
}
