/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: Curses.java 256 2008-07-15 15:54:02Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.cc;

import java.awt.Point;

import javax.swing.JComponent;
import javax.swing.UIManager;

/**
 * A component for displaying curses-style.
 *
 * @author Steve Millington
 * @version $Id: Curses.java 256 2008-07-15 15:54:02Z steve $
 */
public class Curses extends JComponent {

    private static final String UICLASSID = "CursesUI";

    private CursesModel model;

    /**
     * Create a component with the specified width and height.
     * <p>
     * This creates a <code>DefaultCursesModel</code> for the data.
     *
     * @param w The number of columns.
     * @param h The number of rows.
     */
    public Curses(int w, int h) {
        this(new DefaultCursesModel(w, h));
    }

    /**
     * Create a component with the specified model.
     *
     * @param model The model
     */
    public Curses(CursesModel model) {
        this.model = model;

        setFocusable(false);
        setOpaque(true);
        updateUI();
    }

    /**
     * Get the model associated with this component.
     *
     * @return The model.
     */
    public CursesModel getModel() {
        return model;
    }

    /**
     * Set the model associated with this component.
     *
     * @param model The new model.
     */
    public void setModel(CursesModel model) {
        if (model == null) {
            throw new NullPointerException("model");
        }
        final CursesModel oldModel = this.model;
        this.model = model;
        firePropertyChange("model", oldModel, model);
    }

    /**
     * Get the row associated with a given component point.
     *
     * @param p The point
     * @return The row or -1 if the point does not have a corresponding row.
     */
    public int getRow(Point p) {
        return getUI().getRow(p);
    }

    /**
     * Get the column associated with a given component point.
     *
     * @param p The point
     * @return The column or -1 if the point does not have a corresponding
     * column.
     */
    public int getColumn(Point p) {
        return getUI().getColumn(p);
    }

    @Override
    public String toString() {
        return "uk.co.ruelgnoj.orion.cc.Curses[" + model.getRows() + "," + model.getCols() + "]";
    }

    /**
     * Get the UI implementation.
     *
     * @return The UI
     */
    public CursesUI getUI() {
        return (CursesUI) ui;
    }

    /**
     * Set the UI implementation.
     *
     * @param ui The UI
     */
    public void setUI(CursesUI ui) {
        super.setUI(ui);
    }

    @Override
    public String getUIClassID() {
        PackageManager.initialize();
        return UICLASSID;
    }

    @Override
    public void updateUI() {
        setUI((CursesUI) UIManager.getUI(this));
    }
}
