/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: CursesInputField.java 256 2008-07-15 15:54:02Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.cc;

import javax.swing.AbstractAction;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.JComponent;
import javax.swing.KeyStroke;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import uk.co.ruelgnoj.orion.component.util.Activatable;

public class CursesInputField implements Activatable {

    private final JComponent jc;
    private final WritableCursesModel cc;
    private CursesInputFieldListener fl;
    private final StringBuilder value;
    private String labelText;
    private final Color labelColor;
    private final Color findColor;
    private final Color showDefaultColor;
    private String defaultValue;
    private final KeyListener keyHandler;
    private final ActionMap actionMap;
    private final InputMap inputMap;
    private boolean active;

    public CursesInputField(JComponent jc, WritableCursesModel cc) {
        this.jc = jc;
        this.cc = cc;
        active = false;
        value = new StringBuilder();
        findColor = Color.green;
        showDefaultColor = new Color(0x00, 0xa0, 0xa0);
        labelColor = Color.white;
        keyHandler = new KeyHandler();
        actionMap = createActionMap();
        inputMap = createInputMap();
    }

    protected ActionMap createActionMap() {
        final ActionMap am = new ActionMap();
        am.put("cancel", new CancelAction());
        am.put("ok", new ApplyAction());
        return am;
    }

    protected InputMap createInputMap() {
        final InputMap im = new InputMap();
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0, false), "cancel");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0, false), "ok");
        return im;
    }

    @Override
    public boolean isActive() {
        return active;
    }

    public void setLabel(String s) {
        labelText = s;
    }

    public String getLabel() {
        return labelText;
    }

    public void setListener(CursesInputFieldListener fl) {
        this.fl = fl;
    }

    private void redraw() {
        cc.setBatch(true);
        cc.setCursorVisible(true);
        cc.setBG(Color.black);
        cc.setFG(findColor);
        cc.clear();
        cc.moveTo(0, 0);
        cc.setFG(labelColor);
        cc.print(labelText + ": ");
        if (defaultValue == null) {
            cc.setFG(findColor);
            cc.print(value.toString());
        } else {
            cc.setFG(showDefaultColor);
            cc.print(defaultValue);
            cc.moveTo(labelText.length() + 2, 0);
        }
        cc.setBatch(false);
    }

    private KeyListener[] oldListeners = null;
    private ActionMap savedActionMap;
    private InputMap savedInputMap;

    @Override
    public void activate() {
        active = true;
        final JComponent parent = jc;
        oldListeners = parent.getKeyListeners();
        for (KeyListener listener : oldListeners) {
            parent.removeKeyListener(listener);
        }
        cc.setCursorVisible(true);
        redraw();
        //cc.requestFocus();
        parent.addKeyListener(keyHandler);

        savedActionMap = parent.getActionMap();
        savedInputMap = parent.getInputMap();
        parent.setActionMap(actionMap);
        parent.setInputMap(JComponent.WHEN_FOCUSED, inputMap);
    }

    @Override
    public void deactivate() {
        active = false;
        final JComponent parent = jc;
        for (KeyListener listener : oldListeners) {
            parent.addKeyListener(listener);
        }
        oldListeners = null;
        cc.setCursorVisible(false);
        parent.removeKeyListener(keyHandler);
        parent.setActionMap(savedActionMap);
        parent.setInputMap(JComponent.WHEN_FOCUSED, savedInputMap);
    }

    public String getValue() {
        if (defaultValue != null) {
            return defaultValue;
        }
        return value.toString();
    }

    public void setDefaultValue(String nV) {
        defaultValue = nV;
        value.setLength(0);	// Refresh the 'returned' value!
    }

    private class KeyHandler implements KeyListener {

        private boolean ignoreTyped;

        public KeyHandler() {
        }

        @Override
        public void keyPressed(KeyEvent ev) {
            ignoreTyped = ev.isConsumed();
        }

        @Override
        public void keyTyped(KeyEvent ev) {
            if (ev.isConsumed() || ignoreTyped) {
                return;
            }
            char c = ev.getKeyChar();
            switch (c) {
                case 0x08:			// Backspace
                    if (value.length() > 0) {
                        value.setLength(value.length() - 1);
                        redraw();
                    }
                    break;
                case 0x1b:			// ESC
                case '\t':
                case '\r':
                case '\n':
                    ev.consume();
                    break;
                default:
                    value.append(ev.getKeyChar());
                    if (defaultValue != null) {
                        defaultValue = null;
                        redraw();
                    } else {
                        cc.print(ev.getKeyChar());
                    }
                    break;
            }
        }

        @Override
        public void keyReleased(KeyEvent ev) {
            ignoreTyped |= ev.isConsumed();
        }
    }

    public class CancelAction extends AbstractAction {

        @Override
        public void actionPerformed(ActionEvent e) {
            fl.inputFieldCancelled(CursesInputField.this);
        }
    }

    public class ApplyAction extends AbstractAction {

        @Override
        public void actionPerformed(ActionEvent e) {
            fl.inputFieldSet(CursesInputField.this);
        }
    }
}
