/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: DefaultCursesModel.java 256 2008-07-15 15:54:02Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.cc;

import java.awt.Color;
import java.awt.Paint;

import java.util.Arrays;

/**
 * The default model for a curses component. This model is a fixed-dimension
 * writable curses model.
 *
 * @author Steve Millington
 * @version $Id: DefaultCursesModel.java 256 2008-07-15 15:54:02Z steve $
 */
public class DefaultCursesModel extends AbstractCursesModel implements WritableCursesModel {

    private static final int FG_MASK = 0x000000ff;
    private static final int BG_MASK = 0x0000ff00;
    private static final int UL_MASK = 0x00010000;
    private static final int NOT_FG_MASK = 0xffffff00;
    private static final int NOT_BG_MASK = 0xffff00ff;
    private static final int NOT_UL_MASK = 0xfffeffff;
    private static final int BG_OFFSET = 8; // Shift for BG attr
    private static final int COLOR_TABLE_SIZE = 256;

    private int _w;
    private int _h;
    private int _wh;
    private int _cursorOffset;
    private final Paint[] _colorTable;
    private char[] _charBuffer;
    private int[] _attrBuffer;
    private int _currentAttr;

    private int _nextFreeColorTableEntry;
    private boolean _showCursor;

    /**
     * Constructor specifying the size.
     *
     * @param w The width
     * @param h The height
     */
    public DefaultCursesModel(int w, int h) {
        _w = w;
        _h = h;
        _wh = w * h;
        _cursorOffset = 0;
        _colorTable = new Paint[COLOR_TABLE_SIZE];
        _charBuffer = new char[_wh];
        _attrBuffer = new int[_wh];
        _colorTable[0] = Color.white;	// Default BG Color;
        _colorTable[1] = Color.black;	// Default FG Color;
        _currentAttr = 0x00000001;
        _showCursor = false;
        clear();
    }

    
    @Override
    public void setSize(int newWidth, int newHeight) {
        final int newBuffSz = newWidth * newHeight;
        final char[] newCharBuffer = new char[newBuffSz];
        final int[] newAttrBuffer = new int[newBuffSz];
        for (int i = 0; i < newBuffSz; i++) {
            newCharBuffer[i] = ' ';
            newAttrBuffer[i] = _currentAttr;
        }
        final int w1 = Math.min(_w, newWidth);
        final int h1 = Math.min(_h, newHeight);
        for (int i = 0; i < h1; i++) {
            for (int j = 0; j < w1; j++) {
                final int off1 = i * newWidth + j;
                final int off2 = i * _w + j;
                newCharBuffer[off1] = _charBuffer[off2];
                newAttrBuffer[off1] = _attrBuffer[off2];
            }
        }

        _charBuffer = newCharBuffer;
        _attrBuffer = newAttrBuffer;
        _w = newWidth;
        _h = newHeight;
        _wh = newBuffSz;
        if (_cursorOffset >= _wh) {
            _cursorOffset = 0;
        }
        maybeFireEvent();
    }

    @Override
    public void setCursorVisible(boolean b) {
        _showCursor = b;
        maybeFireEvent();
    }

    @Override
    public boolean isCursorVisible() {
        return _showCursor;
    }

    @Override
    public void setBG(Paint c) {
        int i;
        for (i = 0; i < _nextFreeColorTableEntry; i++) {
            if (_colorTable[i].equals(c)) {
                break;
            }
        }
        if (i == _nextFreeColorTableEntry) {
            _colorTable[_nextFreeColorTableEntry++] = c;
        }
        _currentAttr &= NOT_BG_MASK;
        _currentAttr |= i << BG_OFFSET;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Paint getBG() {
        return bgOf(_currentAttr);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Paint getFG() {
        return fgOf(_currentAttr);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setFG(Paint c) {
        int i;
        for (i = 0; i < _nextFreeColorTableEntry; i++) {
            if (_colorTable[i].equals(c)) {
                break;
            }
        }
        if (i == _nextFreeColorTableEntry) {
            _colorTable[_nextFreeColorTableEntry++] = c;
        }
        _currentAttr &= NOT_FG_MASK;
        _currentAttr |= i;
    }

    @Override
    public void clear() {
        final Paint c1 = bgOf(_currentAttr);
        final Paint c2 = fgOf(_currentAttr);
        _colorTable[0] = c1;
        _colorTable[1] = c2;
        _nextFreeColorTableEntry = 2;
        _currentAttr = 0x00000001;

        Arrays.fill(_charBuffer, ' ');
        Arrays.fill(_attrBuffer, 0x00000001);

        maybeFireEvent();
    }

    @Override
    public void moveTo(int x, int y) {
        _cursorOffset = getOffset(x, y);
        maybeFireEvent();
    }

    @Override
    public void print(String s) {
        print(s.toCharArray(), 0, s.length());
        maybeFireEvent();
    }

    @Override
    public void println(String s) {
        print(s.toCharArray(), 0, s.length());
        _cursorOffset += _w;
        _cursorOffset -= _cursorOffset % _w;
        if (_cursorOffset >= _wh) {
            System.arraycopy(_charBuffer, _w, _charBuffer, 0, _w * (_h - 1));
            System.arraycopy(_attrBuffer, _w, _attrBuffer, 0, _w * (_h - 1));
            _cursorOffset -= _w;
            for (int i = 0; i < _w; i++) {
                _charBuffer[_cursorOffset + i] = ' ';
                _attrBuffer[_cursorOffset + i] = _currentAttr;
            }
        }
        maybeFireEvent();
        /*
         y++;
         x=0;
         */
    }

    @Override
    public void print(char[] bytes, int offset, int len) {
        if (_cursorOffset + len >= _wh) {
            len = _wh - _cursorOffset;
        }
        for (int i = 0; i < len; i++) {
            _charBuffer[_cursorOffset + i] = bytes[offset + i];
            _attrBuffer[_cursorOffset + i] = _currentAttr;
        }
        _cursorOffset += len;
        _cursorOffset %= _wh;
        maybeFireEvent();
    }

    @Override
    public void print(char c) {
        _charBuffer[_cursorOffset] = c;
        _attrBuffer[_cursorOffset] = _currentAttr;
        _cursorOffset++;
        _cursorOffset %= _wh;
        maybeFireEvent();
    }

    @Override
    public int getCursorX() {
        return _cursorOffset % _w;
    }

    @Override
    public int getCursorY() {
        return _cursorOffset / _w;
    }

    /**
     * Convert x,y to offset.
     *
     * @param x X
     * @param y Y
     * @return The offset
     */
    private int getOffset(int x, int y) {
        return ((y * _w) + x) % _wh;
    }

    @Override
    public int getCols() {
        return _w;
    }

    @Override
    public int getRows() {
        return _h;
    }

    @Override
    public char getCharAt(int x, int y) {
        return _charBuffer[getOffset(x, y)];
    }

    @Override
    public Paint getForegroundAt(int x, int y) {
        return fgOf(_attrBuffer[getOffset(x, y)]);
    }

    /**
     * Get the foreground paint (colour) from an attribute.
     *
     * @param attr The attribute
     * @return The paint
     */
    private Paint fgOf(int attr) {
        return _colorTable[attr & FG_MASK];
    }

    /**
     * Get the background paint (colour) fron an attribute.
     *
     * @param attr The attribute
     * @return The paint
     */
    private Paint bgOf(int attr) {
        return _colorTable[(attr & BG_MASK) >> BG_OFFSET];
    }

    @Override
    public Paint getBackgroundAt(int x, int y) {
        return bgOf(_attrBuffer[getOffset(x, y)]);
    }

    @Override
    public boolean isUnderLine(int x, int y) {
        return isUnderLine(getOffset(x, y));
    }

    /**
     * Is the given offset underlined?
     *
     * @param offset The offset
     * @return true if underlined
     */
    private boolean isUnderLine(int offset) {
        return (_attrBuffer[offset] & UL_MASK) != 0;
    }

    @Override
    public void setUnderLine(boolean b) {
        if (b) {
            _currentAttr |= UL_MASK;
        } else {
            _currentAttr &= NOT_UL_MASK;
        }
    }

    public boolean isUnderline() {
        return (_currentAttr & UL_MASK) == UL_MASK;

    }

    @Override
    public void scrollUp(int numLines) {
        final int amt = numLines * _w;
        System.arraycopy(_charBuffer, amt, _charBuffer, 0, _wh - amt);
        System.arraycopy(_attrBuffer, amt, _attrBuffer, 0, _wh - amt);
        for (int i = 0; i < amt; i++) {
            _charBuffer[_wh - amt + i] = ' ';
            _attrBuffer[_wh - amt + i] = _currentAttr;
        }
        maybeFireEvent();
    }

    @Override
    public void scrollDown(int numLines) {
        final int amt = numLines * _w;
        System.arraycopy(_charBuffer, 0, _charBuffer, amt, _wh - amt);
        System.arraycopy(_attrBuffer, 0, _attrBuffer, amt, _wh - amt);
        for (int i = 0; i < amt; i++) {
            _charBuffer[i] = ' ';
            _attrBuffer[i] = _currentAttr;
        }
        maybeFireEvent();
    }
}
