/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: StackedCursesModel.java 285 2011-07-19 12:37:42Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package uk.co.ruelgnoj.orion.cc;

import java.awt.Paint;

/**
 * A model that is a vertically stacked version of the 2 supplied models.
 * @author Steve Millington
 * @version $Id: StackedCursesModel.java 285 2011-07-19 12:37:42Z steve $
 */
public class StackedCursesModel extends AbstractCursesModel {
    private final CursesModel m1;
    private final CursesModel m2;
    /**
     * Constructor.
     * @param m1 The topmost model in the stack.
     * @param m2 The bottommost model in the stack.
     * @throws java.lang.NullPointerException if either argument is null.
     * @throws java.lang.IllegalArgumentException if the number of rows in the
     * 2 models is not the same.
     */
    public StackedCursesModel(CursesModel m1, CursesModel m2) {
	this.m1 = m1;
	this.m2 = m2;
	if (m1.getCols() != m2.getCols()) {
	    throw new IllegalArgumentException("Rows not same");
	}
	final Listener listener = new Listener();
	m1.addModelListener(listener);
	m2.addModelListener(listener);
    }

    @Override
    public boolean isCursorVisible() {
	return m1.isCursorVisible() | m2.isCursorVisible();
    }

    @Override
    public boolean isUnderLine(int x, int y) {
	if (y >= m1.getRows()) {
	    return m2.isUnderLine(x, y - m1.getRows());
	} else {
	    return m1.isUnderLine(x, y);
	}
    }

    @Override
    public Paint getBackgroundAt(int x, int y) {
	if (y >= m1.getRows()) {
	    return m2.getBackgroundAt(x, y - m1.getRows());
	} else {
	    return m1.getBackgroundAt(x, y);
	}
    }

    @Override
    public Paint getForegroundAt(int x, int y) {
	if (y >= m1.getRows()) {
	    return m2.getForegroundAt(x, y - m1.getRows());
	} else {
	    return m1.getForegroundAt(x, y);
	}
    }

    @Override
    public char getCharAt(int x, int y) {
	if (y >= m1.getRows()) {
	    return m2.getCharAt(x, y - m1.getRows());
	} else {
	    return m1.getCharAt(x, y);
	}
    }

    @Override
    public int getCols() {
	return m1.getCols();
    }

    @Override
    public int getRows() {
	return m1.getRows() + m2.getRows();
    }

    @Override
    public int getCursorY() {
	//
	// Show the cursor in the lowest segment
	//
	if (m2.isCursorVisible()) {
	    return m2.getCursorY() + m1.getRows();
	} else if (m1.isCursorVisible()) {
	    return m1.getCursorY();
	}
	return 0;
    }

    @Override
    public int getCursorX() {
	//
	// Show the cursor in the lowest segment
	//
	if (m2.isCursorVisible()) {
	    return m2.getCursorX();
	} else if (m1.isCursorVisible()) {
	    return m1.getCursorX();
	}
	return 0;
    }

    /**
     * An adapter to fire events to our listeners when our children
     * fire theirs.
     */
    private class Listener implements CursesModelListener {
	/**
	 * Constructor.
	 */
	public Listener() {
	}

	@Override
        public void dataChanged(CursesModelEvent e) {
	    maybeFireEvent();
	}
    }
}
