/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: WritableCursesModel.java 256 2008-07-15 15:54:02Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.cc;

import java.awt.Paint;

/**
 * A curses data model that is modifiable is a curses style.
 *
 * @author Steve Millington
 * @version $Id: WritableCursesModel.java 256 2008-07-15 15:54:02Z steve $
 */
public interface WritableCursesModel extends CursesModel {

    /**
     * Set the batch status. In order to optimise event generation, a system of
     * batch processing is provided. This means that many modifications can be
     * made to the model with only a single event being generated towards the
     * end of the processing.
     * <p>
     * One might use this as in the following example.
     * <pre>
     * wcm.setBatch(true);
     * wcm.clear();
     * wcm.println("Hello World");
     * wcm.println("From Steve");
     * wcm.setBatch(false);
     * </pre>
     * <p>
     * In this example, the model modification event is <b>only</b> generated as
     * part of the last line of code.
     *
     * @param b True to turn batch on, false off.
     */
    void setBatch(boolean b);

    /**
     * Set the cursor visibility.
     *
     * @param b true to make the cursor visible.
     */
    void setCursorVisible(boolean b);

    /**
     * Set the current background color.
     *
     * @param c The background color.
     */
    void setBG(Paint c);

    /**
     * Set the current foreground color.
     *
     * @param c The foreground color.
     */
    void setFG(Paint c);

    /**
     * Get the current background.
     *
     * @return the background.
     */
    Paint getBG();

    /**
     * Get the current foreground.
     *
     * @return the foreground.
     */
    Paint getFG();

    /**
     * Clear the model to the current background color.
     */
    void clear();

    /**
     * Move the cursor to the specified location.
     *
     * @param x The x location.
     * @param y The y location.
     */
    void moveTo(int x, int y);

    /**
     * Print the supplied string at the current location. Wrapping occurs as
     * necessary.
     *
     * @param s The string to print.
     */
    void print(String s);

    /**
     * Print the supplied string at the current location and move the cursor to
     * the start of the subsequent line. Wrapping occurs as necessary.
     *
     * @param s The string to print.
     */
    void println(String s);

    /**
     * Print the supplied characters at the current location. Wrapping occurs as
     * necessary.
     *
     * @param chars The chars to print.
     * @param offset The start offset into the <code>chars</code> buffer.
     * @param len The number of characters.
     */
    void print(char[] chars, int offset, int len);

    /**
     * Print the supplied character at the current location. Wrapping occurs as
     * necessary.
     *
     * @param c The character to print.
     */
    void print(char c);

    /**
     * Set the current underline state.
     *
     * @param b True to underline
     */
    void setUnderLine(boolean b);

    /**
     * Scroll the whole space up by the specified number of lines.
     *
     * @param numLines The amount to scroll.
     */
    void scrollUp(int numLines);

    /**
     * Scroll the whole space down by the specified number of lines.
     *
     * @param numLines The amount to scroll.
     */
    void scrollDown(int numLines);
    
    /**
     * Resize the model.
     * <p>
     * This may cause data loss if changing to a smaller size.
     *
     * @param newWidth The new width
     * @param newHeight The new height
     */
    void setSize(int newWidth, int newHeight);
}
