/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: BasicCursesUI.java 256 2008-07-15 15:54:02Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.cc.basic;

import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Paint;
import java.awt.Point;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.JComponent;
import javax.swing.plaf.ComponentUI;

import uk.co.ruelgnoj.orion.cc.Curses;
import uk.co.ruelgnoj.orion.cc.CursesModel;
import uk.co.ruelgnoj.orion.cc.CursesModelEvent;
import uk.co.ruelgnoj.orion.cc.CursesModelListener;
import uk.co.ruelgnoj.orion.cc.CursesUI;

/**
 * Generic UI implementation for a curses component.
 *
 * @author Steve Millington
 * @version $Id: BasicCursesUI.java 256 2008-07-15 15:54:02Z steve $
 */
public class BasicCursesUI extends CursesUI implements CursesModelListener, PropertyChangeListener {

    private Curses cc;
    private CursesModel model;
    //
    // Cached values to optimize painting.
    //
    private int charW;
    private int charH;
    private int charA;

    /**
     * Constructor.
     */
    public BasicCursesUI() {
    }

    /**
     * Create a UI for a component.
     *
     * @param c The component (must be a Curses)
     * @return A UI
     */
    public static ComponentUI createUI(JComponent c) {
        return new BasicCursesUI();
    }

    @Override
    public void installUI(JComponent c) {
        this.cc = (Curses) c;
        model = cc.getModel();

        cacheValues();		// Remember some stuff to optimize painting

        cc.addPropertyChangeListener(this);
        model.addModelListener(this);
    }

    @Override
    public void uninstallUI(JComponent c) {
        cc.removePropertyChangeListener(this);
        model.removeModelListener(this);
    }

    @Override
    public void propertyChange(PropertyChangeEvent e) {
        if (null != e.getPropertyName()) {
            switch (e.getPropertyName()) {
                case "model":
                    model.removeModelListener(this);
                    model = cc.getModel();
                    model.addModelListener(this);
                    break;
                case "font":
                    cacheValues();	// Remember some stuff to optimize painting
                    break;
                default:
                    break;
            }
        }
    }

    @Override
    public void dataChanged(CursesModelEvent e) {
        cc.repaint();
    }

    /**
     * Draw a bunch of characters.
     *
     * @param g The graphics
     * @param x The buffer x position
     * @param y The buffer y position
     * @param len The number of chars to draw
     * @param fg The foreground paint (color)
     * @param bg The background paint (color)
     */
    private void drawChars(Graphics2D g,
            int x,
            int y,
            int len,
            Paint fg,
            Paint bg) {
        if (bg != null) {
            g.setPaint(bg);
            g.fillRect(x * charW, y * charH, len * charW, charH);
        }
        for (int i = 0; i < len; i++) {
            if (model.getCharAt(x + i, y) != ' ') {
                x += i;
                len -= i;
                break;
            }
        }
        for (int i = len - 1; i >= 0; i--) {
            if (model.getCharAt(x + i, y) != ' ') {
                len = i + 1;
                break;
            } else if (i == 0) {
                len = 0;
            }
        }

        if (len > 0) {
            final StringBuilder sb = new StringBuilder();
            for (int i = 0; i < len; i++) {
                sb.append(model.getCharAt(x + i, y));
            }

            g.setPaint(fg);
            if (model.isUnderLine(x, y)) {
                g.drawLine(x * charW, y * charH + charA, (x + len) * charW, y * charH + charA);
            }

            g.drawString(sb.toString(), x * charW, y * charH + charA);
        }
    }

    /**
     * Redraw a single character position.
     *
     * @param g The graphics to render into
     * @param x The x coord to draw
     * @param y The y coord to draw
     */
    private void redrawSingle(Graphics2D g, int x, int y) {
        Paint fg = model.getForegroundAt(x, y);
        Paint bg = model.getBackgroundAt(x, y);
        if (model.isCursorVisible()
                && (model.getCursorX() == x)
                && (model.getCursorY() == y)) {
            final Paint c = bg;
            bg = fg;
            fg = c;
        }
        drawChars(g, x, y, 1, fg, bg);
    }

    /**
     * Cache the information about character size etc.
     */
    private void cacheValues() {
        final Font font = cc.getFont();
        if (font != null) {
            final FontMetrics fm = cc.getToolkit().getFontMetrics(font);
            charW = fm.charWidth('W');
            charH = fm.getHeight();
            charA = fm.getAscent();
        }
    }

    @Override
    public void paint(Graphics g1d, JComponent jc) {
        final Insets insets = jc.getInsets();
        final Graphics2D g = (Graphics2D) g1d;
        g.translate(insets.left, insets.top);
        final int w = model.getCols();
        final int h = model.getRows();
        Paint saveBg;
        Paint saveFg;
        Paint currentBg;
        Paint currentFg;
        boolean saveUnderline;
        boolean currentUnderline;
        int saveX;

        final Paint baseBg = model.getBackgroundAt(0, 0);

        g.setPaint(baseBg);
        g.fillRect(0, 0, charW * w, charH * h);

        for (int i = 0; i < h; i++) {
            saveBg = model.getBackgroundAt(0, i);
            saveFg = model.getForegroundAt(0, i);
            saveUnderline = model.isUnderLine(0, i);
            saveX = 0;
            for (int j = 0; j < w; j++) {
                currentBg = model.getBackgroundAt(j, i);
                currentFg = model.getForegroundAt(j, i);
                currentUnderline = model.isUnderLine(j, i);
                if ((currentBg != saveBg)
                        || (currentFg != saveFg)
                        || (currentUnderline != saveUnderline)) {
                    drawChars(g,
                            saveX,
                            i,
                            j - saveX,
                            saveFg,
                            (saveBg == baseBg) ? null : saveBg);
                    saveX = j;
                    saveBg = currentBg;
                    saveFg = currentFg;
                    saveUnderline = currentUnderline;
                }
            }
            drawChars(g,
                    saveX,
                    i,
                    w - saveX,
                    saveFg,
                    (saveBg == baseBg) ? null : saveBg);
        }
        redrawSingle(g, model.getCursorX(), model.getCursorY());

        //
        // Restore any graphics transforms
        //
        g.translate(insets.left, insets.top);
    }

    @Override
    public Dimension getPreferredSize(JComponent c) {
        final Insets insets = c.getInsets();
        final Dimension preferredSize = new Dimension(
                (model.getCols() * charW) + insets.left + insets.right,
                (model.getRows() * charH) + insets.top + insets.bottom);
        return preferredSize;
    }

    @Override
    public int getRow(Point p) {
        cacheValues();
        final Insets insets = cc.getInsets();
        return ((p.y - insets.top) / charH) + 1;
    }

    @Override
    public int getColumn(Point p) {
        cacheValues();
        final Insets insets = cc.getInsets();
        return ((p.x - insets.left) / charW) + 1;
    }
}
