/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: FileListView.java 270 2008-11-06 09:56:06Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.component;

import uk.co.ruelgnoj.orion.dm.OrionFileEntry;
import java.awt.Color;

import java.awt.event.KeyEvent;
import java.util.List;

import java.text.MessageFormat;

import java.util.Collections;
import java.util.Comparator;
import java.util.ResourceBundle;
import javax.swing.AbstractListModel;

import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.KeyStroke;
import uk.co.ruelgnoj.orion.cc.WritableCursesModel;
import uk.co.ruelgnoj.orion.component.util.MethodAction;

public class FileListView extends ListView {

    private static final ListViewRenderer RENDERER = new OrionFileEntryRenderer();
    private static final ResourceBundle BUNDLE = ResourceBundle.getBundle("uk.co.ruelgnoj.orion.component.resources");
    private final String srch;

    public FileListView(OrionComponent orionComponent, String srch, List<OrionFileEntry> files) {
        super(orionComponent, new Model(files));
        setRenderer(RENDERER);
        this.srch = srch;
    }
    private static final Color BASE_COLOR = Color.WHITE;
    private static final Color FILENAME_COLOR = Color.YELLOW;
    private static final Color HITCOUNT_COLOR = Color.CYAN;

    private static class OrionFileEntryRenderer implements ListViewRenderer {

        public OrionFileEntryRenderer() {
        }

        @Override
        public void render(WritableCursesModel wtm,
                ListView view,
                Object o,
                int modelIndex,
                boolean selected) {
            final OrionFileEntry hfe = (OrionFileEntry) o;
            String hitCountString = null;
            int hitCountStart = wtm.getCols();
            if (hfe.getHitCount() >= 0) {
                Object[] args = {
                    hfe.getHitCount()
                };
                hitCountString = MessageFormat.format(
                        view.orionComponent.getBundle().getString(
                                "msg.hitcount"),
                        args);
                hitCountStart -= hitCountString.length();
            }
            String fileName = hfe.getName();
            int nameLength = fileName.length();
            if (nameLength > hitCountStart - 4) {
                fileName = "..." + fileName.substring(
                        8 + nameLength - hitCountStart);
                nameLength = fileName.length();
            }
            final String pathTrunc = Utils.truncateFile(
                    hfe.getPath(), hitCountStart - nameLength - 1);
            if (!selected) {
                wtm.setFG(BASE_COLOR);
            } else {
                wtm.setFG(wtm.getBG());
                wtm.setBG(BASE_COLOR);
            }
            wtm.print(pathTrunc);
            if (!selected) {
                wtm.setFG(FILENAME_COLOR);
            } else {
                wtm.setBG(FILENAME_COLOR);
            }
            wtm.print(fileName);
            if (hitCountString != null) {
                wtm.moveTo(hitCountStart, wtm.getCursorY());
                if (!selected) {
                    wtm.setFG(HITCOUNT_COLOR);
                } else {
                    wtm.setBG(HITCOUNT_COLOR);
                }
                wtm.print(hitCountString);
            }
        }
    }

    private static class Model extends AbstractListModel {

        private final List<OrionFileEntry> fileList;

        public Model(List<OrionFileEntry> fileList) {
            this.fileList = fileList;
        }

        @Override
        public int getSize() {
            return fileList.size();
        }

        @Override
        public Object getElementAt(int index) {
            return fileList.get(index);
        }

        private void sortByName() {
            Collections.sort(fileList, new ByNameComparator());
            fireContentsChanged(this, 0, fileList.size() - 1);
        }

        private void sortByHits() {
            Collections.sort(fileList, new ByHitsComparator());
            fireContentsChanged(this, 0, fileList.size() - 1);
        }

        private void sortByPath() {
            Collections.sort(fileList, new ByPathComparator());
            fireContentsChanged(this, 0, fileList.size() - 1);
        }

        private void sortBySuffix() {
            Collections.sort(fileList, new BySuffixComparator());
            fireContentsChanged(this, 0, fileList.size() - 1);
        }

        private class ByNameComparator implements Comparator<OrionFileEntry> {

            @Override
            public int compare(OrionFileEntry o1, OrionFileEntry o2) {
                return o1.getName().compareTo(o2.getName());
            }
        }

        private class ByPathComparator implements Comparator<OrionFileEntry> {

            @Override
            public int compare(OrionFileEntry o1, OrionFileEntry o2) {
                int ret = o1.getPath().compareToIgnoreCase(o2.getPath());
                if (ret == 0) {
                    ret = o1.getFileName().compareToIgnoreCase(o2.getFileName());
                }
                return ret;
            }
        }

        private class BySuffixComparator implements Comparator<OrionFileEntry> {

            @Override
            public int compare(OrionFileEntry o1, OrionFileEntry o2) {
                final String s1 = o1.getName().substring(o1.getName().lastIndexOf('.') + 1);
                final String s2 = o2.getName().substring(o2.getName().lastIndexOf('.') + 1);
                return s1.compareToIgnoreCase(s2);
            }
        }

        private class ByHitsComparator implements Comparator<OrionFileEntry> {

            @Override
            public int compare(OrionFileEntry o1, OrionFileEntry o2) {
                return o2.getHitCount() - o1.getHitCount();
            }
        }
    }

    @Override
    protected void itemSelected(Object o) {
        final OpenFileSwingWorker ofsw = new OpenFileSwingWorker(orionComponent, ((OrionFileEntry) o).getFileName(), srch);
        ofsw.execute();
    }

    @Override
    public void activate() {
        final WritableCursesModel header = orionComponent.getHeaderArea();
        header.setBatch(true);
        header.setBG(Color.BLACK);
        header.clear();
        header.moveTo(0, 0);
        header.setFG(Color.WHITE);
        header.print(orionComponent.getBundle().getString("msg.searchfor") + ' ');
        header.setFG(Color.YELLOW);
        header.print(srch);
        header.setBatch(false);

        super.activate();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void initializeInputMap(InputMap im) {
        super.initializeInputMap(im);
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_S, KeyEvent.ALT_MASK, false), "sortBySuffix");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_N, KeyEvent.ALT_MASK, false), "sortByName");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_P, KeyEvent.ALT_MASK, false), "sortByPath");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_H, KeyEvent.ALT_MASK, false), "sortByHits");
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void initializeActionMap(ActionMap am) {
        super.initializeActionMap(am);
        am.put("sortBySuffix", new MethodAction(BUNDLE, "actions.filelistview", "sortBySuffix", this, "sortBySuffix"));
        am.put("sortByName", new MethodAction(BUNDLE, "actions.filelistview", "sortByName", this, "sortByName"));
        am.put("sortByHits", new MethodAction(BUNDLE, "actions.filelistview", "sortByHits", this, "sortByHits"));
        am.put("sortByPath", new MethodAction(BUNDLE, "actions.filelistview", "sortByPath", this, "sortByPath"));
    }

    public void sortByName() {
        ((Model) getModel()).sortByName();
    }

    public void sortByPath() {
        ((Model) getModel()).sortByPath();
    }

    public void sortBySuffix() {
        ((Model) getModel()).sortBySuffix();
    }

    public void sortByHits() {
        ((Model) getModel()).sortByHits();
    }
}
