/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: FileViewSettings.java 256 2008-07-15 15:54:02Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.component;

import java.awt.Color;

import java.util.HashMap;
import java.util.Map;
import java.util.ResourceBundle;

import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyleContext;

import uk.co.ruelgnoj.orion.tokenizers.LanguageTokenizer;
import uk.co.ruelgnoj.orion.tokenizers.TokenType;
import uk.co.ruelgnoj.orion.tokenizers.TokenizerFactory;

/**
 * The configuration settings for a file view.
 *
 * @author Steve Millington
 * @version $Id: FileViewSettings.java 256 2008-07-15 15:54:02Z steve $
 */
public class FileViewSettings implements Comparable<FileViewSettings>, Cloneable {

    private static final int BASE_16 = 16;
    private static final int DEFAULT_TABSTOP = 8;

    private static final Map<String, Color> COLOR_COLORS;
    private static final Map<Color, String> COLOR_STRINGS;

    //
    // Static initializer of the 2 hash tables
    //
    static {
        COLOR_COLORS = new HashMap<>();
        COLOR_COLORS.put("black", Color.BLACK);
        COLOR_COLORS.put("blue", Color.BLUE);
        COLOR_COLORS.put("cyan", Color.CYAN);
        COLOR_COLORS.put("darkgray", Color.DARK_GRAY);
        COLOR_COLORS.put("gray", Color.GRAY);
        COLOR_COLORS.put("green", Color.GREEN);
        COLOR_COLORS.put("lightgray", Color.LIGHT_GRAY);
        COLOR_COLORS.put("magenta", Color.MAGENTA);
        COLOR_COLORS.put("orange", Color.ORANGE);
        COLOR_COLORS.put("pink", Color.PINK);
        COLOR_COLORS.put("red", Color.RED);
        COLOR_COLORS.put("white", Color.WHITE);
        COLOR_COLORS.put("yellow", Color.YELLOW);

        COLOR_STRINGS = new HashMap<>();
        COLOR_STRINGS.put(Color.BLACK, "black");
        COLOR_STRINGS.put(Color.BLUE, "blue");
        COLOR_STRINGS.put(Color.CYAN, "cyan");
        COLOR_STRINGS.put(Color.DARK_GRAY, "darkgray");
        COLOR_STRINGS.put(Color.GRAY, "gray");
        COLOR_STRINGS.put(Color.GREEN, "green");
        COLOR_STRINGS.put(Color.LIGHT_GRAY, "lightgray");
        COLOR_STRINGS.put(Color.MAGENTA, "magenta");
        COLOR_STRINGS.put(Color.ORANGE, "orange");
        COLOR_STRINGS.put(Color.PINK, "pink");
        COLOR_STRINGS.put(Color.RED, "red");
        COLOR_STRINGS.put(Color.WHITE, "white");
        COLOR_STRINGS.put(Color.YELLOW, "yellow");
    }

    private final String tokenizerName;
    private final StyleContext styleContext;

    private final String visual;
    private int tabstop;

    /**
     * Constructor.
     *
     * @param tokenizerName The name of the tokenizer these settings are for.
     */
    public FileViewSettings(String tokenizerName) {
        styleContext = new StyleContext();
        initializeStyles();
        this.tokenizerName = tokenizerName;
        final ResourceBundle rb = ResourceBundle.getBundle("uk.co.ruelgnoj.orion.component.resources");
        this.visual = rb.getString("language." + tokenizerName + ".title");
    }

    public String getVisual() {
        return visual;
    }

    public int getTabstop() {
        return tabstop;
    }

    @Override
    public int compareTo(FileViewSettings that) {
        return that.visual.compareTo(visual);
    }

    private void initializeStyles() {
        final Style root = styleContext.addStyle(StyleContext.DEFAULT_STYLE, null);
        StyleConstants.setForeground(root, Color.WHITE);
        StyleConstants.setBackground(root, Color.BLACK);
        Style style;
        style = styleContext.addStyle("plaintext", root);
        style = styleContext.addStyle("whitespace", root);
        style = styleContext.addStyle("newline", style);
        style = styleContext.addStyle("comment", root);
        StyleConstants.setForeground(style, Color.GREEN);
        style = styleContext.addStyle("punctuation", root);
        StyleConstants.setForeground(style, Color.PINK);
        style = styleContext.addStyle("string", root);
        StyleConstants.setForeground(style, Color.CYAN);
        style = styleContext.addStyle("word", root);
        StyleConstants.setForeground(style, Color.WHITE);
        style = styleContext.addStyle("indexed", style);
        StyleConstants.setBackground(style, Color.RED.darker());
        style = styleContext.addStyle("keyword", root);
        StyleConstants.setForeground(style, Color.YELLOW);
        style = styleContext.addStyle("number", root);
        StyleConstants.setForeground(style, Color.RED);
        style = styleContext.addStyle("character", root);
        StyleConstants.setForeground(style, Color.ORANGE);
        style = styleContext.addStyle("linenumber", root);
        StyleConstants.setForeground(style, Color.LIGHT_GRAY);
        style = styleContext.addStyle("error", root);
        StyleConstants.setForeground(style, Color.BLUE);
        style = styleContext.addStyle("cicsapi", root);
        StyleConstants.setForeground(style, Color.CYAN.darker());
        style = styleContext.addStyle("sqlapi", root);
        StyleConstants.setForeground(style, Color.GREEN.darker());
    }

    public void load(Map<String, String> props) {
        final LanguageTokenizer tokenizer
                = TokenizerFactory.getSharedInstance().getTokenizerByName(tokenizerName);
        tabstop = DEFAULT_TABSTOP;
        for (TokenType tt : tokenizer.getPossibleTokenTypes()) {
            final String name = tt.getName();

            loadStyle(props, name);
        }
    }

    public LanguageTokenizer getTokenizer() {
        return TokenizerFactory.getSharedInstance().getTokenizerByName(tokenizerName);
    }

    private Color getFGColor(Map<String, String> props, String s) {
        return getColor(props, "fg." + s);
    }

    private Color getBGColor(Map<String, String> props, String s) {
        return getColor(props, "bg." + s);
    }

    private Color getColor(Map<String, String> props, String s) {
        s = "orion.sourceview." + tokenizerName + ".color." + s;
        final String n = props.get(s);
        if (n == null) {
            return null;
        }
        final Color c = getColor(n);
        if (c == null) {
            return null;
        }
        return c;
    }

    private void putFGColor(Map<String, String> props, String colorName, Color value) {
        putColor(props, "fg." + colorName, value);
    }

    private void putBGColor(Map<String, String> props, String colorName, Color value) {
        putColor(props, "bg." + colorName, value);
    }

    private void putColor(Map<String, String> props, String colorName, Color value) {
        props.put("orion.sourceview." + tokenizerName + ".color." + colorName, getString(value));
    }

    public void save(Map<String, String> props) {
        // System.out.println("FVS.save()");
        final LanguageTokenizer tokenizer
                = TokenizerFactory.getSharedInstance().getTokenizerByName(tokenizerName);
        for (TokenType tt : tokenizer.getPossibleTokenTypes()) {
            final String name = tt.getName();

            saveStyle(props, name);
        }
    }

    private void saveStyle(Map<String, String> props, String name) {
        final Style style = styleContext.getStyle(name);
        putBGColor(props, name, StyleConstants.getBackground(style));
        putFGColor(props, name, StyleConstants.getForeground(style));
    }

    private void loadStyle(Map<String, String> props, String name) {
        final Style style = styleContext.getStyle(name);
        Color c;
        c = getFGColor(props, name);
        if (c != null) {
            StyleConstants.setForeground(style, c);
        }
        c = getBGColor(props, name);
        if (c != null) {
            StyleConstants.setBackground(style, c);
        }
    }

    @Override
    public FileViewSettings clone() {
        try {
            return (FileViewSettings) super.clone();
        } catch (CloneNotSupportedException e) {
            throw new Error(e);
        }
    }

    public StyleContext getStyleContext() {
        return styleContext;
    }

    /**
     * Convert a supplied color string to a color object.
     *
     * @param key The color string
     * @return The color
     */
    private static Color getColor(String key) {
        final String lowerCaseKey = key.toLowerCase();

        if (COLOR_COLORS.containsKey(lowerCaseKey)) {
            return COLOR_COLORS.get(lowerCaseKey);
        }
        try {
            final int value;
            if (lowerCaseKey.startsWith("0x")) {
                value = Integer.parseInt(lowerCaseKey.substring(2), BASE_16);
            } else {
                value = Integer.parseInt(lowerCaseKey);
            }
            return new Color(value);
        } catch (NumberFormatException e) {
            return null;
        }
    }

    /**
     * Convert the supplied color to a string.
     * <p>
     * The returned string is of the form <code>0xaarrggbb</code>, or one of the
     * pre-defined color names.
     *
     * @param key The color to convert
     * @return The string representation
     */
    private static String getString(Color key) {
        if (COLOR_STRINGS.containsKey(key)) {
            return COLOR_STRINGS.get(key);
        }
        return "0x" + Integer.toString(key.getRGB() & 0x00ffffff, BASE_16);
    }
}
