/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: ListView.java 270 2008-11-06 09:56:06Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package uk.co.ruelgnoj.orion.component;

import java.awt.Color;
import java.awt.Paint;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.text.MessageFormat;
import java.util.List;
import javax.swing.AbstractAction;
import javax.swing.AbstractListModel;
import javax.swing.Action;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.KeyStroke;
import javax.swing.ListModel;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;
import uk.co.ruelgnoj.orion.cc.WritableCursesModel;

/**
 * A view of a list of things.
 * This class is designed to be subclasses for particular types of list.
 *
 * @author Steve Millington
 * @version $Id: ListView.java 270 2008-11-06 09:56:06Z steve $
 */
public abstract class ListView extends AbstractView {

    private final MouseListener ml;
    private final MouseWheelListener mwl;
    private final ActionMapHelper amh;
    private int topIndex;
    private int selectedModelIndex;
    private ListModel model;
    private Paint background;
    private ListViewRenderer renderer;
    private boolean displayEnabled;
    private ListDataListener listListener;

    /**
     * Constructor.
     * @param orionComponent The component that the view is displayed on.
     */
    private ListView(OrionComponent orionComponent) {
        super(orionComponent);
        displayEnabled = true;
        background = Color.BLACK;
        this.renderer = new DefaultListViewRenderer();
        this.topIndex = 0;
        this.selectedModelIndex = 0;

        amh = new ActionMapHelper(orionComponent);
        ml = new ListViewMouseListener();
        mwl = (MouseWheelEvent e) -> {
            String direction;
            String amount;
            int notches = e.getWheelRotation();
            if (notches < 0) {
                direction = "Up";
                notches = -notches;
            } else {
                direction = "Down";
            }
            if (e.getScrollType() == MouseWheelEvent.WHEEL_UNIT_SCROLL) {
                amount = "viewport";
            } else { // Scroll Type == MouseWheelEvent.WHEEL_BLOCK_SCROLL) {
                amount = "page";
            }
            final String function = amount + direction;
            enableDisplay(false);
            for (int i = 0; i < notches; i++) {
                if (i == notches - 1) {
                    enableDisplay(true);
                }
                amh.invokeAction(function);
            }
        };
        
        listListener = new ListDataListener() {

            @Override
            public void intervalAdded(ListDataEvent e) {
                displayAll();
            }

            @Override
            public void intervalRemoved(ListDataEvent e) {
                displayAll();
            }

            @Override
            public void contentsChanged(ListDataEvent e) {
                displayAll();
            }
        };

    }

    private int modelToView(int modelIndex) {
        if (modelIndex < topIndex) {
            return -1;
        }
        if (modelIndex >= topIndex + orionComponent.getMainArea().getRows()) {
            return -1;
        }

        return modelIndex - topIndex;
    }

    public ListView(OrionComponent orionComponent, ListModel model) {
        this(orionComponent);
        this.model = model;
        this.model.addListDataListener(listListener);
    }

    public ListView(OrionComponent orionComponent, List<String>  list) {
        this(orionComponent);
        this.model = new SimpleModel(list);
        this.model.addListDataListener(listListener);
    }

    private void enableDisplay(boolean b) {
        displayEnabled = b;
    }

    /** {@inheritDoc} */
    @Override
    protected void initializeInputMap(InputMap im) {
        im.put(KeyStroke.getKeyStroke('H'), "hunt");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0, false), "lineDown");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_UP, 0, false), "lineUp");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_DOWN, 0, false), "pageDown");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_UP, 0, false), "pageUp");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_F, KeyEvent.CTRL_MASK, false), "pageDown");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_B, KeyEvent.CTRL_MASK, false), "pageUp");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_D, KeyEvent.CTRL_MASK, false), "debug");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0, false), "select");
    }

    /** {@inheritDoc} */
    @Override
    protected void initializeActionMap(ActionMap am) {
        am.put("pageDown", new PageDownAction());
        am.put("pageUp", new PageUpAction());
        am.put("select", new SelectAction());
        am.put("lineUp", new LineUpAction());
        am.put("lineDown", new LineDownAction());
        am.put("viewportUp", new ViewportUpAction());
        am.put("viewportDown", new ViewportDownAction());
    }

    public void setRenderer(ListViewRenderer renderer) {
        this.renderer = renderer;
    }

    public int getTopIndex() {
        return topIndex;
    }

    public ListModel getModel() {
        return model;
    }

    public int getVisibleItemCount() {
        return Math.min(orionComponent.getMainArea().getRows(), model.getSize() - topIndex);
    }

    private void setCurrent(int row) {
        if (row + topIndex >= model.getSize()) {
            throw new IllegalArgumentException();
        }
        if (row + topIndex < 0) {
            throw new IllegalArgumentException();
        }
        final WritableCursesModel mainArea = orionComponent.getMainArea();

        final int oldViewLine = modelToView(selectedModelIndex);
        selectedModelIndex = topIndex + row;
        final int newViewLine = modelToView(selectedModelIndex);


        mainArea.setBatch(true);
        displayLine(oldViewLine);
        displayLine(newViewLine);
        mainArea.setBatch(false);
        displayPos();
        msgNone();
    }

    private boolean isCurrent(int row) {
        return modelToView(selectedModelIndex) == row;
    }

    private void displayPos() {
        final WritableCursesModel header = orionComponent.getHeaderArea();
        header.setBatch(true);
        header.setFG(Color.LIGHT_GRAY);
        String s = orionComponent.getBundle().getString("listview.header");
        final Object[] inserts = {selectedModelIndex + 1, model.getSize()};
        // Add extra space at the start to ensure the overwriting
        // of any spurious characters
        s = ' ' + (new MessageFormat(s, orionComponent.getLocale())).format(inserts, new StringBuffer(), null).toString();
        header.moveTo(header.getCols() - s.length(), 1);
        header.print(s);
        header.setBatch(false);
    }

    protected synchronized void displayAll() {
        if (!displayEnabled) {
            return;
        }
        final WritableCursesModel mainArea = orionComponent.getMainArea();
        mainArea.setBatch(true);
        mainArea.setBG(background);
        mainArea.clear();
        final int rows = getVisibleItemCount();
        for (int i = 0; i < rows; i++) {
            displayLine(i);
        }
        mainArea.setBatch(false);
    }

    private void displayLine(int viewLineNum) {
        final WritableCursesModel mainArea = orionComponent.getMainArea();
        final int modelIndex = viewLineNum + topIndex;

        mainArea.moveTo(1, viewLineNum);
        mainArea.setBG(background);
        renderer.render(mainArea, this, model.getElementAt(modelIndex), modelIndex, modelIndex == selectedModelIndex);
        mainArea.setBG(background);
    }

    @Override
    public void activate() {
        super.activate();

        // This is TERRIBLE - we need a better way!!
        orionComponent.getMousableComponent().addMouseListener(ml);
        orionComponent.getMousableComponent().addMouseWheelListener(mwl);

        displayPos();

        displayAll();
        msgNone();
    }

    @Override
    public void deactivate() {
        super.deactivate();
        // This is TERRIBLE - we need a better way!!
        orionComponent.getMousableComponent().removeMouseListener(ml);
        orionComponent.getMousableComponent().removeMouseWheelListener(mwl);
        orionComponent.getMessenger().info(null);
    }

    protected abstract void itemSelected(Object o);

    /**
     * Action to go to the next line.
     */
    private class LineDownAction extends AbstractAction {

        /**
         * Constructor.
         */
        public LineDownAction() {
        }

        /** {@inheritDoc} */
        @Override
        public void actionPerformed(ActionEvent e) {

            final WritableCursesModel mainArea = orionComponent.getMainArea();
            if (selectedModelIndex == model.getSize() - 1) {
                msgBottom();
                return;
            }
            selectedModelIndex++;
            if (modelToView(selectedModelIndex) == -1) {
                topIndex++;
                mainArea.setBatch(true);
                mainArea.scrollUp(1);
                final int viewRows = orionComponent.getMainArea().getRows();
                displayLine(viewRows - 1);
                displayLine(viewRows - 2);
                mainArea.setBatch(false);
                displayPos();
                msgNone();
            } else {
                mainArea.setBatch(true);
                displayLine(selectedModelIndex - topIndex);
                displayLine(selectedModelIndex - topIndex - 1);
                mainArea.setBatch(false);
                displayPos();
                msgNone();
            }
        }
    }

    private void msgBottom() {
        orionComponent.getMessenger().error(orionComponent.getBundle().getString("msg.bottom"));
    }

    private void msgTop() {
        orionComponent.getMessenger().error(orionComponent.getBundle().getString("msg.top"));
    }

    private void msgNone() {
        orionComponent.getMessenger().info(getNullMessage());
    }

    /**
     * Action to go to the previous line.
     */
    private class LineUpAction extends AbstractAction {

        /**
         * Constructor.
         */
        public LineUpAction() {
        }

        /** {@inheritDoc} */
        @Override
        public void actionPerformed(ActionEvent e) {
            if (selectedModelIndex == 0) {
                msgTop();
                return;
            }
            msgNone();
            selectedModelIndex--;
            if (modelToView(selectedModelIndex) == -1) {
                topIndex--;
                final WritableCursesModel mainArea = orionComponent.getMainArea();
                mainArea.setBatch(true);
                mainArea.scrollDown(1);
                displayLine(0);
                displayLine(1);
                mainArea.setBatch(false);
                displayPos();
            } else {
                final WritableCursesModel mainArea = orionComponent.getMainArea();
                mainArea.setBatch(true);
                displayLine(selectedModelIndex - topIndex);
                displayLine(selectedModelIndex - topIndex + 1); // The old current line
                mainArea.setBatch(false);
                displayPos();
            }
        }
    }

    private class ViewportUpAction extends AbstractAction {

        public ViewportUpAction() {
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            if (selectedModelIndex == 0) {
                msgTop();
                displayAll();
                return;
            }

            msgNone();

            final WritableCursesModel mainArea = orionComponent.getMainArea();
            if (topIndex == 0) {
                selectedModelIndex--;
                mainArea.setBatch(true);
                displayLine(modelToView(selectedModelIndex));
                displayLine(modelToView(selectedModelIndex + 1));
                mainArea.setBatch(false);
                displayPos();
            } else {
                topIndex--;
                mainArea.setBatch(true);
                mainArea.scrollDown(1);
                displayLine(0);
                if (modelToView(selectedModelIndex) == -1) {
                    selectedModelIndex--;
                    displayLine(modelToView(selectedModelIndex));
                    displayPos();
                }
                mainArea.setBatch(false);
            }
        }
    }

    private class ViewportDownAction extends AbstractAction {

        public ViewportDownAction() {
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            if (selectedModelIndex == model.getSize() - 1) {
                msgTop();
                displayAll();
                return;
            }
            msgNone();

            final WritableCursesModel mainArea = orionComponent.getMainArea();

            mainArea.setBatch(true);
            final int viewRows = orionComponent.getMainArea().getRows();
            if (topIndex + viewRows >= model.getSize()) {
                selectedModelIndex++;
                displayLine(modelToView(selectedModelIndex));
                displayLine(modelToView(selectedModelIndex - 1));
                displayPos();
            } else {
                mainArea.scrollUp(1);
                topIndex++;
                if (modelToView(selectedModelIndex) == -1) {
                    selectedModelIndex++;
                    displayLine(modelToView(selectedModelIndex));
                    displayPos();
                }
                displayLine(viewRows - 1);
            }
            
            mainArea.setBatch(false);
        }
    }

    /**
     * Action to select the current line.
     */
    private class SelectAction extends AbstractAction {

        /**
         * Constructor.
         */
        public SelectAction() {
        }

        /** {@inheritDoc} */
        @Override
        public void actionPerformed(ActionEvent e) {
            itemSelected(model.getElementAt(selectedModelIndex));
        }
    }

    /**
     * Action to go to the previous page of items.
     */
    private class PageUpAction extends AbstractAction {

        /**
         * Constructor.
         */
        public PageUpAction() {
        }

        /** {@inheritDoc} */
        @Override
        public void actionPerformed(ActionEvent e) {
            if (topIndex == 0) {
                msgTop();
            } else {
                msgNone();
                final int newTopIndex = Math.max(0, topIndex - orionComponent.getMainArea().getRows());
                final int moved = topIndex - newTopIndex;
                topIndex = newTopIndex;
                selectedModelIndex -= moved;
                displayAll();
                displayPos();
            }
        }
    }

    /**
     * Action to go to the next page of items.
     */
    private class PageDownAction extends AbstractAction {

        /**
         * Constructor.
         */
        public PageDownAction() {
        }

        /** {@inheritDoc} */
        @Override
        public void actionPerformed(ActionEvent e) {
            final int newPlace = topIndex + orionComponent.getMainArea().getRows();
            final int moved = newPlace - topIndex;
            if (newPlace < model.getSize()) {
                topIndex = newPlace;
                selectedModelIndex += moved;
                if (selectedModelIndex >= model.getSize()) {
                    selectedModelIndex = model.getSize() - 1;
                }
                displayAll();
                displayPos();
                msgNone();
            } else {
                msgBottom();
            }
        }
    }

    protected String getNullMessage() {
        return null;
    }

    /**
     * A simple model of the data.
     */
    private class SimpleModel extends AbstractListModel {

        private final List<String>  list;

        /**
         * Constructor.
         * @param list the data for the list.
         */
        public SimpleModel(List<String>  list) {
            this.list = list;
        }

        @Override
        public int getSize() {
            return list.size();
        }

        @Override
        public Object getElementAt(int index) {
            return list.get(index);
        }
    }

    /**
     * Listener to allow mouse-wheel scrolling.
     */
    private class ListViewMouseListener extends MouseAdapter {

        /**
         * Constructor.
         */
        public ListViewMouseListener() {
        }

        /** {@inheritDoc} */
        @Override
        public void mousePressed(MouseEvent e) {
            if ((!e.isPopupTrigger()) && (e.getClickCount() == 1)) {
                final int row = orionComponent.getMainAreaRow(e.getPoint());
                try {
                    setCurrent(row - 1);
                    e.consume();
                } catch (IllegalArgumentException ex) {
                    orionComponent.getMessenger().error(orionComponent.getBundle().getString("msg.invalidrow"));
                }
            }
        }

        /** {@inheritDoc} */
        @Override
        public void mouseClicked(MouseEvent e) {
            if ((!e.isPopupTrigger()) && (e.getClickCount() % 2 == 0)) {
                final int row = orionComponent.getMainAreaRow(e.getPoint());
                //
                // Make sure we only do it if we are on the selected
                // row
                //
                if (isCurrent(row - 1)) {
                    final Action action = orionComponent.getActionMap().get("select");
                    if ((action != null) && action.isEnabled()) {
                        action.actionPerformed(new ActionEvent(e.getSource(), ActionEvent.ACTION_PERFORMED, "select"));
                    }
                }
            }
        }
    }
}
