/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: Messenger.java 256 2008-07-15 15:54:02Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package uk.co.ruelgnoj.orion.component;

import java.awt.Color;
import java.awt.Toolkit;

import uk.co.ruelgnoj.orion.cc.WritableCursesModel;
import uk.co.ruelgnoj.orion.component.util.Activatable;

/**
 * A controller for the status line that can produce messages nicely formatted.
 *
 * @author Steve Millington
 * @version $Id: Messenger.java 256 2008-07-15 15:54:02Z steve $
 */
public class Messenger implements Activatable {
    private static enum Severity {
        INFO,
        WARNING,
        ERROR
    };
    
    private final WritableCursesModel	statusArea;
    private String currentMessage;
    private Severity currentMessageSeverity;
    private boolean active;

    /**
     * Constructor.
     * @param statusArea The are to control.
     */
    public Messenger(WritableCursesModel statusArea) {
	this.statusArea = statusArea;
	currentMessage = null;
        currentMessageSeverity = Severity.INFO;
        active = false;
    }

    /**
     * Replace the current message with the new one.
     * @param s The new message.
     */
    private void msg(String s) {
	//
	// Only do anything if the message is different to the last one
	//
	if (currentMessage == s) {
	    return;
	}

	if ((currentMessage == null)
		|| (s == null)
		|| !currentMessage.equals(s)) {
	    currentMessage = s;
	    redraw();
	}
    }

    /**
     * Redraw the message.
     */
    private void redraw() {
    	statusArea.setBatch(true);
	if (currentMessage != null) {
	    statusArea.setFG(Color.WHITE);
	    switch (currentMessageSeverity) {
	    	case INFO:
		    statusArea.setBG(Color.BLACK);
		    break;
	    	case WARNING:
		    statusArea.setBG(Color.BLUE);
		    break;
	    	case ERROR:
		    statusArea.setBG(Color.RED);
		    break;
	    }
	    statusArea.clear();
	    statusArea.moveTo(0, 0);
	    statusArea.print(currentMessage);
	} else {
	    statusArea.setBG(Color.BLACK);
	    statusArea.clear();
	}
	statusArea.setBatch(false);
    }

    /**
     * Set the current message to be the supplied <i>error</i> string.
     * @param s The message.
     */
    public void error(String s) {
	currentMessageSeverity = Severity.ERROR;
	msg(s);
	Toolkit.getDefaultToolkit().beep();
    }

    /**
     * Set the current message to be the supplied <i>warning</i> string.
     * @param s The message.
     */
    public void warning(String s) {
	currentMessageSeverity = Severity.WARNING;
	msg(s);
    }

    /**
     * Set the current message to be the supplied <i>informational</i> string.
     * @param s The message.
     */
    public void info(String s) {
	currentMessageSeverity = Severity.INFO;
	msg(s);
    }

    /**
     * Clear the message.
     */
    public void clear() {
	currentMessageSeverity = Severity.INFO;
	currentMessage = null;
	redraw();
    }

    @Override
    public void activate() {
        active = true;
    	redraw();
    }

    @Override
    public void deactivate() {
        active = false;
    }
    
    @Override
    public boolean isActive() {
        return active;
    }
}
