/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: OrionComponent.java 282 2009-10-17 08:24:51Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.component;

import uk.co.ruelgnoj.orion.dm.OrionDataModel;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Paint;
import java.awt.Point;

import java.awt.event.ActionEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Collections;

import java.util.HashMap;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.Stack;
import java.util.TreeSet;
import java.util.regex.Pattern;

import javax.swing.AbstractAction;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.JComponent;
import javax.swing.JLayer;
import javax.swing.JPanel;
import javax.swing.KeyStroke;
import javax.swing.plaf.LayerUI;

import uk.co.ruelgnoj.orion.cc.Curses;
import uk.co.ruelgnoj.orion.cc.CursesInputField;
import uk.co.ruelgnoj.orion.cc.CursesModel;
import uk.co.ruelgnoj.orion.cc.DefaultCursesModel;
import uk.co.ruelgnoj.orion.cc.CursesInputFieldListener;
import uk.co.ruelgnoj.orion.cc.StackedCursesModel;
import uk.co.ruelgnoj.orion.cc.WritableCursesModel;

import uk.co.ruelgnoj.orion.component.debug.DebugAction;
import uk.co.ruelgnoj.orion.component.fileview.FileView;
import uk.co.ruelgnoj.orion.component.util.Activatable;

import uk.co.ruelgnoj.orion.tokenizers.TokenizerFactory;

/**
 * This bean is a single component whose visible aspect is to allow the user of
 * it to traverse through a Orion Database.
 *
 * @author Steve Millington
 * @version $Id: OrionComponent.java 282 2009-10-17 08:24:51Z steve $
 */
public class OrionComponent extends JPanel {

    private WritableCursesModel titlesModel;
    private WritableCursesModel mainModel;
    private WritableCursesModel statusModel;
    private Stack<View> views;
    private Curses mainSpace;
    private Activatable currentStatusAreaController;
    private Messenger mess;
    private CursesInputField cif;
    private OrionInputFieldListener ifl;
    private boolean firePropertyChangeEvents;
    private int rows;
    private int columns;
    private Map<String, FileViewSettings> settings;
    private OrionDataModel model;
    private Resizer resizer;
    private Pattern lastSearch = Pattern.compile("");
    private final MyOverlay lui;

    /**
     * Construct a 'default' component.
     */
    public OrionComponent() {
        this(Collections.unmodifiableMap(new HashMap<String, String>()));
    }

    /**
     * Construct a component, specifying its structure from a Properties object.
     *
     * @param props The properties Object
     */
    public OrionComponent(Map<String, String> props) {
        //
        // Do not Use BorderLayout because it calls getPreferredSize too
        // often. This partial BorderLayout is good enough for our
        // purposes.
        //
        this.setLayout(new PartialBorderLayout());

        setFocusable(true);
        lui = new MyOverlay();
        initialize(props);
        // This custom layerUI will fill the layer with translucent green
        // and print out all mouseMotion events generated within its borders

//	mainSpace.setFocusable(true);
    }

    /**
     * Initialize the instance from a bunch of configuration properties.
     *
     * @param props The configuration properties.
     */
    private void initialize(Map<String, String> props) {
        firePropertyChangeEvents = true;

        settings = createSettings();

        views = new Stack<>();
        views.push(new RootView());

        final OrionMouseHandler omh = new OrionMouseHandler();

        configure(props);

        titlesModel = new DefaultCursesModel(columns, 2);
        mainModel = new DefaultCursesModel(columns, rows - 3);
        statusModel = new DefaultCursesModel(columns, 1);
        final CursesModel wholeModel = new StackedCursesModel(
                titlesModel,
                new StackedCursesModel(
                        mainModel,
                        statusModel));
        mainSpace = new Curses(wholeModel);
        mainSpace.setInheritsPopupMenu(true);
        mainSpace.setFont(getFont());
        mainSpace.addMouseListener(omh);

        //-----------------------------------------------------------------
        // Test Code
        //
        //mainSpace.setBorder(javax.swing.BorderFactory.createMatteBorder(
        //	    10,20,30,40,java.awt.Color.GREEN));
        //-----------------------------------------------------------------
        mess = new Messenger(statusModel);
        cif = new CursesInputField(this, statusModel);
        ifl = new OrionInputFieldListener();
        cif.setListener(ifl);
        // create the layer for the panel using our custom layerUI
        final JLayer<Curses> layer = new JLayer<>(mainSpace, lui);

        this.add(BorderLayout.CENTER, layer);

        initializeInputsAndActions();

        setStatusAreaController(mess);

        addComponentListener(new ComponentAdapter() {
            @Override
            public void componentResized(ComponentEvent ce) {
                final Font font = getFont();
                if (font != null) {
                    final FontMetrics fm = getToolkit().getFontMetrics(font);
                    final int charW = fm.charWidth('W');
                    final int charH = fm.getHeight();
                    final Insets insets = getInsets();
                    final Dimension size = getSize();
                    final int cols = (size.width - insets.left - insets.right) / charW;
                    final int rows = (size.height - insets.top - insets.bottom) / charH;
                    titlesModel.setSize(cols, 2);
                    statusModel.setSize(cols, 1);
                    mainModel.setSize(cols, rows - 3);
                }
            }
        });
    }

    public JComponent getMousableComponent() {
        return mainSpace;
    }

    /**
     * Initialize all of the 'global' actions and their key mappings.
     */
    private void initializeInputsAndActions() {
        final InputMap iMap = getInputMap();
        iMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), "back");
        iMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_F4, 0), "debug");
        iMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_F3, 0), "back");
        iMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_R, InputEvent.CTRL_DOWN_MASK), "backAll");

        final ActionMap aMap = getActionMap();
        aMap.put("back", new BackAction());
        aMap.put("backAll", new BackToStartAction());
        aMap.put("hunt", new HuntAction());
        aMap.put("debug", new DebugAction(this));

    }

    @Override
    public void addNotify() {
        super.addNotify();
        if (getView() == null) {
            setView(new WelcomeView(this));
        }
        resizer = new Resizer();
        mainSpace.addPropertyChangeListener(resizer);
    }

    @Override
    public void removeNotify() {
        mainSpace.removePropertyChangeListener(resizer);
        super.removeNotify();
    }

    private View getView() {
        final View view = views.peek();
        return (view instanceof RootView) ? null : view;
    }

    private static Map<String, FileViewSettings> createSettings() {
        final TokenizerFactory fac = TokenizerFactory.getSharedInstance();
        final Set<String> tokenizerNames = fac.getRegisteredTokenizerNames();
        final Map<String, FileViewSettings> map = new HashMap<>();
        for (String name : tokenizerNames) {
            map.put(name, new FileViewSettings(name));
        }
        return map;
    }

    /**
     * This function is used internally to perform the setting of the various
     * properties of this class from a properties file.
     *
     * @param props The Properties object.
     */
    private void configure(Map<String, String> props) {
        firePropertyChangeEvents = false;
        try {
            for (Map.Entry entry : settings.entrySet()) {
                final FileViewSettings fvs = (FileViewSettings) entry.getValue();
                fvs.load(props);
            }

            String s;

            s = props.get("orion.rows");
            rows = 40;
            try {
                rows = Integer.parseInt(s);
            } catch (Exception e) {
                // Ignore
            }
            s = props.get("orion.columns");
            columns = 80;
            try {
                columns = Integer.parseInt(s);
            } catch (Exception e) {
                // Ignore
            }

            String fontString = props.get("orion.font");
            if (fontString == null) {
                fontString = "Monospaced--12";
            }
            setFont(Font.decode(fontString));
        } finally {
            firePropertyChangeEvents = true;
        }
    }

    /**
     * Set the properties of this object from a Properties object.
     *
     * @param displayProps The properties object.
     */
    public void setDisplayProperties(Map<String, String> displayProps) {
        configure(displayProps);
        final View v = getView();
        v.deactivate();
        v.activate();
    }

    @Override
    public void setFont(Font newFont) {
        super.setFont(newFont);
        if (mainSpace != null) {
            mainSpace.setFont(newFont);
        }
    }

    @Override
    protected void firePropertyChange(String propertyName,
            Object oldValue,
            Object newValue) {
        if (firePropertyChangeEvents) {
            super.firePropertyChange(propertyName, oldValue, newValue);
        }
    }

    /**
     * Push the supplied view to the top, causing it to be displayed.
     *
     * @param v The new view.
     * @throws NullPointerException if the new view is null.
     */
    public void setView(View v) {
        final View oldView = views.peek();
        oldView.deactivate();
        views.push(v);

        //
        // Note that we must set the input/action maps BEFORE we activate.
        // Otherwise WelcomeView does not work properly.
        //
        final InputMap im = v.getInputMap();
        if (im != null) {
            getInputMap().setParent(im);
        }
        final ActionMap am = v.getActionMap();
        if (am != null) {
            getActionMap().setParent(am);
        }
        v.activate();
    }

    /**
     * Pop a view off the view stack.
     */
    public void popView() {
        final View oldView = views.pop();
        oldView.deactivate();
        final View newView = views.peek();
        //
        // Note that we must set the input/action maps BEFORE we
        // activate.
        // Otherwise WelcomeView does not work properly.
        //
        final InputMap im = newView.getInputMap();
        getInputMap().setParent(im);
        final ActionMap am = newView.getActionMap();
        if (am != null) {
            getActionMap().setParent(am);
        }
        newView.activate();
        firePropertyChange("view", oldView, getView());
    }

    /**
     * Pop a view off the view stack.
     */
    private void popAllViews() {
        final View oldView = views.pop();

        View newView = oldView;
        while (getView() != null) {
            newView = views.pop();
        }
        views.push(newView);
        if (oldView == newView) {
            return;
        }
        oldView.deactivate();

        //
        // Note that we must set the input/action maps BEFORE we
        // activate.
        // Otherwise WelcomeView does not work properly.
        //
        final InputMap im = newView.getInputMap();
        getInputMap().setParent(im);
        final ActionMap am = newView.getActionMap();
        if (am != null) {
            getActionMap().setParent(am);
        }
        newView.activate();
        firePropertyChange("view", oldView, getView());
    }

    public WritableCursesModel getHeaderArea() {
        return titlesModel;
    }

    public WritableCursesModel getMainArea() {
        return mainModel;
    }

    public Messenger getMessenger() {
        return mess;
    }

    private void setStatusAreaController(Activatable sac) {
        if (currentStatusAreaController != null) {
            currentStatusAreaController.deactivate();
        }
        currentStatusAreaController = sac;
        currentStatusAreaController.activate();
        //	requestFocusInWindow();
    }

    public void startInput(String prefix,
            String initialValue,
            OrionInputHandler handler) {
        cif.setLabel(prefix);
        cif.setDefaultValue(initialValue);
        ifl.setHandler(handler);
        setStatusAreaController(cif);
    }

    /**
     * Perform a hunt operation on the supplied string.
     *
     * @param searchString The string to hunt for
     */
    public void searchFor(String searchString) {
        final HuntSwingWorker hsw = new HuntSwingWorker(this, searchString);
        hsw.execute();
    }

    /**
     * Get the name of the currently viewed file.
     * <p>
     * If a file is not currently being viewed, then <code>null</code> is
     * returned.
     *
     * @return The name.
     */
    public String getFileName() {
        final View v = getView();
        if (v instanceof FileView) {
            return ((FileView) v).getFileName();
        }
        return null;
    }

    /**
     * Get the location within the currently viewed file. This method only has
     * meaning if the current view is a file. If this is not the case, then this
     * method returns <code>-1</code>.
     *
     * @return The location (if possible) or -1 if inappropriate.
     */
    public int getFileLocation() {
        final View v = getView();
        if (v instanceof FileView) {
            return ((FileView) v).getFileLocation();
        }
        return -1;
    }

    public Set<FileViewSettings> getSettingsSet() {
        final Set<FileViewSettings> ret = new TreeSet<>();
        for (Map.Entry<String, FileViewSettings> entry : settings.entrySet()) {
            ret.add(entry.getValue());
        }
        return ret;
    }

    public FileViewSettings getSettings(String language) {
        return settings.get(language);
    }

    public int getRows() {
        return rows;
    }

    public void setRows(int rows) {
        this.rows = rows;
    }

    public int getColumns() {
        return columns;
    }

    public void setColumns(int columns) {
        this.columns = columns;
    }

    public OrionDataModel getModel() {
        return model;
    }

    public void setModel(OrionDataModel model) {
        final OrionDataModel oldModel = this.model;
        this.model = model;
        firePropertyChange("model", oldModel, model);
    }

    public Pattern getLastSearch() {
        return lastSearch;
    }

    public void setLastSearch(Pattern pattern) {
        lastSearch = pattern;
    }

    public int getMainAreaRow(Point point) {
        final int row = mainSpace.getRow(point) - titlesModel.getRows();
        if ((row < 0) || (row >= mainSpace.getModel().getRows())) {
            return -1;
        }
        return row;
    }

    public int getMainAreaColumn(Point point) {
        final int col = mainSpace.getColumn(point);
        if ((col < 0) || (col >= mainSpace.getModel().getCols())) {
            return -1;
        }
        return col;
    }

    // PACKAGE privacy
    ResourceBundle getBundle() {
        return ResourceBundle.getBundle(
                "uk.co.ruelgnoj.orion.component.resources",
                getLocale());
    }

    public void setLocked(boolean b) {
        lui.setLocked(b);
    }

    /**
     * Class to handle resize events.
     */
    private class Resizer implements PropertyChangeListener {

        /**
         * Constructor.
         */
        public Resizer() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void propertyChange(PropertyChangeEvent ev) {
            final View v = getView();
            if (v != null) {
                v.deactivate();
                v.activate();
            }
        }
    }

    /**
     *
     * Handle Orion Mouse events.
     */
    private class OrionMouseHandler extends MouseAdapter {

        /**
         * Constructor.
         */
        public OrionMouseHandler() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void mousePressed(MouseEvent e) {
            requestFocus();
        }
    }

    /**
     * Action to pop a view from the stack.
     */
    private class BackAction extends AbstractAction {

        /**
         * Constructor.
         */
        public BackAction() {
            super("Back");
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void actionPerformed(ActionEvent e) {
            if (isEnabled()) {
                popView();
            }
        }
    }

    /**
     * Action to pop a view from the stack.
     */
    private class BackToStartAction extends AbstractAction {

        /**
         * Constructor.
         */
        public BackToStartAction() {
            super("BacktoStart");
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void actionPerformed(ActionEvent e) {
            if (isEnabled()) {
                popAllViews();
            }
        }
    }

    private class OrionInputFieldListener implements CursesInputFieldListener {

        private OrionInputHandler handler;

        public OrionInputFieldListener() {
            handler = null;
        }

        public void setHandler(OrionInputHandler handler) {
            this.handler = handler;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void inputFieldCancelled(CursesInputField f) {
            mess.info("");
            setStatusAreaController(mess);

            handler = null;

            // A bit of a hack, but when someone cancels the hunt from a welcome
            // view then we want to pop the view stack. For the Orion application
            // this will quit, and for the orion applet, this will cause the view
            // to be refreshed, and thus the worst that will happen is that the
            // user will see a flicker.
            if (views.peek() instanceof WelcomeView) {
                popView();
            }
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void inputFieldSet(CursesInputField f) {
            mess.info("");
            setStatusAreaController(mess);

            final OrionInputHandler oldHandler = handler;

            handler = null;

            oldHandler.handleInput(f.getValue());
        }
    }

    /**
     * Action to initiate a hunt operation. This does not perform the hunt, but
     * merely transforms the component to request the user to type a hunt
     * string.
     */
    private class HuntAction extends AbstractAction implements OrionInputHandler {

        /**
         * Constructor.
         */
        public HuntAction() {
            super("Hunt");
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void actionPerformed(ActionEvent e) {
            startInput(getBundle().getString("hunt"), "", this);
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void handleInput(String input) {
            if (input.length() == 0) {
                mess.error("No string specified.");
                return;
            }
            searchFor(input);
        }
    }

    /**
     * The root view that does nothing, but is just the base of the stack.
     */
    private static class RootView implements View {

        /**
         * Constructor.
         */
        public RootView() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void activate() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void deactivate() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public ActionMap getActionMap() {
            return null;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public InputMap getInputMap() {
            return null;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean isActive() {
            return false;
        }
    }

    private abstract static class LocakbleOverlay<T extends JComponent> extends LayerUI<T> {

        private boolean locked;

        public LocakbleOverlay() {
        }

        public final boolean isLocked() {
            return locked;
        }

        public final void setLocked(boolean locked) {
            this.locked = locked;
        }
    }

    private static class MyOverlay extends LocakbleOverlay<Curses> {

        private final Paint thePaint;

        public MyOverlay() {
            this(new Color(0x7f, 0x7f, 0x7f, 0x7f));
        }

        public MyOverlay(Paint thePaint) {
            this.thePaint = thePaint;
        }

        @Override
        public void paint(Graphics g, JComponent c) {
            super.paint(g, c);
            if (isLocked()) {
                final Graphics2D g2 = (Graphics2D) g.create();
                final int w = c.getWidth();
                final int h = c.getHeight();
                g2.setPaint(thePaint);
                g2.fillRect(0, 0, w, h);

                g2.dispose();
            }
        }
    }
}
