/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: Utils.java 256 2008-07-15 15:54:02Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package uk.co.ruelgnoj.orion.component;

/**
 * A set of static utility functions for Orion component implementation.
 *
 * @author Steve Millington
 * @version $Id: Utils.java 256 2008-07-15 15:54:02Z steve $
 */
public final class Utils {
    private static final String ONE_DOT = ".";
    private static final String THREE_DOT = "...";
    private static final String FIVE_DOT = ".....";
    private static final int LENGTH_FOR_ONLY_DOTS = FIVE_DOT.length();
    /**
     * Prevent construction.
     */
    private Utils() {
    }

    /**
     * Truncate the name of a file to at most the supplied length.
     * <p>If the count is &lt; 5 characters, and the name is longer than that,
     * then a correctly lengthed string of dots is returned.
     * <p>Otherwise, the leading path components are compressed into a '...'
     * clause.
     * <p>For example:
     * <code>
     * <pre>
     * truncateFile("/abcdefg/hi",8)
     * </pre>
     * </code>
     * returns
     * <code>
     * <pre>
     * "/.../hi"
     * </pre>
     * </code>
     * @param fileName The name of the file to truncate
     * @param count the truncate amount
     * @return The (possibly truncated) string.
     * @throws java.lang.NullPointerException if <code>fileName</code> is null.
     * @throws java.lang.IllegalArgumentException if <code>count</code> is
     * negative.
     */
    public static String truncateFile(String fileName, int count) {
	if (fileName == null) {
	    throw new NullPointerException("fileName");
	}
	if (count < 0) {
	    throw new IllegalArgumentException("count<0");
	}
	if (fileName.length() <= count) {
	    return fileName;
	}
	if (count <= LENGTH_FOR_ONLY_DOTS) {
	    return FIVE_DOT.substring(LENGTH_FOR_ONLY_DOTS - count);
	}

	final StringBuilder ret = new StringBuilder();
	if (fileName.startsWith("/")) {
	    ret.append('/');
	    fileName = fileName.substring(1);
	}
	boolean dots = false;
	while (ret.length() + fileName.length() > count) {
	    final int off = fileName.indexOf('/');
	    if (off < 0) {
		ret.append(THREE_DOT);
		final int n = count - ret.length();
		fileName = fileName.substring(fileName.length() - n);
	    } else {
		if (!dots) {
		    ret.append(THREE_DOT);
		    ret.append('/');
		    dots = true;
		}
		fileName = fileName.substring(off + 1);
	    }
	}
	ret.append(fileName);
	return ret.toString();
    }
}
