/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: WelcomeView.java 256 2008-07-15 15:54:02Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.component;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.util.Random;
import java.util.ResourceBundle;
import javax.swing.Action;
import javax.swing.InputMap;
import javax.swing.KeyStroke;
import javax.swing.Timer;
import uk.co.ruelgnoj.orion.product.Product;
import uk.co.ruelgnoj.orion.cc.WritableCursesModel;

class WelcomeView extends AbstractView {

    private static final int BLINK_PERIOD = 2000;
    private static final int BLINK_DURATION = 200;
    private static final Color[] COLORS = {Color.RED, Color.GREEN, Color.YELLOW, new Color(150, 150, 255), Color.WHITE, Color.PINK, Color.CYAN, Color.YELLOW.darker()};

    //
    // If these pictures change, care must be taken to ensure that the animation
    // gets updated with it.
    //
    private static final String[] LOGO = {"               ^                      ", "               \\\\                     ", "      \\ /       \\\\                    ", "     --v-        \\\\                   ", "    (    )        ||                  ", "   (  o o )       ||                  ", "    |    >        ||                  ", "    |   -|        ||                  ", "     \\  /         ||                \\ ", "     (  )         ||      ===========>", "    _/  \\________^||                / ", "  _/  />|________-||                  ", " / ___^ |         ||                  ", "/_/ |   |         ||                  ", "V   |   |         ||                  ", "    |___|        //                   ", "    * * *       //                    ", "    |---|      //                     ", "   / ---|      V                      ", "   |/  \\\\                             ", "   ||  | \\                            ", "   ||   \\|                            ", "   |)   ||                            ", "   ||   ||                            ", "   ||_  ||_                           ", "   \\__> \\__>                          "};

    private static final String[] COLORMAP = {"               1                      ", "               11                     ", "      2 2       11                    ", "     2222        11                   ", "    2    2        11                  ", "   2  0 0 2       11                  ", "    2    2        11                  ", "    2   22        11                  ", "     2  2         11                3 ", "     2  2         11      333333333333", "    22  222222222211                3 ", "  22  22222222222211                  ", " 2 2222 2         11                  ", "222 2   2         11                  ", "2   2   2         11                  ", "    22222        11                   ", "    42424       11                    ", "    22222      11                     ", "   222222      1                      ", "   22  22                             ", "   22  2 2                            ", "   22   22                            ", "   22   22                            ", "   22   22                            ", "   222  222                           ", "   2222 2222                          "};

    //----------------------------------------------------------------------
    // EYES
    //----------------------------------------------------------------------
    private static final String OPEN = "o o";
    private static final String CLOSED = "- -";
    private static final String WINK_RIGHT = "o -";
    private static final String WINK_LEFT = "- o";
    private static final String STARS = "* *";
    //----------------------------------------------------------------------
    private final Timer blinkOpenTimer;
    private final Timer blinkCloseTimer;
    private Random rnd;

    public WelcomeView(OrionComponent orionComponent) {
        super(orionComponent);

        rnd = new Random();

        //
        // The timer for blink Opening
        //
        blinkOpenTimer = new Timer(BLINK_PERIOD, (e) -> setEyes(OPEN));
        // Ensure the 'open'
        blinkOpenTimer.setInitialDelay(BLINK_PERIOD + BLINK_DURATION);

        //
        // The timer for blink Closing
        //
        blinkCloseTimer = new Timer(BLINK_PERIOD, (e) -> {
            //
            // Work out what sort of blink to do
            //
            final int n = rnd.nextInt();
            String eyes;
            if (n % 16 == 0) {
                eyes = STARS;
            } else if (n % 8 == 0) {
                eyes = WINK_LEFT;
            } else if (n % 4 == 0) {
                eyes = WINK_RIGHT;
            } else {
                eyes = CLOSED;
            }
            setEyes(eyes);
        });
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void initializeInputMap(InputMap im) {
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0), "hunt");
    }

    private void display() {
        final WritableCursesModel mainArea = orionComponent.getMainArea();
        final Color bgColor = Color.BLACK;
        final Color fgColor = Color.WHITE;
        mainArea.setBatch(true);

        mainArea.setBG(bgColor);
        mainArea.setFG(fgColor);
        mainArea.clear();
        if (isPictureShowing()) {
            final int rows = mainArea.getRows();
            final int cols = mainArea.getCols();

            final int r = (rows - LOGO.length - 1) / 2;
            final int c = (cols - LOGO[0].length() - 1) / 2;
            for (int i = 0; i < LOGO.length; i++) {
                mainArea.moveTo(c, r + i);
                for (int j = 0; j < LOGO[i].length(); j++) {
                    final char ch1 = COLORMAP[i].charAt(j);
                    if (ch1 != ' ') {
                        mainArea.setFG(COLORS[ch1 - '0']);
                    }
                    mainArea.print(LOGO[i].charAt(j));
                }
            }
        }

        mainArea.setBatch(false);
    }

    //
    // Draw the Orion name and version information
    //
    private void setTitles() {
        final ResourceBundle rb = Product.getBundle(orionComponent.getLocale());
        final WritableCursesModel headerArea = orionComponent.getHeaderArea();
        headerArea.setBatch(true);
        headerArea.setBG(Color.BLACK);
        headerArea.clear();

        final int cols = headerArea.getCols();
        final String desc = rb.getString("orion.productname");
        final String rest = " " + rb.getString("orion.version");
        final int len = desc.length() + rest.length();

        headerArea.setUnderLine(true);
        headerArea.moveTo((cols - len) / 2, 1);
        headerArea.setFG(Color.YELLOW);
        headerArea.print(desc);
        headerArea.setFG(Color.CYAN);
        headerArea.print(rest);
        headerArea.setUnderLine(false);
        headerArea.setBatch(false);
    }

    @Override
    public void activate() {
        super.activate();

        setTitles();

        display();

        if (isPictureShowing()) {
            blinkOpenTimer.restart();
            blinkCloseTimer.restart();
        }

        //
        // Now do the 'hunt' action so that we get the initial search up
        //
        final Action huntAction = orionComponent.getActionMap().get("hunt");
        if (huntAction != null) {
            if (huntAction.isEnabled()) {
                final ActionEvent e = new ActionEvent(orionComponent, ActionEvent.ACTION_PERFORMED, "hunt");
                huntAction.actionPerformed(e);
            }
        }
    }

    @Override
    public void deactivate() {
        if (isPictureShowing()) {
            blinkOpenTimer.stop();
            blinkCloseTimer.stop();
        }
        super.deactivate();
    }

    /**
     * Set the eyes.
     *
     * @param eyes A string containing the eyes. This must be 3 characters in
     * length.
     */
    private void setEyes(String eyes) {
        final WritableCursesModel mainArea = orionComponent.getMainArea();
        final int rows = mainArea.getRows();
        final int cols = mainArea.getCols();
        final int r = (rows - LOGO.length - 1) / 2;
        final int c = (cols - LOGO[0].length() - 1) / 2;
        synchronized (mainArea) {
            mainArea.setBatch(true);
            mainArea.moveTo(c + 6, r + 5);
            mainArea.setFG(COLORS[0]);
            mainArea.print(eyes);
            mainArea.setBatch(false);
        }
    }

    /**
     * Determine if the picture is to be displayed.
     * <p>
     * A picture is only displayed if there is enough space in the component for
     * it.
     *
     * @return if the picture is to be displayed.
     */
    private boolean isPictureShowing() {
        final WritableCursesModel mainArea = orionComponent.getMainArea();
        final int rows = mainArea.getRows();
        final int cols = mainArea.getCols();

        final int r = (rows - LOGO.length - 1) / 2;
        final int c = (cols - LOGO[0].length() - 1) / 2;
        return (r >= 0) && (c >= 0);
    }
}
