/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: BraceMatchAction.java 285 2011-07-19 12:37:42Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.component.fileview;

import java.awt.event.ActionEvent;

import java.util.List;

import javax.swing.AbstractAction;

import uk.co.ruelgnoj.orion.tokenizers.LanguageToken;
import uk.co.ruelgnoj.orion.tokenizers.TokenTypes;

/**
 * Find the matching brace for the current position.
 *
 * @author Steve Millington
 * @version $Id: BraceMatchAction.java 285 2011-07-19 12:37:42Z steve $
 */
public class BraceMatchAction extends AbstractAction {

    private final Internals internals;

    /**
     * Constructor.
     *
     * @param internals The file view
     */
    public BraceMatchAction(Internals internals) {
        this.internals = internals;
    }

    @Override
    public void actionPerformed(ActionEvent e) {
        final TokenizedFile tf = internals.getTokenizedFile();
        final char incChar = tf.charAt(internals.getCursorOffset());

        final LanguageToken tok = internals.getCurrentToken();
        if (tok.getType() != TokenTypes.TT_PUNC) {
            internals.error("putonbracket");
            return;
        }

        final int oldCursorLine = internals.getCursorLine();
        final int oldCursorCol = internals.getCursorColumn();
        final boolean down;
        final char decChar;
        switch (incChar) {
            case '{':
                decChar = '}';
                down = true;
                break;
            case '[':
                decChar = ']';
                down = true;
                break;
            case '(':
                decChar = ')';
                down = true;
                break;

            case '}':
                decChar = '{';
                down = false;
                break;
            case ']':
                decChar = '[';
                down = false;
                break;
            case ')':
                decChar = '(';
                down = false;
                break;
            default:
                internals.error("putonbracket");
                return;
        }

        internals.clearMessages();

        final boolean found;
        if (down) {
            found = moveForwardMatching(incChar, decChar);
        } else {
            found = moveBackwardMatching(incChar, decChar);
        }
        if (found) {
            internals.setCursorMagic(internals.getCursorColumn());

            internals.adjustViewportAndDisplay(oldCursorLine, oldCursorCol);
        }
    }

    /**
     * Move forwards to the matching brace.
     *
     * @param incCh The opening brace character
     * @param decCh The closing brace character
     * @return true if a match was found.
     */
    private boolean moveForwardMatching(char incCh, char decCh) {
        final TokenizedFile tf = internals.getTokenizedFile();
        int ind = internals.getCurrentTokenIndex();
        int dOff = internals.getCursorOffset() + 1 - internals.getCurrentToken().getStartOffset();
        int nesting = 1;

        final int startRow = internals.getCursorLine();

        for (int row = startRow; row < tf.getLineCount(); row++) {
            final List<LanguageToken> rowVec = tf.getLine(row);
            for (; ind < rowVec.size(); ind++) {
                final LanguageToken tok = rowVec.get(ind);
                if (tok.getType() == TokenTypes.TT_PUNC) {
                    int off = tok.getStartOffset() + dOff;
                    dOff = 0;
                    for (; off <= tok.getEndOffset(); off++) {
                        final char thisChar = tf.charAt(off);
                        if (thisChar == incCh) {
                            nesting++;
                        } else if (thisChar == decCh) {
                            nesting--;
                            if (nesting == 0) {
                                moveTo(row, off);
                                return true;
                            }
                        }
                    }
                }
            }
            ind = 0;
        }
        return false;
    }

    /**
     * Move the 'internals' to the supplied place.
     *
     * @param newCursorLine The new cursor line
     * @param off The new cursor offset
     */
    private void moveTo(int newCursorLine, int off) {
        final TokenizedFile tf = internals.getTokenizedFile();
        internals.setCursorLine(newCursorLine);
        internals.setCursorOffset(off);

        final List<LanguageToken> tlv = tf.getLine(internals.getCursorLine());
        int so = tlv.get(0).getStartOffset();
        int newCol = 0;
        while (so < internals.getCursorOffset()) {
            if (tf.charAt(so) == '\t') {
                newCol = ((newCol / internals.getTabstop()) + 1) * internals.getTabstop();
            } else {
                newCol++;
            }
            so++;
        }
        internals.setCursorColumn(newCol);
    }

    /**
     * Move backwards to the matching brace.
     *
     * @param incCh The closing brace character
     * @param decCh The opening brace character
     * @return true if a match was found.
     */
    private boolean moveBackwardMatching(char incCh, char decCh) {
        final TokenizedFile tf = internals.getTokenizedFile();
        int off = internals.getCursorOffset() - 1;
        int ind = internals.getCurrentTokenIndex();
        int dOff = off - internals.getCurrentToken().getStartOffset();
        int nesting = 1;

        for (int row = internals.getCursorLine(); row >= 0; row--) {
            final List<LanguageToken> rowVec = tf.getLine(row);
            if (ind == -1) {
                ind = rowVec.size() - 1;
            }
            for (; ind >= 0; ind--) {
                final LanguageToken tok = rowVec.get(ind);
                if (tok.getType() == TokenTypes.TT_PUNC) {
                    if (dOff >= -1) {
                        off = tok.getStartOffset() + dOff;
                        dOff = -2; // -2 is used cos of starting on a
                        // single char thing
                    } else {
                        off = tok.getEndOffset();
                    }
                    for (; off >= tok.getStartOffset(); off--) {
                        final char thisChar = tf.charAt(off);
                        if (thisChar == incCh) {
                            nesting++;
                        } else if (thisChar == decCh) {
                            nesting--;
                            if (nesting == 0) {
                                moveTo(row, off);
                                return true;
                            }
                        }
                    }
                }
            }
            ind = -1;
        }
        return false;
    }
}
