/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: FileView.java 285 2011-07-19 12:37:42Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.component.fileview;

import java.awt.Color;

import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;

import java.io.IOException;

import java.text.MessageFormat;

import java.util.List;
import java.util.ResourceBundle;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.AbstractAction;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.KeyStroke;

import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyleContext;

import uk.co.ruelgnoj.orion.cc.WritableCursesModel;

import uk.co.ruelgnoj.orion.component.AbstractView;
import uk.co.ruelgnoj.orion.component.ActionMapHelper;
import uk.co.ruelgnoj.orion.component.FileViewSettings;
import uk.co.ruelgnoj.orion.component.HuntSpecificSwingWorker;
import uk.co.ruelgnoj.orion.component.HuntSwingWorker;
import uk.co.ruelgnoj.orion.component.Messenger;
import uk.co.ruelgnoj.orion.component.OpenFileSwingWorker;
import uk.co.ruelgnoj.orion.component.OrionComponent;
import uk.co.ruelgnoj.orion.component.OrionInputHandler;
import uk.co.ruelgnoj.orion.component.Utils;

import uk.co.ruelgnoj.orion.tokenizers.LanguageToken;
import uk.co.ruelgnoj.orion.tokenizers.LanguageTokenizer;
import uk.co.ruelgnoj.orion.tokenizers.TokenType;
import uk.co.ruelgnoj.orion.tokenizers.TokenTypes;
import uk.co.ruelgnoj.orion.tokenizers.TokenizerFactory;

/**
 * View some sort of text file.
 *
 * @author Steve Millington
 * @version $Id: FileView.java 285 2011-07-19 12:37:42Z steve $
 */
public final class FileView extends AbstractView {

    private static final String PREFIX = "msg.fileview.";
    // Keep same as HTMLSourceCodeFormatter.java
    private static final String LINE_NUMBER_PREFIX = "L";

    /* ******* DEBUG VERSIONS
     private static final char DISPLAYABLE_NL		= '>';
     private static final char DISPLAYABLE_CR		= '>';
     private static final char DISPLAYABLE_TABFILL	= '.';
     private static final char DISPLAYABLE_TAB		= '+';
     ***********************/
    private static final char DISPLAYABLE_NL = ' ';
    private static final char DISPLAYABLE_CR = ' ';
    private static final char DISPLAYABLE_TABFILL = ' ';
    private static final char DISPLAYABLE_TAB = ' ';
    private static final TokenType TT_INDEXED = new TokenTypeIndexed();
    private Internals internals;
    private final MouseListener ml;
    private final MouseWheelListener mwl;
    private final ActionMapHelper amh;
    private final String fileName;
    private final char[] searchArray;
    private int vpTopLine;
    private int vpLeftColumn;
    // Made package privacy to all the actions to access it.
    private TokenizedFile tf;
    private int cursorOffset;
    private int cursorLine;
    private int cursorColumn;
    private int cursorMagic;
    private boolean displayEnabled;
    private FileViewSettings settings;

    /**
     * Constructor.
     *
     * @param orionComponent The host component
     * @param fileName The name of the file to view
     * @param searchString The string to initially hunt for.
     * @throws IOException .
     */
    public FileView(OrionComponent orionComponent,
            String fileName,
            String searchString,
            CharSequence content) {
        super(orionComponent);
        displayEnabled = true;
        this.fileName = fileName;
        this.searchArray = searchString.toCharArray();

        this.vpTopLine = 0;
        this.vpLeftColumn = 0;
        this.cursorOffset = 0;
        this.cursorLine = 0;
        this.cursorColumn = 0;
        this.cursorMagic = 0;

        amh = new ActionMapHelper(orionComponent);
        ml = new MouseAdapter() {

            @Override
            public void mousePressed(MouseEvent e) {
                FileView.this.info(null);
                goToRowCol(
                        FileView.this.orionComponent.getMainAreaRow(
                                e.getPoint()),
                        FileView.this.orionComponent.getMainAreaColumn(
                                e.getPoint()));
            }

            @Override
            public void mouseClicked(MouseEvent e) {
                if (e.getClickCount() % 2 == 0) {
                    amh.invokeAction("select");
                }
            }
        };

        mwl = new MouseWheelListener() {

            @Override
            public void mouseWheelMoved(MouseWheelEvent e) {
                String direction;
                String amount;
                int notches = e.getWheelRotation();
                if (notches < 0) {
                    direction = "Up";
                    notches = -notches;
                } else {
                    direction = "Down";
                }
                if (e.getScrollType() == MouseWheelEvent.WHEEL_UNIT_SCROLL) {
                    amount = "viewport";
                } else { // Scroll Type == MouseWheelEvent.WHEEL_BLOCK_SCROLL) {
                    amount = "page";
                }
                final String function = amount + direction;
                enableDisplay(false);
                for (int i = 0; i < notches; i++) {
                    if (i == notches - 1) {
                        enableDisplay(true);
                    }
                    amh.invokeAction(function);
                }
            }
        };

        // Note: We do NOT display() here cos this will be done when we get
        //       activated.
        tokenize(content);	// This parses the data, only first time through
    }

    Messenger getMessenger() {
        return orionComponent.getMessenger();
    }

    private void enableDisplay(boolean b) {
        displayEnabled = b;
    }

//    /**
//     * Create a view. This can be better than using the constructor since if
//     * there is any exception, and <code>ExceptionView</code> is return"ed
//     * instead.
//     * @param orionComponent The parent Orion component.
//     * @param fileName The name of the file.
//     * @param searchString The string to search for.
//     * @return Either a <code>FileView</code> or an <code>ExceptionView</code>.
//     * @see uk.co.ruelgnoj.orion.component.ExceptionView
//     */
//    public static View createView(OrionComponent orionComponent,
//            String fileName,
//            String searchString) {
//        try {
//            return new FileView(orionComponent, fileName, searchString);
//        } catch (OrionDataModelException e) {
//            return new ExceptionView(orionComponent, e);
//        }
//    }
    @Override
    protected void initializeInputMap(InputMap im) {
        im.put(KeyStroke.getKeyStroke('H'),
                "hunt");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_DOWN, 0, false),
                "pageDown");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_UP, 0, false),
                "pageUp");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_F, KeyEvent.CTRL_MASK, false),
                "pageDown");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_B, KeyEvent.CTRL_MASK, false),
                "pageUp");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0, false),
                "select");
        im.put(KeyStroke.getKeyStroke(
                KeyEvent.VK_HOME, KeyEvent.CTRL_MASK, false),
                "top");
        im.put(KeyStroke.getKeyStroke(
                KeyEvent.VK_END, KeyEvent.CTRL_MASK, false),
                "bottom");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_HOME, 0, false),
                "startOfLine");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_END, 0, false),
                "endOfLine");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, 0, false),
                "characterLeft");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, 0, false),
                "characterRight");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_UP, 0, false),
                "lineUp");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0, false),
                "lineDown");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_J, 0, false),
                "lineDown");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_K, 0, false),
                "lineUp");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_H, 0, false),
                "characterLeft");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_L, 0, false),
                "characterRight");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_N, 0, false),
                "nextToken");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_W, 0, false),
                "nextWord");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_B, 0, false),
                "previousWord");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_P, 0, false),
                "previousToken");
        im.put(KeyStroke.getKeyStroke(
                KeyEvent.VK_G, KeyEvent.SHIFT_MASK, false),
                "bottom");
        im.put(KeyStroke.getKeyStroke('*'),
                "select");
        im.put(KeyStroke.getKeyStroke('%'),
                "braceMatch");
        im.put(KeyStroke.getKeyStroke('$'),
                "endOfLine");
        im.put(KeyStroke.getKeyStroke('0'),
                "startOfLine");
        im.put(KeyStroke.getKeyStroke('/'),
                "findForward");
        im.put(KeyStroke.getKeyStroke('?'),
                "findBackward");
        im.put(KeyStroke.getKeyStroke('f'),
                "openFile");
        im.put(KeyStroke.getKeyStroke(':'),
                "gotoLine");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_N, KeyEvent.CTRL_MASK, false),
                "nextFind");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_P, KeyEvent.CTRL_MASK, false),
                "previousFind");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, KeyEvent.CTRL_MASK, false),
                "javaNativeHunt");
        im.put(
                KeyStroke.getKeyStroke(
                        KeyEvent.VK_S, KeyEvent.CTRL_MASK, false),
                "showInBrowser");
    }

    @Override
    protected void initializeActionMap(ActionMap am) {
        // Hack!
        this.internals = new InternalsImpl();

        am.put("bottom", new BottomAction());
        am.put("braceMatch", new BraceMatchAction(internals));
        am.put("characterLeft", new CharacterLeftAction(internals));
        am.put("characterRight", new CharacterRightAction(internals));
        am.put("endOfLine", new EndOfLineAction(internals));
        am.put("findBackward", new FindAction(true));
        am.put("findForward", new FindAction(false));
        am.put("gotoLine", new GotoLineAction());
        am.put("javaNativeHunt", new JavaNativeHuntAction());
        am.put("lineDown", new LineDownAction(internals));
        am.put("lineUp", new LineUpAction(internals));
        am.put("nextFind", new NextFindAction(false));
        am.put("nextToken", new NextTokenAction(false));
        am.put("nextWord", new NextWordAction(internals, false));
        am.put("openFile", new OpenFileAction());
        am.put("pageDown", new PageDownAction());
        am.put("pageUp", new PageUpAction());
        am.put("previousFind", new NextFindAction(true));
        am.put("previousToken", new NextTokenAction(true));
        am.put("previousWord", new NextWordAction(internals, true));
        am.put("select", new SelectAction());
        am.put("startOfLine", new StartOfLineAction(internals));
        am.put("top", new TopAction());
        am.put("viewportDown", new ViewportDownAction());
        am.put("viewportUp", new ViewportUpAction());
    }

    private static class TokenTypeIndexed implements TokenType {

        public TokenTypeIndexed() {
        }

        @Override
        public String getName() {
            return "indexed";
        }

        @Override
        public String getDisplayName() {
            return "indexed";
        }

        @Override
        public boolean isIndexable() {
            return true;
        }
    }

    //
    // Tokenize the dataBuffer, to produce a set of lines of tokens.
    //
    private void tokenize(CharSequence content) {
        final LanguageTokenizer lt
                = TokenizerFactory.getSharedInstance().getTokenizer(fileName);
        tf = new TokenizedFile(lt, content);
        settings = orionComponent.getSettings(lt.getLanguage());

        final Navigator nav = tf.navigator();
        while (nav.hasNext()) {
            final LanguageToken tok = nav.next();
            if (tok.getType().isIndexable()) {
                if (tf.tokenContains(tok, searchArray)) {
                    tok.setType(TT_INDEXED);
                }
            }
        }

        if (searchArray.length > 0) {
            goToToken(false);
        }
    }

    private String format(String msgID, Object[] inserts) {
        final String tmp = getBundle().getString(msgID);
        return (new MessageFormat(tmp, orionComponent.getLocale())).format(
                inserts, new StringBuffer(), null).toString();
    }

    //
    // Display the positional information - currently in the top window
    //
    private void displayPos() {
        final WritableCursesModel header = orionComponent.getHeaderArea();
        header.setBatch(true);
        header.setFG(Color.lightGray);
        final Object[] args = {
            cursorLine + 1,
            tf.getLineCount(),
            cursorColumn + 1
        };
        final String s = format(PREFIX + "pos", args);

        //
        // Blank out any previous message
        //
        header.moveTo(1, 1);
        final int nSpaces = header.getCols() - s.length() - 1;
        for (int i = 0; i < nSpaces; i++) {
            header.print(' ');
        }

        //
        // Now put the message out
        //
        header.print(s);

        header.setBatch(false);
    }

    private void displayAllLines() {
        if (!displayEnabled) {
            return;
        }
        final StyleContext styleContext = settings.getStyleContext();
        final WritableCursesModel mainArea = orionComponent.getMainArea();

        int screenLineIndex = 0;
        int off = 0;
        mainArea.moveTo(0, 0);

        final Navigator nav = tf.navigator();
        nav.gotoLine(vpTopLine);
        LanguageToken tok;
        while (nav.hasNext() && (screenLineIndex < mainArea.getRows())) {
            tok = nav.next();
            if (tok.getType() == TokenTypes.TT_NEWLINE) {
                off = 0;
                screenLineIndex++;
                if (screenLineIndex < mainArea.getRows()) {
                    mainArea.moveTo(0, screenLineIndex);
                }
            } else if (off < vpLeftColumn + mainArea.getCols()) {
                final String styleName = tok.getType().getName();
                final Style style = styleContext.getStyle(styleName);
                mainArea.setFG(StyleConstants.getForeground(style));
                mainArea.setBG(StyleConstants.getBackground(style));
                mainArea.setUnderLine(StyleConstants.isUnderline(style));
                for (int k = 0; k < tok.getLength(); k++) {
                    if (off >= vpLeftColumn + mainArea.getCols()) {
                        break;
                    }
                    final char thisChar = tf.charAt(tok.getStartOffset() + k);
                    if (thisChar == '\t') {
                        final int nl = ((off / getTabstop()) + 1) * getTabstop();
                        final int o1 = off;
                        for (int kk = 0; kk < nl - o1; kk++) {
                            if (off >= vpLeftColumn) {
                                if (kk == nl - o1 - 1) {
                                    mainArea.print(DISPLAYABLE_TAB);
                                } else {
                                    mainArea.print(DISPLAYABLE_TABFILL);
                                }
                            }
                            off++;
                        }
                    } else if (thisChar == '\n') {
                        if (off >= vpLeftColumn) {
                            mainArea.print(DISPLAYABLE_NL);
                        }
                        off++;
                    } else if (thisChar == '\r') {
                        if (off >= vpLeftColumn) {
                            mainArea.print(DISPLAYABLE_CR);
                        }
                        off++;
                    } else {
                        if (off >= vpLeftColumn) {
                            mainArea.print(thisChar);
                        }
                        off++;
                    }
                }
            }
        }
    }

    //
    // Display the whole screen. This involves drawing each and every bit of
    // the screen. Hopefully, when the curses library gets beefed up, this
    // will be called less often. This is because for single line scrolling,
    // there should be a scrollup() method in the WritableCursesModel class.
    //
    void displayAll() {
        final WritableCursesModel mainArea = orionComponent.getMainArea();

        mainArea.setBatch(true);
        final Style style = settings.getStyleContext().getStyle(
                StyleContext.DEFAULT_STYLE);
        mainArea.setFG(StyleConstants.getForeground(style));
        mainArea.setBG(StyleConstants.getBackground(style));
        mainArea.setUnderLine(StyleConstants.isUnderline(style));
        mainArea.clear();

        displayAllLines();

        //
        // Finally move the cursor, and get the thing to paint by setting
        // back to interactive mode.
        //
        mainArea.moveTo(cursorColumn - vpLeftColumn, cursorLine - vpTopLine);
        mainArea.setBatch(false);

        //
        // Now update the position information
        //
        displayPos();
    }

    @Override
    public void activate() {
        super.activate();

        final WritableCursesModel mainArea = orionComponent.getMainArea();
        // This is TERRIBLE - we need a better way!!
        orionComponent.getMousableComponent().addMouseListener(ml);
        orionComponent.getMousableComponent().addMouseWheelListener(mwl);
        mainArea.setCursorVisible(true);

        final String prefix = getBundle().getString(PREFIX + "fileprefix");
        final WritableCursesModel header = orionComponent.getHeaderArea();
        header.setBatch(true);
        header.clear();
        header.moveTo(0, 0);
        header.setFG(Color.white);
        header.print(prefix);
        header.print(' ');
        header.setFG(Color.yellow);
        header.print(
                Utils.truncateFile(fileName,
                        header.getCols() - prefix.length() - 1));
        header.setBatch(false);

        displayAll();
    }

    @Override
    public void deactivate() {
        super.deactivate();

        final WritableCursesModel mainArea = orionComponent.getMainArea();
        mainArea.setCursorVisible(false);
        // This is TERRIBLE - we need a better way!!
        orionComponent.getMousableComponent().removeMouseListener(ml);
        orionComponent.getMousableComponent().removeMouseWheelListener(mwl);
    }

    //
    // This routine moves the cursorOffset and cursorColumn to the 'magic'
    // position. On input to this routine, the cursorLine and cursorMagic
    // variables must be set, and the others are changed based on those 2
    // pieces of information.
    // Note: It does not adjust the viewport.
    //
    void setCursorToMagic() {
        final List<LanguageToken> tlv = tf.getLine(cursorLine);
        cursorOffset = tlv.get(0).getStartOffset();

        switch (tf.charAt(cursorOffset)) {
            case '\r':
            case '\n':
                cursorColumn = 0;
                return;
            case '\t':
                cursorColumn = getTabstop() - 1;
                break;
            default:
                cursorColumn = 0;
                break;
        }

        //
        // Loop down the characters, moving the cursor right the correct
        // number of places, until either the cursor is at (or after) the
        // magic position, or we hit a newline.
        //
        boolean newlineHit = false;
        while ((cursorColumn < cursorMagic) && !newlineHit) {
            cursorOffset++;
            switch (tf.charAt(cursorOffset)) {
                case '\t':
                    cursorColumn++;
                    cursorColumn = (cursorColumn / getTabstop()) * getTabstop() + getTabstop() - 1;
                    break;
                case '\r':
                case '\n':
                    cursorOffset--;
                    newlineHit = true;
                    break;
                default:
                    cursorColumn++;
                    break;
            }
        }
    }

    //
    // Make the cursor be in the viewport (from a horizontal perspective).
    // We first keep moving left until the cursor is visible, then we move
    // right till it is visible.
    //
    void adjustViewportHoriz() {
        final int cols = orionComponent.getMainArea().getCols();
        while (vpLeftColumn > cursorColumn) {
            vpLeftColumn -= getTabstop();
        }
        while (cursorColumn >= vpLeftColumn + cols) {
            vpLeftColumn += getTabstop();
        }
    }

    void displayCursor() {
        final WritableCursesModel mainArea = orionComponent.getMainArea();
        mainArea.moveTo(cursorColumn - vpLeftColumn, cursorLine - vpTopLine);
        displayPos();
    }

    private boolean isCursorInViewportHeightwise() {
        final int rows = orionComponent.getMainArea().getRows();
        return (cursorLine >= vpTopLine) && (cursorLine < (vpTopLine + rows));
    }

    boolean isCursorInViewportWidthwise() {
        return (cursorColumn >= vpLeftColumn) && (cursorColumn < (vpLeftColumn + orionComponent.getMainArea().getCols()));
    }

    void adjustViewportAndDisplay(int oldCursorLine,
            int oldCursorCol) {
        if (adjustViewport(oldCursorLine, oldCursorCol)) {
            displayAll();
        } else {
            displayCursor();
        }
    }

    private boolean adjustViewport(int oldCursorLine, int oldCursorCol) {
        boolean viewportAdjusted = false;

        if (!isCursorInViewportHeightwise()) {
            viewportAdjusted = true;
            final int rows = orionComponent.getMainArea().getRows();
            final int moveDistance = Math.abs(oldCursorLine - cursorLine);
            if (moveDistance < rows) {
                if (oldCursorLine > cursorLine) {
                    vpTopLine = cursorLine;
                } else {
                    vpTopLine = cursorLine - (rows - 1);
                }
            } else {
                vpTopLine = cursorLine - (rows / 2);
            }
            vpTopLine = Math.min(vpTopLine, tf.getLineCount() - rows);
            vpTopLine = Math.max(vpTopLine, 0);
        }

        if (!isCursorInViewportWidthwise()) {
            adjustViewportHoriz();
            viewportAdjusted = true;
        }

        return viewportAdjusted;
    }

    private boolean isLastLineVisible() {
        final int rows = orionComponent.getMainArea().getRows();
        return tf.getLineCount() - vpTopLine - 1 < rows;
    }

    int getCurrentTokenIndex() {
        final List<LanguageToken> v = tf.getLine(cursorLine);
        int i;
        for (i = 0; i < v.size(); i++) {
            final LanguageToken tok = v.get(i);
            if (tok.getStartOffset() > cursorOffset) {
                return i - 1;
            }
        }
        return i - 1;
    }

    private LanguageToken getCurrentToken() {
        return getToken(cursorLine, getCurrentTokenIndex());
    }

    private LanguageToken getToken(int line, int tokenIndex) {
        final List<LanguageToken> v = tf.getLine(cursorLine);
        if (tokenIndex >= v.size()) {
            return null;
        }
        return v.get(tokenIndex);
    }

    private void goToTokenForward() {
        int startRow;
        int startIndex;
        List<LanguageToken> currentRowVec;
        int currentRow;
        int currentIndex;
        int endRow;
        int endIndex;

        final int oldCursorLine = cursorLine;
        final int oldCursorCol = cursorColumn;

        endRow = cursorLine;
        endIndex = getCurrentTokenIndex();

        //
        // Move to the next token as the starting point
        //
        currentRowVec = tf.getLine(endRow);
        startIndex = endIndex + 1;
        startRow = endRow;
        if (startIndex >= currentRowVec.size()) {
            startIndex = 0;
            startRow++;
        }
        if (startRow >= tf.getLineCount()) {
            startRow = 0;
        }

        currentRow = startRow;
        currentIndex = startIndex;
        currentRowVec = tf.getLine(startRow);

        //
        // loop until we reach the 'next' thingy
        //
        while ((endRow != currentRow) || (endIndex != currentIndex)) {
            //
            // Skip any blank lines!
            //
            while (currentRowVec.isEmpty()) {
                currentRow++;
                if (currentRow >= tf.getLineCount()) {
                    currentRow = 0;
                    warning("wrapped");
                }
                currentRowVec = tf.getLine(currentRow);
                currentIndex = 0;
            }
            final LanguageToken tok = currentRowVec.get(currentIndex);
            if (tok.getType() == TT_INDEXED) {
                break;
            }

            //
            // Go to the 'next' token
            //
            currentIndex++;
            if (currentIndex >= currentRowVec.size()) {
                currentIndex = 0;
                currentRow++;
                if (currentRow >= tf.getLineCount()) {
                    currentRow = 0;
                    warning("wrapped");
                }
            }
            currentRowVec = tf.getLine(currentRow);
        }

        //
        // Find out where the X psition of the start of the word is!
        //
        cursorOffset = currentRowVec.get(currentIndex).getStartOffset();
        cursorLine = currentRow;
        final int rowStartOffset = currentRowVec.get(0).getStartOffset();

        cursorColumn = 0;
        for (int i = rowStartOffset; i < cursorOffset; i++) {
            switch (tf.charAt(i)) {
                case '\t':
                    cursorColumn
                            = (((cursorColumn / getTabstop()) + 1) * getTabstop());
                    break;
                default:
                    cursorColumn++;
                    break;
            }
        }

        cursorMagic = cursorColumn;

        adjustViewportAndDisplay(oldCursorLine, oldCursorCol);
    }

    private void goToTokenBack() {
        int startRow;
        int startIndex;
        List<LanguageToken> currentRowVec;
        int currentRow;
        int currentIndex;
        int endRow;
        int endIndex;

        final int oldCursorLine = cursorLine;
        final int oldCursorCol = cursorColumn;

        endRow = cursorLine;
        endIndex = getCurrentTokenIndex();

        //
        // Move to the prev token as the starting point
        //
        currentRowVec = tf.getLine(endRow);
        startIndex = endIndex - 1;
        startRow = endRow;
        if (startIndex < 0) {
            startRow--;
            currentRowVec = tf.getLine(startRow);
            startIndex = currentRowVec.size() - 1;
        }
        if (startRow < 0) {
            startRow = tf.getLineCount() - 1;
        }

        currentRow = startRow;
        currentIndex = startIndex;
        currentRowVec = tf.getLine(startRow);

        //
        // loop until we reach the 'next' thingy
        //
        while ((endRow != currentRow) || (endIndex != currentIndex)) {
            //
            // Skip any blank lines!
            //
            while (currentRowVec.isEmpty()) {
                currentRow--;
                if (currentRow < 0) {
                    currentRow = tf.getLineCount() - 1;
                    warning("wrapped");
                }
                currentRowVec = tf.getLine(currentRow);
                currentIndex = currentRowVec.size() - 1;
            }
            final LanguageToken tok = currentRowVec.get(currentIndex);
            if (tok.getType() == TT_INDEXED) {
                break;
            }

            //
            // Go to the 'next' token
            //
            currentIndex--;
            if (currentIndex < 0) {
                currentRow--;
                if (currentRow < 0) {
                    currentRow = tf.getLineCount() - 1;
                    warning("wrapped");
                }
                currentRowVec = tf.getLine(currentRow);
                currentIndex = currentRowVec.size() - 1;
            }
        }

        //
        // Find out where the X psition of the start of the word is!
        //
        cursorOffset = currentRowVec.get(currentIndex).getStartOffset();
        cursorLine = currentRow;
        final int rowStartOffset = currentRowVec.get(0).getStartOffset();

        cursorColumn = 0;
        for (int i = rowStartOffset; i < cursorOffset; i++) {
            switch (tf.charAt(i)) {
                case '\t':
                    cursorColumn
                            = (((cursorColumn / getTabstop()) + 1) * getTabstop());
                    break;
                default:
                    cursorColumn++;
                    break;
            }
        }

        cursorMagic = cursorColumn;

        adjustViewportAndDisplay(oldCursorLine, oldCursorCol);
    }

    private void goToToken(boolean back) {
        if (back) {
            goToTokenBack();
        } else {
            goToTokenForward();
        }
    }

    /**
     * Go to the end of the current line.
     * <p>
     * This method does <b>not</b> move the viewport or redisplay. It is only
     * responsible for logically moving the cursor.
     */
    void goToEndOfLine() {
        final List<LanguageToken> tlv = tf.getLine(cursorLine);
        int so = tlv.get(0).getStartOffset();
        final LanguageToken lastToken = tlv.get(tlv.size() - 1);
        cursorOffset = lastToken.getStartOffset() - 1;
        if (tlv.size() < 2) {
            cursorOffset = 0;
        }
        cursorColumn = 0;
        while (so < cursorOffset) {
            if (tf.charAt(so) == '\t') {
                cursorColumn
                        = ((cursorColumn / getTabstop()) + 1) * getTabstop();
            } else {
                cursorColumn++;
            }
            so++;
        }
    }

    int getTabstop() {
        return settings.getTabstop();
    }

    private void goToOffset(int off) {
        final int oldCursorLine = cursorLine;
        final int oldCursorCol = cursorColumn;

        cursorOffset = off;
        int line = tf.getLineCount() - 1;
        for (int i = 1; i < tf.getLineCount(); i++) {
            final LanguageToken thisToken = tf.getLine(i).get(0);
            if (thisToken.getStartOffset() > off) {
                line = i - 1;
                break;
            }
        }
        cursorLine = line;
        cursorColumn = 0;
        final List<LanguageToken> crv = tf.getLine(line);
        final int rowStartOffset = crv.get(0).getStartOffset();
        for (int i = rowStartOffset; i < cursorOffset; i++) {
            switch (tf.charAt(i)) {
                case '\t':
                    cursorColumn
                            = (((cursorColumn / getTabstop()) + 1) * getTabstop());
                    break;
                default:
                    cursorColumn++;
                    break;
            }
        }

        cursorMagic = cursorColumn;

        adjustViewportAndDisplay(oldCursorLine, oldCursorCol);
    }

    private boolean rfind(Pattern pattern) {
        //
        // There is no REGEX API for reverse finding. Because of that, we
        // have to use a brute force method of doing it.
        //
        final Matcher matcher = pattern.matcher(tf.getData());
        final int so = cursorOffset;
        boolean brc = matcher.find(0);
        if (brc) {
            int ts = -1;
            while (brc) {
                final int tmp = matcher.start();
                if (tmp < so) {
                    ts = tmp;
                    brc = matcher.find();
                } else {
                    break;
                }
            }
            if (ts == -1) {
                ts = matcher.start();
                while (matcher.find()) {
                    ts = matcher.start();
                }
            }
            goToOffset(ts);
            if (ts >= so) {
                warning("wrapped");
            }
            return true;
        }
        return false;
    }

    private boolean find(Pattern pattern) {
        final Matcher matcher = pattern.matcher(tf.getData());
        final int so = cursorOffset + 1;
        final boolean brc = matcher.find(cursorOffset + 1) || matcher.find(0);
        if (brc) {
            goToOffset(matcher.start());
            if (cursorOffset < so) {
                warning("wrapped");
            }
            return true;
        }
        return false;
    }

    private void goToLine(int num) {
        info(null);
        final int oldCursorLine = cursorLine;
        final int oldCursorCol = cursorColumn;
        cursorLine = num - 1;
        setCursorToMagic();
        adjustViewportAndDisplay(oldCursorLine, oldCursorCol);
    }

    /**
     * Get the name of the currently viewed file. This method is primarily used
     * by the outer layers for scripting purposes.
     *
     * @return The (relative) name of the file.
     */
    public String getFileName() {
        return fileName;
    }

    /**
     * Get the location of the current view of the file. This is defined as
     * being the index of the top visible line.
     *
     * @return The location of the top visible line. This is a one-based index.
     */
    public int getFileLocation() {
        return vpTopLine + 1;
    }

    /**
     * Go to a specific row/column pair.
     * <p>
     * No movement is done if this is not in the file, and an error is produced.
     *
     * @param row The row
     * @param col The column
     */
    private void goToRowCol(int row, int col) {
        if ((row <= 0) || (col <= 0)) {
            return;
        }
        final int newCursorLine = vpTopLine + row - 1;
        if (newCursorLine >= tf.getLineCount()) {
            error("bottom");
            return;
        }

        cursorLine = newCursorLine;

        cursorMagic = vpLeftColumn + col - 1;
        setCursorToMagic();

        if (!isCursorInViewportWidthwise()) {
            adjustViewportHoriz();
            displayAll();
        } else {
            displayCursor();
        }
    }

    private class NextFindAction extends AbstractAction {

        private boolean reverse;

        public NextFindAction(boolean reverse) {
            super();
            this.reverse = reverse;
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            info(null);
            final boolean found;
            if (reverse) {
                found = rfind(orionComponent.getLastSearch());
            } else {
                found = find(orionComponent.getLastSearch());
            }
            if (!found) {
                error("searchnotfound", orionComponent.getLastSearch().pattern());
            }
        }
    }

    private class FindAction extends AbstractAction implements OrionInputHandler {

        private final boolean reverse;

        public FindAction(boolean reverse) {
            super();
            this.reverse = reverse;
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            final String s = reverse ? "findback" : "findforward";
            orionComponent.startInput(
                    getBundle().getString(PREFIX + s),
                    orionComponent.getLastSearch().pattern(),
                    this);
        }

        @Override
        public void handleInput(String input) {
            final Pattern p = Pattern.compile(input, Pattern.MULTILINE);
            orionComponent.setLastSearch(p);
            amh.invokeAction(reverse ? "previousFind" : "nextFind");
        }
    }

    private class OpenFileAction extends AbstractAction implements OrionInputHandler {

        public OpenFileAction() {
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            orionComponent.startInput(
                    getBundle().getString(PREFIX + "openfile"),
                    "",
                    this);
        }

        @Override
        public void handleInput(String input) {
            if (input.length() == 0) {
                error("nofile");
                return;
            }
            final OpenFileSwingWorker ofsw = new OpenFileSwingWorker(orionComponent, input, "");
            ofsw.execute();
        }
    }

    private class GotoLineAction extends AbstractAction
            implements OrionInputHandler {

        public GotoLineAction() {
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            orionComponent.startInput(
                    getBundle().getString(PREFIX + "line"),
                    "",
                    this);
        }

        @Override
        public void handleInput(String input) {
            try {
                final int num = Integer.parseInt(input);
                if ((num <= 0) || (num > tf.getLineCount())) {
                    error("linenumoutofrange");
                } else {
                    goToLine(num);
                }
            } catch (NumberFormatException e) {
                error("badlinenum");
            }
        }
    }

    private class NextTokenAction extends AbstractAction {

        private final boolean reverse;

        public NextTokenAction(boolean reverse) {
            super();
            this.reverse = reverse;
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            info(null);
            goToToken(reverse);
        }
    }

    private class TopAction extends AbstractAction {

        public TopAction() {
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            info(null);
            final int oldCursorLine = cursorLine;
            final int oldCursorCol = cursorColumn;

            cursorOffset = 0;
            cursorLine = 0;
            cursorColumn = 0;
            cursorMagic = 0;

            adjustViewportAndDisplay(oldCursorLine, oldCursorCol);
        }
    }

    private class BottomAction extends AbstractAction {

        public BottomAction() {
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            info(null);
            final int oldCursorLine = cursorLine;
            final int oldCursorCol = cursorColumn;

            cursorOffset = tf.getCharCount() - 1;
            cursorLine = tf.getLineCount() - 1;

            goToEndOfLine();

            cursorMagic = cursorColumn;

            adjustViewportAndDisplay(oldCursorLine, oldCursorCol);
        }
    }

    private class SelectAction extends AbstractAction {

        public SelectAction() {
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            final LanguageToken tok = getCurrentToken();
            if (tok != null) {
                if (tok.getType().isIndexable()) {
                    final String word = tf.tokenAsString(tok);
                    final HuntSpecificSwingWorker hsw = new HuntSpecificSwingWorker(orionComponent, word);
                    hsw.execute();
                    return;
                }
            }
            error("putonword");
        }
    }

    private class ViewportUpAction extends AbstractAction {

        public ViewportUpAction() {
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            if (cursorLine == 0
                    || vpTopLine == 0) {
                error("top");
                displayAll();
                return;
            }

            info(null);

            final int oldCursorLine = cursorLine;
            final int oldCursorCol = cursorColumn;

            cursorLine = vpTopLine - 1;
            adjustViewport(cursorLine + 1, oldCursorCol);
            cursorLine = oldCursorLine;
            if (!isCursorInViewportHeightwise()) {
                cursorLine--;
                setCursorToMagic();

                adjustViewport(oldCursorLine, oldCursorCol);
            }
            displayAll();
        }
    }

    private class ViewportDownAction extends AbstractAction {

        public ViewportDownAction() {
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            if (cursorLine == tf.getLineCount() - 1) {
                error("bottom");
                displayAll();
                return;
            }

            info(null);

            final int oldCursorLine = cursorLine;
            final int oldCursorCol = cursorColumn;

            final int rows = orionComponent.getMainArea().getRows();
            cursorLine = vpTopLine + rows;
            adjustViewport(cursorLine - 1, oldCursorCol);
            cursorLine = oldCursorLine;
            if (!isCursorInViewportHeightwise()) {
                cursorLine++;
                setCursorToMagic();

                adjustViewport(oldCursorLine, oldCursorCol);
            }
            displayAll();
        }
    }

    private class PageUpAction extends AbstractAction {

        public PageUpAction() {
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            if (vpTopLine == 0) {
                error("top");
                return;
            }

            info(null);

            final int rowsToSubtract = orionComponent.getMainArea().getRows();
            cursorLine -= rowsToSubtract;
            if (cursorLine < 0) {
                cursorLine = 0;
            }
            vpTopLine -= rowsToSubtract;
            if (vpTopLine < 0) {
                vpTopLine = 0;
            }

            setCursorToMagic();

            adjustViewportHoriz();

            displayAll();
        }
    }

    private class PageDownAction extends AbstractAction {

        public PageDownAction() {
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            info(null);
            final int rowsToAdd = orionComponent.getMainArea().getRows();
            final int newCursorLine = cursorLine + rowsToAdd;
            if (newCursorLine >= tf.getLineCount()) {
                if (isLastLineVisible()) {
                    error("bottom");
                    return;
                } else {
                    final int oldCursorLine = cursorLine;
                    final int oldCursorCol = cursorColumn;
                    cursorLine = tf.getLineCount() - 1;
                    setCursorToMagic();

                    adjustViewportAndDisplay(oldCursorLine, oldCursorCol);
                    return;
                }
            }
            cursorLine = newCursorLine;

            vpTopLine += rowsToAdd;
            if (vpTopLine + rowsToAdd > tf.getLineCount()) {
                vpTopLine = tf.getLineCount() - rowsToAdd;
                if (vpTopLine < 0) {
                    vpTopLine = 0;
                }
            }
            setCursorToMagic();

            adjustViewportHoriz();

            displayAll();
        }
    }

    private class JavaNativeHuntAction extends AbstractAction {

        public JavaNativeHuntAction() {
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            final LanguageToken tok = getCurrentToken();
            if (tok != null && tok.getType().isIndexable()) {
                final Navigator nav = tf.navigator();
                nav.gotoLine(cursorLine);
                if (rowContainsNative(nav)) {
                    final String pack = getPackage(nav);
                    final String clazz = getClass(nav);
                    final String token = tf.tokenAsString(tok);
                    final String full = "Java_" + pack + '_' + clazz + '_' + token;
                    final String word = full.replace('.', '_');
                    final HuntSwingWorker hsw = new HuntSwingWorker(orionComponent, word);
                    hsw.execute();
                }
            }
            error("putonword");
        }

        private boolean rowContainsNative(Navigator nav) {
            while (nav.hasNext()) {
                final LanguageToken tok = nav.next();
                if (tok.getType() == TokenTypes.TT_NEWLINE) {
                    return false;
                } else if (tok.getType() == TokenTypes.TT_KEYWORD) {
                    if (tf.tokenContains(tok, "native")) {
                        return true;
                    }
                }
            }
            return false;
        }

        private String getClass(Navigator nav) {
            nav.gotoLine(0);
            int state = BASE;
            while (nav.hasNext()) {
                final LanguageToken tok = nav.next();
                switch (state) {
                    case BASE:
                        if ((tok.getType() == TokenTypes.TT_KEYWORD) && tf.tokenContains(tok, "class")) {
                            state = PACKAGE;
                        }
                        break;
                    case PACKAGE:
                        if (tok.getType() == TokenTypes.TT_WORD) {
                            return tf.tokenAsString(tok);
                        }
                        break;
                    default:
                        throw new IllegalStateException();
                }
            }
            return null;
        }

        private String getPackage(Navigator nav) {
            nav.gotoLine(0);
            final StringBuilder sb = new StringBuilder();
            int state = BASE;
            loop:
            while (nav.hasNext()) {
                final LanguageToken tok = nav.next();
                switch (state) {
                    case BASE:
                        if ((tok.getType() == TokenTypes.TT_KEYWORD) && tf.tokenContains(tok, "package")) {
                            state = PACKAGE;
                        }
                        break;
                    case PACKAGE:
                        if ((tok.getType() == TokenTypes.TT_NEWLINE) || (tok.getType() == TokenTypes.TT_COMMENT)) {
                            break loop;
                        } else {
                            sb.append(tf.tokenAsString(tok));
                        }
                        break;
                    default:
                        throw new IllegalStateException();
                }
            }
            if (sb.length() > 0) {
                sb.setLength(sb.length() - 1);
            }
            return sb.toString().trim();
        }
    }
    private static final int BASE = 0;
    private static final int PACKAGE = 1;

    void error(String s, Object ... inserts) {
        getMessenger().error(format(PREFIX + s, inserts));
    }

    void warning(String s) {
        getMessenger().warning(getBundle().getString(PREFIX + s));
    }

    void info(String s) {
        final String message;
        if (s == null) {
            message = s;
        } else {
            message = getBundle().getString(PREFIX + s);
        }
        getMessenger().info(message);
    }

    private ResourceBundle getBundle() {
        return ResourceBundle.getBundle(
                "uk.co.ruelgnoj.orion.component.resources",
                orionComponent.getLocale());
    }

    private class InternalsImpl implements Internals {

        public InternalsImpl() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void setCursorMagic(int magic) {
            cursorMagic = magic;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public int getCursorMagic() {
            return cursorMagic;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void setCursorColumn(int column) {
            cursorColumn = column;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public int getCursorColumn() {
            return cursorColumn;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void setCursorLine(int line) {
            cursorLine = line;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public int getCursorLine() {
            return cursorLine;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public int getCursorOffset() {
            return cursorOffset;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void setCursorOffset(int offset) {
            cursorOffset = offset;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public int getTabstop() {
            return FileView.this.getTabstop();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public TokenizedFile getTokenizedFile() {
            return tf;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public int getLineCount() {
            return tf.getLineCount();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public int getCurrentTokenIndex() {
            return FileView.this.getCurrentTokenIndex();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void error(String message) {
            FileView.this.error(message);
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void warning(String message) {
            FileView.this.warning(message);
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void info(String message) {
            FileView.this.info(message);
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void clearMessages() {
            FileView.this.info(null);
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void adjustViewportAndDisplay(int oldCursorLine,
                int oldCursorCol) {
            FileView.this.adjustViewportAndDisplay(oldCursorLine, oldCursorCol);
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void displayAll() {
            FileView.this.displayAll();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void displayCursor() {
            FileView.this.displayCursor();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void adjustViewportHoriz() {
            FileView.this.adjustViewportHoriz();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void goToEndOfLine() {
            FileView.this.goToEndOfLine();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void setCursorToMagic() {
            FileView.this.setCursorToMagic();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean isCursorInViewportWidthwise() {
            return FileView.this.isCursorInViewportWidthwise();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken getCurrentToken() {
            return FileView.this.getCurrentToken();
        }
    }
}
