/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: TokenizedFile.java 285 2011-07-19 12:37:42Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.component.fileview;

import java.util.ArrayList;
import java.util.List;

import uk.co.ruelgnoj.orion.tokenizers.LanguageToken;
import uk.co.ruelgnoj.orion.tokenizers.LanguageTokenizer;
import uk.co.ruelgnoj.orion.tokenizers.TokenTypes;

public class TokenizedFile {

    private final LanguageTokenizer tokenizer;
    private final CharSequence data;
    private final List<List<LanguageToken>> lines;

    public TokenizedFile(LanguageTokenizer tokenizer, CharSequence data) {
        ensureNotNull(tokenizer, "tokenizer");
        ensureNotNull(data, "data");

        this.tokenizer = tokenizer;
        this.data = data;

        lines = new ArrayList<>();
        initialize();
    }

    private static void ensureNotNull(Object obj, String name) {
        if (obj == null) {
            throw new NullPointerException(name);
        }
    }

    private void initialize() {
        tokenizer.setBuffer(data);
        List<LanguageToken> v = new ArrayList<>();

        for (LanguageToken tok : tokenizer) {
            v.add(tok);

            if (tok.getType() == TokenTypes.TT_NEWLINE) {
                lines.add(v);
                v = new ArrayList<>();
            }
        }
        if (!v.isEmpty()) {
            lines.add(v);
        }
    }

    public boolean tokenContains(LanguageToken token, String text) {
        return token.asString(data).equals(text);
    }

    public boolean tokenContains(LanguageToken token, char[] text) {
        return tokenContains(token, new String(text));
    }

    public String tokenAsString(LanguageToken token) {
        return token.asString(data);
    }

    public int getLineCount() {
        return lines.size();
    }

    public List<LanguageToken> getLine(int index) {
        if ((index < 0) || (index >= getLineCount())) {
            throw new IllegalArgumentException("index=" + index);
        }
        return lines.get(index);
    }

    public char charAt(int index) {
        return data.charAt(index);
    }

    public int getCharCount() {
        return data.length();
    }

    public Navigator navigator() {
        return new NavigatorImpl();
    }

    public CharSequence getData() {
        return data;
    }

    private class NavigatorImpl implements Navigator {

        private int currentLineIndex;
        private int tokenIndex;

        public NavigatorImpl() {
            currentLineIndex = 0;
            tokenIndex = -1;
        }

        @Override
        public void gotoLine(int lineIndex) {
            if ((lineIndex < 0) || (lineIndex >= getLineCount())) {
                throw new IllegalArgumentException();
            }
            currentLineIndex = lineIndex;
            tokenIndex = -1;
        }

        @Override
        public LanguageToken gotoCharacter(int charIndex) {
            throw new Error();
        }

        @Override
        public boolean hasPrevious() {
            return (currentLineIndex > 0) || (tokenIndex > 0);
        }

        @Override
        public boolean hasNext() {
            return (currentLineIndex != getLineCount() - 1)
                    || (tokenIndex != getLine(getLineCount() - 1).size() - 1);
        }

        @Override
        public LanguageToken previous() {
            if (!hasPrevious()) {
                throw new IllegalStateException();
            }
            tokenIndex--;
            if (tokenIndex < 0) {
                currentLineIndex--;
                tokenIndex = getLine(currentLineIndex).size() - 1;
            }
            return current();
        }

        @Override
        public LanguageToken next() {
            if (!hasNext()) {
                throw new IllegalStateException();
            }
            tokenIndex++;
            if (tokenIndex >= getLine(currentLineIndex).size()) {
                currentLineIndex++;
                tokenIndex = 0;
            }
            return current();
        }

        @Override
        public LanguageToken current() {
            return getLine(currentLineIndex).get(tokenIndex);
        }
    }
}
