/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: MnemonicHelper.java 285 2011-07-19 12:37:42Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package uk.co.ruelgnoj.orion.component.util;

import java.awt.event.KeyEvent;

import java.lang.reflect.Field;

import java.util.MissingResourceException;
import java.util.ResourceBundle;

import javax.swing.AbstractButton;
import javax.swing.Action;
import javax.swing.JLabel;

/**
 * A Helper class that assigns a Mnemonic to a component when supplied with a
 * ResourceBundle and item name.
 * The name and mnemonic of the component come from the ResourceBundle.
 * @author Steve Millington
 * @version $Id: MnemonicHelper.java 285 2011-07-19 12:37:42Z steve $
 */
public final class MnemonicHelper {

    /**
     * Private Constructor.
     */
    private MnemonicHelper() {
    }

    /**
     * Sets a Button's properties as defined by the resource bundle.
     * The mnemonic should be defined as follows:
     * <code>&lt;itemname&gt;.mnemonic</code>
     * @param aButton The Button.
     * @param bundle The source of the information defining the Button.
     * @param itemName The name of the Button within the ResourceBundle.
     * @return A Button.
     */
    public static AbstractButton assignMnemonic(
	    			AbstractButton aButton,
	    			ResourceBundle bundle,
				String itemName) {
	// CHECKSTYLE-OFF: EmptyBlockCheck
	try {
	    final String s = bundle.getString(itemName + ".mnemonic");
	    final int n = getKeyCode(s);
	    aButton.setMnemonic(n);
	} catch (MissingResourceException e) {
	    // This is fine - no mnemonic required
	} catch (IllegalArgumentException e) {
	    // Mnemonic code wrong - warn the tester!
	    System.out.println(": Warning, : '" + itemName
		    + "' has illegal mnemonic. " + e.getMessage());
	}
	return aButton;
	// CHECKSTYLE-ON: EmptyBlockCheck
    }

    /**
     * Sets a JLabel's properties as defined by the resource bundle.
     * The mnemonic should be defined as follows:
     * <code>&lt;itemname&gt;.mnemonic</code>
     * @param aJLabel The JLabel.
     * @param bundle The source of the information defining the JLabel.
     * @param itemName The name of the JLabel within the ResourceBundle.
     * @return A JLabel.
     */
    public static JLabel assignMnemonic(
	    			JLabel aJLabel,
	    			ResourceBundle bundle,
				String itemName) {
	// CHECKSTYLE-OFF: EmptyBlockCheck
	try {
	    final String s = bundle.getString(itemName + ".mnemonic");
	    final int n = getKeyCode(s);
	    aJLabel.setDisplayedMnemonic(n);
	} catch (MissingResourceException e) {
	    // This is fine - no mnemonic required
	} catch (IllegalArgumentException e) {
	    // Mnemonic code wrong - warn the tester!
	    System.out.println(": Warning, label: '" + itemName
		    + "' has illegal mnemonic. " + e.getMessage());
	}
	// CHECKSTYLE-ON: EmptyBlockCheck
	return aJLabel;
    }

    /**
     * Sets an Action's properties as defined by the resource bundle.
     * The mnemonic should be defined as follows:
     * <code>&lt;itemname&gt;.mnemonic</code>
     * @param anAction The Action.
     * @param bundle The source of the information defining the Action.
     * @param itemName The name of the Action within the ResourceBundle.
     */
    public static void assignMnemonic(
	    			Action anAction,
	    			ResourceBundle bundle,
				String itemName) {
	// CHECKSTYLE-OFF: EmptyBlockCheck
	try {
	    final String s = bundle.getString(itemName + ".mnemonic");
	    final int n = getKeyCode(s);
	    anAction.putValue(Action.MNEMONIC_KEY, n);
	} catch (MissingResourceException e) {
	    // This is fine - no mnemonic required
	} catch (IllegalArgumentException e) {
	    // Mnemonic code wrong - warn the tester!
	    System.out.println(": Warning, action: '" + itemName
		    + "' has illegal mnemonic. " + e.getMessage());
	}
	// CHECKSTYLE-ON: EmptyBlockCheck
    }

    /**
     * Get the KeyCode from a string.
     * <p>The Key Code is determined by prefixing the uppercased string with
     * <code>VK_</code> and then finding the field in the <code>KeyEvent</code>
     * class.
     * <p>For example:
     * <pre>
     * HOME ---- KeyEvent.VK_HOME
     * home ---- KeyEvent.VK_HOME
     * a ------- KeyEvent.VK_A
     * A ------- KeyEvent.VK_A
     * </pre>
     * @param keyCodeString The stringified code.
     * @return The Key Code
     * @throws NullPointerException If <code>s</code> is
     * <code>null</code>
     * @throws IllegalArgumentException If <code>keyCodeString</code>
     * cannot be converted to a code.
     * @see java.awt.event.KeyEvent
     */
    private static int getKeyCode(String keyCodeString) {
	if (keyCodeString == null) {
	    throw new NullPointerException("keyCodeString");
	}
	final String lookup = "VK_" + keyCodeString.toUpperCase();
	try {
	    final Field f = KeyEvent.class.getField(lookup);
	    final int ret = f.getInt(null);
	    return ret;
	} catch (IllegalAccessException | NoSuchFieldException e) {
	    throw new IllegalArgumentException("keyCodeString=" + keyCodeString, e);
	}
    }
}
