/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: RBAction.java 256 2008-07-15 15:54:02Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.component.util;

import java.util.MissingResourceException;
import java.util.ResourceBundle;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ImageIcon;
import javax.swing.KeyStroke;

/**
 * An action taking all of its values from a resource bundle.
 * <p>
 * Each action must have entries in the supplied resource bundle. The action is
 * conceptually part of a set and has a name.
 * <p>
 * The resources used by this class are:
 * <ul>
 * <li><code>action.%setName%.%actionName%</code></li>
 * <li><code>action.%setName%.%actionName%.icon</code></li>
 * <li><code>action.%setName%.%actionName%.name</code></li>
 * <li><code>action.%setName%.%actionName%.description</code></li>
 * <li><code>action.%setName%.%actionName%.accelerator</code></li>
 * </ul>
 * <p>
 * All of these are optional, with defaults being used where appropriate.
 *
 * @author Steve Millington
 * @version $Id: RBAction.java 256 2008-07-15 15:54:02Z steve $
 */
public abstract class RBAction extends AbstractAction {

    /**
     * Constructor.
     *
     * @param theBundle The resource bundle to obtain information from.
     * @param actionSet The name of the action set
     * @param actionName The name of the action
     * @throws NullPointerException if any parameter is null
     * @throws IllegalArgumentException if either name is empty
     */
    public RBAction(ResourceBundle theBundle,
            String actionSet,
            String actionName) {
        checkNotNull(theBundle, "theBundle");
        checkNotNull(actionSet, "actionSet");
        checkNotNull(actionName, "actionName");
        checkNotEmpty(actionSet, "actionSet");
        checkNotEmpty(actionName, "actionName");

        putValue("HREF", "action." + actionSet + "." + actionName);

        String s;
        try {
            s = theBundle.getString(actionSet + "." + actionName + ".icon");
            final ImageIcon image = new ImageIcon(getClass().getResource(s));
            putValue(Action.SMALL_ICON, image);
            // CHECKSTYLE-OFF: EmptyBlockCheck
        } catch (MissingResourceException ex) {
            // Don't worry
        }
        // CHECKSTYLE-ON: EmptyBlockCheck

        try {
            s = theBundle.getString(actionSet + "." + actionName + ".name");
        } catch (MissingResourceException ex) {
            s = actionName;
        }
        putValue(Action.NAME, s);
        putValue(Action.DEFAULT, s);

        try {
            s = theBundle.getString(
                    actionSet + "." + actionName + ".description");
        } catch (MissingResourceException ex) {
            s = actionName;
        }
        putValue(Action.SHORT_DESCRIPTION, s);
        putValue(Action.LONG_DESCRIPTION, s);

	//
        // Mnemonics and accelerators.
        //
        MnemonicHelper.assignMnemonic(this, theBundle, actionSet + "." + actionName);
        try {
            final String acc = theBundle.getString(
                    actionSet + "." + actionName + ".accelerator");
            final KeyStroke ke = KeyStroke.getKeyStroke(acc);
            if (ke != null) {
                putValue(Action.ACCELERATOR_KEY, ke);
            }
            // CHECKSTYLE-OFF: EmptyBlockCheck
        } catch (MissingResourceException ex) {
            // Don't worry
        }
        // CHECKSTYLE-ON: EmptyBlockCheck
    }

    /**
     * Ensure that the supplied object is not null.
     *
     * @param o The object to check
     * @param message The exception message
     */
    private void checkNotNull(Object o, String message) {
        if (o == null) {
            throw new NullPointerException(message);
        }
    }

    /**
     * Ensure that the supplied string is not empty.
     *
     * @param s The string to check
     * @param message The exception message
     */
    private void checkNotEmpty(String s, String message) {
        if (s.length() == 0) {
            throw new IllegalArgumentException(message);
        }
    }
}
