/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: AbstractJDBCOrionDataModel.java 256 2008-07-15 15:54:02Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package uk.co.ruelgnoj.orion.defdm;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.nio.CharBuffer;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;
import uk.co.ruelgnoj.orion.dm.OrionDataModel;
import uk.co.ruelgnoj.orion.dm.OrionDataModelException;
import uk.co.ruelgnoj.orion.dm.OrionFileEntry;

/**
 *
 * @author Steve Millington
 * @version $Id: AbstractJDBCOrionDataModel.java 256 2008-07-15 15:54:02Z steve $
 */
public abstract class AbstractJDBCOrionDataModel implements OrionDataModel {

    private static final String PROJECTS_TABLE = "PROJECTS";
    private String tagTable_;
    private String fileTable_;
    private String baseDirectory_;
//    private Connection conn_;
    public AbstractJDBCOrionDataModel() throws OrionDataModelException {
    }

    protected final void init() throws OrionDataModelException {
        Connection conn = null;
        try {
            conn = borrowConnection();
            init(conn);
        } catch (SQLException e) {
            throw new OrionDataModelException(e);
        } finally {
            returnConnection(conn);
        }
    }

    private void init(Connection conn) throws SQLException {
        final String txt = "SELECT TAGTABLE,FILETABLE,BASEDIR FROM " + PROJECTS_TABLE + " WHERE NAME = " + quote(getProjectName());
        try (Statement stmt = conn.createStatement(); ResultSet rs = stmt.executeQuery(txt)) {
            rs.next();
            tagTable_ = rs.getString(1);
            fileTable_ = rs.getString(2);
            baseDirectory_ = rs.getString(3);
        }
    }
    
    protected abstract String getProjectName();

    protected abstract Connection borrowConnection() throws SQLException;

    protected abstract void returnConnection(Connection conn);

    @Override
    public final List<String> getTags(String pattern) throws OrionDataModelException {
        Connection conn = null;
        try {
            conn = borrowConnection();
            return getTags(conn, pattern);
        } catch (SQLException e) {
            throw new OrionDataModelException(e);
        } finally {
            returnConnection(conn);
        }
    }

    public final List<String> getTags(Connection conn, String pattern) throws OrionDataModelException {
        try {
            final PreparedStatement ps = conn.prepareStatement("SELECT DISTINCT TAG from " + tagTable_ + " WHERE TAGU LIKE ? ORDER BY TAGU,TAG");
            final List<String> ret = new ArrayList<>();
            ps.setString(1, pattern.toUpperCase() + "%");
            try (ResultSet rs = ps.executeQuery()) {
                while (rs.next()) {
                    ret.add(rs.getString(1));
                }
            }
            return ret;
        } catch (SQLException ex) {
            throw new OrionDataModelException(ex);
        }
    }

    @Override
    public final List<OrionFileEntry> getFiles(String tagString) throws OrionDataModelException {
        Connection conn = null;
        try {
            conn = borrowConnection();
            return getFiles(conn, tagString);
        } catch (SQLException e) {
            throw new OrionDataModelException(e);
        } finally {
            returnConnection(conn);
        }
    }

    public final List<OrionFileEntry> getFiles(Connection conn, String tagString) throws OrionDataModelException {
        try {
            final List<OrionFileEntry> ret = new ArrayList<>();
            final PreparedStatement ps = conn.prepareStatement(
                    "SELECT " + fileTable_ + ".FILE," + tagTable_ + ".HITS FROM " + fileTable_ + "," + tagTable_ + " WHERE INDEX = FILEINDEX AND TAG = ? ORDER BY FILE");

            ps.setString(1, tagString);
            try (ResultSet rs = ps.executeQuery()) {
                while (rs.next()) {
                    final OrionFileEntry ofe = new OrionFileEntry(rs.getString(1), rs.getInt(2));
                    ret.add(ofe);
                }
            }
            return ret;
        } catch (SQLException ex) {
            throw new OrionDataModelException(ex);
        }
    }

    @Override
    public final CharSequence getFile(String fileName) throws NullPointerException, OrionDataModelException {
        final File f = new File(baseDirectory_, fileName);
        final char[] dataBuffer = new char[(int) f.length()];

        try (FileReader fr = new FileReader(f);BufferedReader br = new BufferedReader(fr)) {
            br.read(dataBuffer);	// Read the whole file here.
            return CharBuffer.wrap(dataBuffer);
        } catch (IOException e) {
            throw new OrionDataModelException(e);
        }
    }

    @Override
    public final CharSequence getFile(OrionFileEntry fileEntry) throws OrionDataModelException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void destroy() {
    }

    private String quote(String s) {
        final StringBuilder sb = new StringBuilder(s.length() + 2);
        sb.append('\'').append(s).append('\'');
        return sb.toString();
    }
}
