/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: FileOrionDataModel.java 256 2008-07-15 15:54:02Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.defdm;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.RandomAccessFile;

import java.nio.CharBuffer;
import java.util.ArrayList;
import java.util.List;

import java.util.regex.Matcher;
import java.util.regex.Pattern;
import uk.co.ruelgnoj.orion.dm.OrionDataModel;
import uk.co.ruelgnoj.orion.dm.OrionDataModelException;
import uk.co.ruelgnoj.orion.dm.OrionFileEntry;

/**
 * An Orion data model based on a local file system tags file.
 * <p>
 * This class is capable of coping with the underlying file changing underneath
 * it.
 *
 * @author Steve Millington
 * @version $Id: FileOrionDataModel.java 256 2008-07-15 15:54:02Z steve $
 */
public class FileOrionDataModel implements OrionDataModel {

    private static final int CURRENT_VERSION = 0x01;
    private static final int CURRENT_TYPE = 0xabcdef01;
    private static final int BYTE_OF_FF = 0xff;
    private final File tagsFile;		// The file
    private long fileModifyTime;	// The last mod time of the file
    private RandomAccessFile raf;
    private File baseDirectory;
    private long databaseOffset;
    private int version;

    /**
     * Construct a FileOrionDataModel Object.
     *
     * @param fileName The full path to the Tags File.
     *
     * @exception OrionDataModelException If the file is not readable or is of
     * an incorrect format.
     */
    public FileOrionDataModel(String fileName) throws OrionDataModelException {
        this(new File(fileName));
    }

    /**
     * Construct a FileOrionDataModel Object.
     *
     * @param file The full path to the Tags File.
     *
     * @exception OrionDataModelException If the file is not readable or is of
     * an incorrect format.
     */
    public FileOrionDataModel(File file) throws OrionDataModelException {
        tagsFile = file;
        initialize();
    }

    /**
     * Initialize this data model.
     * <p>
     * This initialization is performed at construction time and when the
     * underlying file changes.
     *
     * @throws IOException if any IO errors occur.
     */
    private void initialize() throws OrionDataModelException {
        try {
            //
            // First free up the old file.
            //
            if (raf != null) {
                raf.close();
                raf = null;
            }

            raf = new RandomAccessFile(tagsFile, "r");

            //
            // Probably an old database - no header so reset the raf
            //
            final int fileType = raf.readInt();
            if (fileType != CURRENT_TYPE) {
                //
                // Probably an old database - no header so reset the raf
                //
                version = -1;
                raf.seek(0);
            } else {
                version = raf.readInt();
                if (version != CURRENT_VERSION) {
                    throw new IOException("Not a database");
                }
            }

            //
            // Read the header
            //
            final String dir = raf.readUTF();
            raf.readInt(); // List of files offset
            databaseOffset = (long) raf.readInt();

            baseDirectory = new File(dir);

            //
            // Determine if the base directory is absolute. If not, it specifies
            // a location relative to the tags file.
            //
            if (!baseDirectory.isAbsolute()) {
                baseDirectory = new File(tagsFile.getParentFile(), dir);
            }

            fileModifyTime = tagsFile.lastModified();
        } catch (IOException ex) {
            throw new OrionDataModelException(ex);
        }
    }

    /**
     * Ensure that the file we are using is up-to-date.
     *
     * @throws OrionDataModelException when IO problems occur.
     */
    private void checkFileUpdate() throws OrionDataModelException {
        //
        // If the file has a different modification time, then we need to
        // re-initialize ourself
        //
        final long newTime = tagsFile.lastModified();
        if (fileModifyTime != newTime) {
            initialize();
        }
    }

    /**
     * Destroy this instance.
     */
    @Override
    public void destroy() {
        if (raf != null) {
            try {
                raf.close();
                // CHECKSTYLE-OFF: EmptyBlockCheck
            } catch (IOException e) {
                // Ignore
            }
            // CHECKSTYLE-ON: EmptyBlockCheck
            raf = null;
        }
    }

    /**
     * Get all the tags that match the specified pattern.
     *
     * @param pattern The pattern to match.
     * <p>
     * <b>Note:</b>This implementation only supports xxxx* patterns</p>
     *
     * @return A list containing all the matching tags.
     *
     * @exception OrionDataModelException If there is a problem reading the
     * tagsfile.
     */
    @Override
    public synchronized List<String> getTags(String pattern) throws OrionDataModelException {
        checkFileUpdate();
        final List<String> ret;
        try {
            if (pattern.startsWith("*")) {

                pattern = pattern.substring(1);
                ret = new ArrayList<>();
                treeMatch(getRoot(), Pattern.compile(pattern, Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE), ret);
            } else {
                final Tag t = getSubtree(pattern);
                if (t != null) {
                    ret = new ArrayList<>();
                    treeMatch(t, pattern, ret);
                } else {
                    ret = null;
                }
            }

            return ret;
        } catch (IOException ex) {
            throw new OrionDataModelException(ex);
        }
    }

    /**
     * Get the root tag of the database.
     *
     * @return The root tag.
     * @throws IOException if necessary
     */
    private Tag getRoot() throws IOException {
        raf.seek(databaseOffset);
        final Tag t = new Tag();
        return t;
    }

    /**
     * This recursive method is used to traverse the tree to get all the
     * required matches.
     *
     * @param tag The 'current' tag.
     * @param pattern The pattern to search for.
     * @param list The list to add the matching tags to.
     *
     * @exception IOException If there was any file error.
     */
    private void treeMatch(Tag tag, String pattern, List<String> list)
            throws IOException {
        Tag t;
        pattern = pattern.toUpperCase();

        t = tag.getLt();
        if (t != null) {
            treeMatch(t, pattern, list);
        }

        if (tag.getTagUC().startsWith(pattern)) {
            list.add(tag.getTag());
        }

        t = tag.getGt();
        if (t != null) {
            treeMatch(t, pattern, list);
        }
    }

    /**
     * This recursive method is used to traverse the tree to get all the
     * required matches.
     *
     * @param tag The 'current' tag.
     * @param pattern The pattern to search for.
     * @param list The list to add the matching tags to.
     *
     * @exception IOException If there was any file error.
     */
    private void treeMatch(Tag tag, Pattern pattern, List<String> list) throws IOException {
        Tag t;

        t = tag.getLt();
        if (t != null) {
            treeMatch(t, pattern, list);
        }

        final Matcher matcher = pattern.matcher(tag.getTag());
        if (matcher.find(0)) {
            list.add(tag.getTag());
        }

        t = tag.getGt();
        if (t != null) {
            treeMatch(t, pattern, list);
        }
    }

    /**
     * This method is used to find the top of a subtree that could contain
     * matches for the given pattern.
     *
     * @param tag The 'base' search string.
     * @return the root of a subtree
     * @exception IOException If there was any file error.
     */
    private Tag getSubtree(String tag) throws IOException {
        Tag t = getRoot();
        final String tagUC = tag.toUpperCase();
        while (true) {
            String s = t.getTagUC();
            if (s.length() > tag.length()) {
                s = s.substring(0, tag.length());
            }
            final int cmp = tagUC.compareTo(s);
            if (cmp == 0) {
                return t;
            } else if (cmp < 0) {
                t = t.getLt();
                if (t == null) {
                    return null;
                }
            } else {
                t = t.getGt();
                if (t == null) {
                    return null;
                }
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public CharSequence getFile(OrionFileEntry fileEntry) throws OrionDataModelException {
        return getFile(fileEntry.getFileName());
    }

    /**
     * Get the contents of a file as a character array.
     *
     * @param fileName The name of the file to get the contents of.
     * @return The contents as a char array
     * @throws OrionDataModelException if something goes wrong.
     */
    @Override
    public CharSequence getFile(String fileName) throws OrionDataModelException {
        checkFileUpdate();
        final File f = new File(baseDirectory, fileName);
        try (FileReader fr = new FileReader(f);
                BufferedReader br = new BufferedReader(fr)) {
            final char[] dataBuffer = new char[(int) f.length()];
            // Read the whole file here.
            br.read(dataBuffer);
            return CharBuffer.wrap(dataBuffer);
        } catch (IOException e) {
            throw new OrionDataModelException(e);
        }

    }

    /**
     * Get the files associated with a given tag.
     *
     * @param tagString The tag to get the files for.
     * @throws OrionDataModelException if an error occurs.
     * @return A list of the matching files.
     */
    @Override
    public synchronized List<OrionFileEntry> getFiles(String tagString) throws OrionDataModelException {
        try {
            checkFileUpdate();
            raf.seek(databaseOffset);
            Tag t = new Tag();
            final String tagStringUC = tagString.toUpperCase();
            while (!t.getTag().equals(tagString)) {
                int cmp = tagStringUC.compareTo(t.getTagUC());
                if (cmp == 0) {
                    cmp = tagString.compareTo(t.getTag());
                }
                if (cmp == 0) {
                    throw new Error("Internal Logic Error");
                } else if (cmp < 0) {
                    t = t.getLt();
                    if (t == null) {
                        return null;
                    }
                } else {
                    t = t.getGt();
                    if (t == null) {
                        return null;
                    }
                }
            }
            return t.getFiles();
        } catch (IOException ex) {
            throw new OrionDataModelException(ex);
        }
    }

    /**
     * A tag within the file.
     */
    private class Tag {

        private final String tag;
        private final String tagUC;
        private final long lt;
        private final long gt;
        private final long pos;

        /**
         *
         * Constructor.
         * <p>
         * Reads the data from the file</p>
         *
         * @throws IOException if it all goes blurred
         */
        public Tag() throws IOException {
            lt = (long) raf.readInt();
            gt = (long) raf.readInt();
            final int tagLen = ((int) raf.readByte()) & BYTE_OF_FF;
            final byte[] bytes = new byte[tagLen];
            raf.read(bytes);
            tag = new String(bytes);
            pos = raf.getFilePointer();
            tagUC = tag.toUpperCase();
        }

        /**
         * Get the raw tag string.
         *
         * @return The string
         */
        public String getTag() {
            return tag;
        }

        /**
         * Get the tag string in upper case.
         *
         * @return the tag
         */
        public String getTagUC() {
            return tagUC;
        }

        /**
         * Get the files for the tag.
         *
         * @return the file entries
         * @throws IOException if necessary
         */
        public List<OrionFileEntry> getFiles() throws IOException {
            final List<OrionFileEntry> list = new ArrayList<>();
            raf.seek(pos);
            long l;
            while ((l = (long) raf.readInt()) != -1) {
                final long ll = raf.getFilePointer();
                raf.seek(l);
                final String fileName = raf.readUTF();
                raf.seek(ll);
                final int hitCount;
                if (version == -1) {
                    hitCount = -1;
                } else {
                    hitCount = raf.readInt();
                }
                list.add(new OrionFileEntry(fileName, hitCount));
            }
            return list;
        }

        /**
         * Get the next tag less than this one.
         *
         * @return the next tag
         * @throws IOException if necessary
         */
        public Tag getLt() throws IOException {
            if (lt >= 0L) {
                raf.seek(lt);
                return new Tag();
            }
            return null;
        }

        /**
         * Get the next tag greater than this one.
         *
         * @return the next tag
         * @throws IOException if necessary
         */
        public Tag getGt() throws IOException {
            if (gt >= 0L) {
                raf.seek(gt);
                return new Tag();
            }
            return null;
        }
    }
}
