/*
 * -----------------------------------------------------------------------
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU Library General Public License as published
 *    by the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 * 
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 * 
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ----------------------------------------------------------------------
 */
package uk.co.ruelgnoj.orion.dm;

import java.util.List;

/**
 * This interface is the source of data for an Orion component.
 * <p>It provides access to orion search capabilities, and to allow the
 * obtaining of file data.
 * @author Steve Millington
 * @version $Id: OrionDataModel.java 206 2008-01-14 18:20:55Z steve $
 */
public interface OrionDataModel {
    /**
     * Get all the tags that match the specified pattern.
     *
     * @param pattern The pattern to match.
     *
     * @return A list containing all the matching tags. If there are no
     * matches then an empty list is returned.
     *
     * @exception OrionDataModelException If any comms fails.
     */
    List<String> getTags(String pattern) throws OrionDataModelException;

    /**
     * Get the files associated with a given tag.
     *
     * @param tagString The tag string
     *
     * @return A list of the file entries of the files that contain the
     * supplied tag.
     * An empty list is returned if there is no match.
     *
     * @exception OrionDataModelException If any comms fails.
     */
    List<OrionFileEntry> getFiles(String tagString) throws OrionDataModelException;

    /**
     * Get the contents of a file as a character array.
     *
     * @param fileName The name of the file to get. The format of the name is
     * implementation dependant, but is normally based on some 'root'
     * directory.
     *
     * @return The contents of the file.
     *
     * @exception OrionDataModelException If any comms fails.
     */
    CharSequence getFile(String fileName) throws OrionDataModelException;

    /**
     * Get the contents of a file as a character array.
     *
     * @param fileEntry The the file to get.
     *
     * @return The contents of the file.
     *
     * @exception OrionDataModelException If any comms fails.
     */
    CharSequence getFile(OrionFileEntry fileEntry) throws OrionDataModelException;

    /**
     * Destroy the model.
     * <p>This method gives individual models the chance to perform any
     * termination processing they require. For example, the implementation
     * may close any open files.
     */
    void destroy();
}
