/*
orion: Provides a set of source code browsing utilities.
Copyright (C) 2008  Steve Millington

$Id: ProjectInfo.java 271 2009-04-15 21:53:25Z steve $

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.servlet;

import uk.co.ruelgnoj.orion.defdm.impl.OrionRequestHandlerLocal;
import com.thoughtworks.xstream.annotations.XStreamAlias;
import com.thoughtworks.xstream.annotations.XStreamAsAttribute;
import java.io.FileNotFoundException;
import java.sql.SQLException;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import uk.co.ruelgnoj.orion.dm.OrionDataModel;
import uk.co.ruelgnoj.orion.dm.OrionDataModelException;
import uk.co.ruelgnoj.orion.defdm.impl.OrionRequestHandler;
import uk.co.ruelgnoj.orion.project.Category;
import uk.co.ruelgnoj.orion.project.Project;
import uk.co.ruelgnoj.orion.project.ProjectFactory;
import uk.co.ruelgnoj.orion.servlet.util.Strings;

/**
 * Container for information about a project.
 */
@XStreamAlias("project")
public class ProjectInfo implements Comparable<ProjectInfo> {

    private transient List<Project> _project;
    private transient SourceCodeFormatter _sourceFormatter;
    private transient OrionDataModel _dataModel;
    private transient OrionRequestHandler _requestHandler;
    @XStreamAsAttribute
    private String name;
    private String description;
    @XStreamAlias("database")
    private String _url;
    private static final Comparator<String> STRING_COMPARATOR = Strings.getNaturalComparator();

    /**
     * Constructor.
     */
    public ProjectInfo() {
        /*
        _url = "RDB";
        name = "N";
        commentary = "COMM";
        title = "TIT";
        description = "DES";
        roles = new HashSet<Role>();
        roles.add(new Role("fred"));
        roles.add(new Role("bill"));
         */
    }

    public void setRawDatabaseName(String rawDatabaseName) {
        _url = rawDatabaseName;
        _project = null;
        _dataModel = null;
        _requestHandler = null;
    }

    public SourceCodeFormatter getSourceFormatter() {
        if (_sourceFormatter == null) {
            _sourceFormatter = new SourceCodeFormatter();
        }
        return _sourceFormatter;
    }

    public OrionDataModel getDataModel() {
        initProject();
        return _dataModel;
    }

    public OrionRequestHandler getRequestHandler() {
        initProject();
        return _requestHandler;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public void setDescription(String description) {
        this.description = description;
    }

    public void setUrl(String _url) {
        this._url = _url;
    }

    public String getDescription() {
        return description;
    }

    public String getDatabase() {
        return ServletUtils.expand(_url);
    }

    public boolean isWellConfigured() {
        return (getRequestHandler() != null) && (getDataModel() != null) && (getSourceFormatter() != null);
    }

    @Override
    public boolean equals(Object o) {
        if (!(o instanceof ProjectInfo)) {
            return false;
        }

        final ProjectInfo that = (ProjectInfo) o;
        return this.getName().equals(that.getName());
    }

    @Override
    public int hashCode() {
        return getName().hashCode();
    }

    @Override
    public int compareTo(ProjectInfo that) {
        return STRING_COMPARATOR.compare(this.getName(), that.getName());
    }

    private void initProject() {
        if (_project == null) {
            final String urlString = ServletUtils.expand(_url);
            final List<Category> cats = ProjectFactory.getCategories(urlString);
            if (cats.isEmpty()) {
                _project = Collections.emptyList();
            } else {
                _project = ProjectFactory.getCategories(urlString).get(0).getProjects();
            }
        }
        if (_dataModel == null && !_project.isEmpty()) {
            try {
                _dataModel = _project.get(0).createDataModel();
            } catch (OrionDataModelException e) {
                if (e.getCause() instanceof SQLException
                        || e.getCause() instanceof FileNotFoundException) {
                    System.err.println("Could not load datamodel for " + _project.get(0).getName() + ": " + e.getCause().getMessage());
                } else {
                    e.printStackTrace();
                }
            }
        }
        if (_requestHandler == null && _dataModel != null) {
            _requestHandler = new OrionRequestHandlerLocal(_dataModel);
        }
    }
}
