/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: ProjectManager.java 271 2009-04-15 21:53:25Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.servlet;

import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.io.xml.DomDriver;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.util.ArrayList;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.stream.Collectors;

/**
 * Container for information about a project.
 *
 * @author Steve Millington
 * @version $Id: ProjectManager.java 271 2009-04-15 21:53:25Z steve $
 */
public class ProjectManager {

    private final Map<String, ProjectInfo> projectsMap;
    private final List<CategoryInfo> categories;
    private Set<ProjectInfo> projects;
    private Throwable error;

    public ProjectManager(File projectFile) {
        projectsMap = new HashMap<>();
        categories = new ArrayList<>();
        projects = Collections.emptySet();
        try (FileInputStream fis = new FileInputStream(projectFile)) {
            initialize(fis);
            error = null;
        } catch (IOException e) {
            error = e;
        }
    }

    /**
     * Constructor for testing.
     */
    ProjectManager(InputStream is) throws IOException {
        error = null;
        projectsMap = new HashMap<>();
        categories = new ArrayList<>();
        projects = Collections.emptySet();
        initialize(is);
    }

    private void initialize(InputStream is) throws IOException {
        final XStream xStream = new XStream(new DomDriver());
        // I would like to do this with annotations, but I dont know how!!
        xStream.processAnnotations(OrionProjectConfiguration.class);
        xStream.processAnnotations(ProjectInfo.class);
        final OrionProjectConfiguration opc = (OrionProjectConfiguration) xStream.fromXML(is);
        for (CategoryInfo category : opc.getCategories()) {
            categories.add(category);
            for (ProjectInfo pi : category.getProjects()) {
                projectsMap.put(pi.getName(), pi);
            }
        }

        final Set<ProjectInfo> pis = new TreeSet<>();
        final Set<Map.Entry<String, ProjectInfo>> ss = projectsMap.entrySet();
        for (Map.Entry<String, ProjectInfo> ent : ss) {
            pis.add(ent.getValue());
        }
        projects = Collections.unmodifiableSet(pis);

    }

    public ProjectInfo getProject(String projectName) {
        return projectsMap.get(projectName);
    }

    public Set<ProjectInfo> getProjects() {
        return projects;
    }

    public Set<ProjectInfo> getWellConfiguredProjects() {
        return projects.stream().filter((p) -> p.isWellConfigured()).collect(Collectors.toSet());
    }

    public List<CategoryInfo> getCategories() {
        return categories;
    }

    public boolean hasError() {
        return error != null;
    }

    public String getErrorAsHTML() {
        if (error == null) {
            return "";
        } else {
            final StringBuilder sb = new StringBuilder();
            sb.append("<div class=\"error\">");
            sb.append("<p>An error occured during initialization:</p>");
            ServletUtils.htmlFor(sb, error);
            sb.append("</div>");
            return sb.toString();
        }
    }
}
