/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: ServletUtils.java 271 2009-04-15 21:53:25Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uk.co.ruelgnoj.orion.servlet;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.StringTokenizer;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

/**
 * Servlet Utilities.
 *
 * @author Steve Millington
 * @version $Id: ServletUtils.java 271 2009-04-15 21:53:25Z steve $
 */
public final class ServletUtils {

    private static final String ORION_COOKIE_NAME = "Orion";
    private static final int COOKIE_LIFETIME = 60 * 60 * 24 * 365 * 3;
    private static final Map<Object, Object> SYSPROP_MAP = new LazySystemPropertiesMap();

    /**
     * Constructor.
     */
    private ServletUtils() {
    }

    /**
     * Get the locale for the request. This comes from the 'Accept-Language'
     * header.
     *
     * @param req The request
     * @return the locale
     */
    public static Locale getPrimaryLocale(HttpServletRequest req) {
        final String acceptLanguage = req.getHeader("Accept-Language");
        if (acceptLanguage == null) {
            return Locale.getDefault();
        }
        final StringTokenizer st = new StringTokenizer(acceptLanguage, ",");
        if (st.hasMoreTokens()) {
            String lang = st.nextToken();
            final int semi = lang.indexOf(';');
            if (semi != -1) {
                lang = lang.substring(0, semi);
            }
            lang = lang.trim();
            final Locale ret;
            final int dash = lang.indexOf('-');
            if (dash != -1) {
                ret = new Locale(lang.substring(0, dash),
                        lang.substring(dash + 1));
            } else {
                ret = new Locale(lang, "");
            }
            return ret;
        }
        return Locale.getDefault();
    }

    /**
     * Get the value of the Orion cookie. If the cookie does not exists, make
     * sure that one is returned on the response to this request.
     *
     * @param req The request
     * @param res The response
     * @return The value
     */
    public static String getCookieValue(HttpServletRequest req, HttpServletResponse res) {
        final Cookie[] cookies = req.getCookies();
        Cookie cookie = null;
        if (cookies != null) {
            for (Cookie cookie1 : cookies) {
                if (cookie1.getName().equals(ORION_COOKIE_NAME)) {
                    cookie = cookie1;
                    break;
                }
            }
        }

        if (cookie == null) {
            final String s = req.getRemoteAddr() + System.currentTimeMillis();
            final StringBuilder sb = new StringBuilder();
            for (int i = 0; i < s.length(); i++) {
                final char c = s.charAt(i);
                switch (c) {
                    case '.':
                        break;
                    default:
                        sb.append(c);
                        break;
                }
            }
            cookie = new Cookie(ORION_COOKIE_NAME, sb.toString());
            cookie.setMaxAge(COOKIE_LIFETIME);
            //
            // Ensure that the cookie is returned for all servlet requests
            // within this web application.
            //
            String path = req.getContextPath();
            if (path.length() == 0) {
                path = "/";
            }
            cookie.setPath(path);

            res.addCookie(cookie);
        }

        return cookie.getValue();
    }

    /**
     * Make the characters in a string safe to send in HTML. This means ensuring
     * there are no less thans or ampersands.
     *
     * @param s The string
     * @return A string of the answer.
     */
    public static String htmlSafe(CharSequence s) {
        return htmlSafe(s, 0, s.length());
    }

    /**
     * Make the characters in a buffer safe to send in HTML. This means ensuring
     * there are no less thans or ampersands.
     *
     * @param buffer the buffer
     * @param so The start offset
     * @param len The length
     * @return A string of the answer.
     */
    public static String htmlSafe(CharSequence buffer, int so, int len) {
        final StringBuilder sb = new StringBuilder();
        for (int i = 0; i < len; i++) {
            final char c = buffer.charAt(so + i);
            switch (c) {
                case '"':
                    sb.append("&quot;");
                    break;
                case '&':
                    sb.append("&amp;");
                    break;
                case '<':
                    sb.append("&lt;");
                    break;
                case '>':
                    sb.append("&gt;");
                    break;
                case '\n':
                    break;
                case '\r':
                    break;
                default:
                    sb.append(c);
                    break;
            }
        }
        return sb.toString();
    }

    public static String expand(String src) {
        return expand(SYSPROP_MAP, src);
    }

    public static String expand(Map<Object, Object> values, String src) {
        return expand(values, Collections.emptySet(), src);
    }

    /**
     * Expand values in a string based on values in a map.
     *
     * @param values the map
     * @param src The source string
     * @return The expanded string
     */
    private static String expand(Map<Object, Object> values, Set<Object> exceptions, String src) {
        if (src.indexOf('$') < 0) {
            return src;
        }

        final StringBuilder ret = new StringBuilder();
        while (src.length() > 0) {
            final int start = src.indexOf('$');
            if (start < 0) {
                ret.append(src);
                src = "";
            } else {
                ret.append(src.substring(0, start));
                src = src.substring(start);
                if (src.indexOf('{') != 1) {
                    ret.append(src.charAt(0));
                    src = src.substring(1);
                } else {
                    final int end = src.indexOf('}');
                    if (end < 0) {
                        ret.append(src.charAt(0));
                        src = src.substring(1);
                    } else {
                        final String key = src.substring(2, end);
                        final String expanded;
                        if (exceptions.contains(key)) {
                            expanded = "${" + key + '}';
                        } else {
                            final String val = (String) values.get(key);
                            if (val == null) {
                                expanded = "${" + key + '}';
                            } else {
                                final Set<Object> newExceptions;
                                if (exceptions.isEmpty()) {
                                    newExceptions = Collections.singleton((Object) key);
                                } else {
                                    newExceptions = new HashSet<>(exceptions);
                                    newExceptions.add(key);
                                }

                                expanded = expand(values, newExceptions, val);
                            }
                        }
                        ret.append(expanded);
                        src = src.substring(end + 1);
                    }
                }
            }
        }
        return ret.toString();
    }

    public static void htmlFor(StringBuilder sb, Throwable t) {
        sb.append("<pre>\n");
        sb.append(htmlSafe(t.getClass().getName()));
        if (t.getMessage() != null) {
            sb.append(": ");
            sb.append(t.getMessage());
        }
        sb.append('\n');
        final StackTraceElement[] stes = t.getStackTrace();
        for (StackTraceElement ste : stes) {
            sb.append("    at ");
            sb.append(htmlSafe(ste.getClassName()));
            sb.append('.');
            sb.append(htmlSafe(ste.getMethodName()));
            sb.append('(');
            if (ste.isNativeMethod()) {
                sb.append("Native Method");
            } else if (ste.getFileName() == null) {
                sb.append("Unknown Source");
            } else {
                sb.append(htmlSafe(ste.getFileName()));
                sb.append(':');
                sb.append(Integer.toString(ste.getLineNumber()));
            }
            sb.append(")\n");
        }

        if (t.getCause() != null) {
            sb.append("Caused by ");
            htmlFor(t.getCause());
        }
        sb.append("\n</pre>");
    }

    public static String htmlFor(Throwable t) {
        final StringBuilder sb = new StringBuilder();
        htmlFor(sb, t);
        return sb.toString();
    }

    private static class LazySystemPropertiesMap implements Map<Object, Object> {

        @Override
        public int size() {
            return System.getProperties().size();
        }

        @Override
        public boolean isEmpty() {
            return System.getProperties().isEmpty();
        }

        @Override
        public boolean containsKey(Object key) {
            if (!(key instanceof String)) {
                return false;
            }
            return System.getProperty((String) key) != null;
        }

        @Override
        public boolean containsValue(Object value) {
            return System.getProperties().containsValue(value);
        }

        @Override
        public Object get(Object key) {
            if (!(key instanceof String)) {
                return null;
            }
            return System.getProperty((String) key);
        }

        @Override
        public Object put(Object key, Object value) {
            return System.getProperties().put(key, value);
        }

        @Override
        public Object remove(Object key) {
            return System.getProperties().remove(key);
        }

        @Override
        public void putAll(Map<? extends Object, ? extends Object> m) {
            System.getProperties().putAll(m);
        }

        @Override
        public void clear() {
            System.getProperties().clear();
        }

        @Override
        public Set<Object> keySet() {
            return System.getProperties().keySet();
        }

        @Override
        public Collection<Object> values() {
            return System.getProperties().values();
        }

        @Override
        public Set<Entry<Object, Object>> entrySet() {
            return System.getProperties().entrySet();
        }
    }
}
