/*
 orion: Provides a set of source code browsing utilities.
 Copyright (C) 2008  Steve Millington

 $Id: SourceCodeFormatter.java 225 2008-03-24 18:39:02Z steve $

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package uk.co.ruelgnoj.orion.servlet;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;

import java.nio.CharBuffer;

import uk.co.ruelgnoj.orion.tokenizers.LanguageToken;
import uk.co.ruelgnoj.orion.tokenizers.LanguageTokenizer;
import uk.co.ruelgnoj.orion.tokenizers.TokenTypes;
import uk.co.ruelgnoj.orion.tokenizers.TokenizerFactory;

/**
 * This is a basic HTML formatting utility. It is unfinished, but
 * the basic functionality is there.
 *
 * Enhancements to the parser APIs would make this a bit simpler!
 * @author Steve Millington
 * @version $Id: SourceCodeFormatter.java 225 2008-03-24 18:39:02Z steve $
 */
public class SourceCodeFormatter {
    private static final TokenizerFactory TF = new TokenizerFactory();

    // Keep same as FileView.java
    private static final String LINE_NUMBER_PREFIX = "L";

    private boolean lineNumbering;

    /**
     * Constructor.
     */
    public SourceCodeFormatter() {
        lineNumbering = true;
    }

    public void setLineNumbering(boolean b) {
        lineNumbering = b;
    }

    public boolean getLineNumbering() {
        return lineNumbering;
    }

    /**
     * Format a source file to the writer.
     * @param sourceFile The file to format
     * @param pw The destination for the HTML
     * @throws IOException if something goes wrong
     */
    void format(File sourceFile, PrintWriter pw) throws IOException {
	final char[] ca;
        try (FileReader fis = new FileReader(sourceFile)) {
            ca = new char[(int)sourceFile.length()];
            fis.read(ca, 0, (int)sourceFile.length());
        }

	format(sourceFile.getName(), ca, pw);
    }

    /**
     * Format a char array.
     * @param name The name of the file from which the array came
     * @param ca The char array
     * @param pw The destination
     */
    void format(String name, char[] ca, PrintWriter pw) {
	format(name, CharBuffer.wrap(ca), pw);
    }

    /**
     * Format a char sequence.
     * @param name The name of the file from which the sequence came
     * @param buffer The char sequence
     * @param pw The destination
     */
    void format(String name, CharSequence buffer, PrintWriter pw) {
	int linenum = 1;		// Start Line numbers at 1
	final LanguageTokenizer lt = TF.getTokenizer(name);

	pw.println("<pre>");
	pw.print("<a id=\"" + LINE_NUMBER_PREFIX + linenum + "\">");
        if (lineNumbering) {
            pw.print(leftPad(linenum, '0', 5));
            pw.print(' ');
	}
	pw.print("</a>");
        linenum++;
	lt.setBuffer(buffer);
	for (LanguageToken tok : lt) {
	    if (tok.getType() == TokenTypes.TT_NEWLINE) {
		pw.println();
		pw.print("<a id=\"" + LINE_NUMBER_PREFIX + linenum + "\">");
                if (lineNumbering) {
                    pw.print(leftPad(linenum, '0', 5));
                    pw.print(' ');
		}
		// Put line numbers in at some point
		pw.print("</a>");
		linenum++;
	    } else if (tok.getType() == TokenTypes.TT_WHITESPACE) {
		pw.print(htmlSafe(buffer, tok));
	    } else {
		final String text = htmlSafe(buffer, tok);
		if (text.length() > 0) {
		    pw.print("<span class=\"");
		    pw.print(tok.getType().getName());
		    pw.print("\">");
		    pw.print(text);
		    pw.print("</span>");
		}
	    }
	}
	pw.println("</pre>");
    }

    /**
     * Pad a number.
     * @param number The number to pad
     * @param padChar the character to pad with
     * @param width the target width of the field.
     * @return A string of the answer.
     */
    private static String leftPad(int number, char padChar, int width) {
        final StringBuilder ret = new StringBuilder();
        final String numberString = Integer.toString(number);
        for (int i = 0; i < width - numberString.length(); i++) {
            ret.append(padChar);
        }
        ret.append(numberString);
        return ret.toString();
    }

    /**
     * Make the characters in a buffer safe to send in HTML. This means
     * ensuring there are no less thans or ampersands.
     * @param buffer the buffer
     * @param tok The token to obtain the length details of
     * @return A string of the answer.
     */
    private static String htmlSafe(CharSequence buffer, LanguageToken tok) {
	return ServletUtils.htmlSafe(buffer, tok.getStartOffset(), tok.getLength());
    }
}
