////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: FileIterator.java 285 2011-07-19 12:37:42Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tags;

import java.io.File;
import java.io.FileFilter;

import java.util.Arrays;
import java.util.Iterator;
import java.util.Stack;

/**
 * An iterator through a directory of files.
 * <p>The iterator order is <b>not</b> defined by this class and should not be
 * relied upon.
 *
 * @author Steve Millington
 * @version $Id: FileIterator.java 285 2011-07-19 12:37:42Z steve $
 */
public class FileIterator implements Iterator<File> {

    private static final FileFilter NO_DIRS = new Dirs(false);
    private static final FileFilter DIRS = new Dirs(true);

    private Stack<StackEntry> stack;
    private File[] current;
    private int index;
    private FileFilter filter;
    private FileFilter fFilter;
    private FileFilter dFilter;

    /**
     * Iterate the files in the supplied directory.
     * All files and subdirectories are produced.
     * @param topLevelDirectory The directory to iterate over.
     */
    public FileIterator(File topLevelDirectory) {
	this(topLevelDirectory, (f)->true);
    }

    /**
     * Iterate the files in the supplied directory.
     * @param topLevelDirectory The directory to iterate over.
     * @param filter The filter defining which files are produced.
     */
    public FileIterator(File topLevelDirectory,
	    		FileFilter filter) {
	fFilter = new MuxFilter(NO_DIRS, filter);
	dFilter = new MuxFilter(DIRS, filter);
	stack = new Stack<>();
	push(topLevelDirectory);
    }

    /**
     * Push a directory onto the stack.
     * @param dir The directory
     */
    private void push(File dir) {
	current = dir.listFiles(fFilter);
	Arrays.sort(current);
	index = 0;
	final StackEntry se = new StackEntry(dir);
	if (se.files.length != 0) {
	    stack.push(se);
	}
	if (current.length == 0) {
	    pop();
	}
    }

    /**
     * Pop a directory from the stack.
     */
    private void pop() {
	if (stack.isEmpty()) {
	    current = null;
	    index = 0;
	} else {
	    final StackEntry se = stack.peek();
	    final File f = se.files[se.index++];
	    if (se.index == se.files.length) {
		stack.pop();
	    }
	    push(f);
	}
    }

    @Override
    public boolean hasNext() {
	return current != null;
    }

    @Override
    public File next() {
	final File next = current[index];
	index++;
	if (index == current.length) {
	    pop();
	}
	return next;
    }

    /**
     * {@inheritDoc}
     * @throws java.lang.UnsupportedOperationException always
     */
    @Override
    public void remove() {
	throw new UnsupportedOperationException();
    }

    /**
     * An entry on the stack.
     */
    private class StackEntry {
	private File[] files;
	private int index;
	/**
	 * Constructor.
	 * @param f The file
	 */
	public StackEntry(File f) {
	    files = f.listFiles(dFilter);
	    Arrays.sort(files);
	    index = 0;
	}
    }

    /**
     * A filter only returning directories.
     */
    private static class Dirs implements FileFilter {
	private final boolean b;
	/**
	 * Constructor.
	 * @param b True to only accept directories, false to never accept
	 * directories.
	 */
	public Dirs(boolean b) {
	    this.b = b;
	}

        @Override
	public boolean accept(File f) {
	    return f.isDirectory() == b;
	}
    }
}
