////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: FileTagDatabase.java 285 2011-07-19 12:37:42Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tags;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.RandomAccessFile;

/**
 * A tag database getting its information from a file.
 * @author Steve Millington
 * @version $Id: FileTagDatabase.java 285 2011-07-19 12:37:42Z steve $
 */
public class FileTagDatabase implements TagDatabase {

    private static final int FILE_TYPE_ID = (int) 0xabcdef01L;
    private static final int FILE_VERSION = 0x1;
    private final Tree tags_;
    private final RandomAccessFile raf_;
    private final String baseDirectory_;
    private final String dbName_;
    private int databaseOffset_;
    private int listsOfFilesOffset_;
    private Integer anInt_;

    /**
     * Constructor.
     * @param srcDir The location of the source files.
     * @param dbName The name of the destination database
     * @throws TagDatabaseException if an error occurs.
     */
    public FileTagDatabase(String srcDir, String dbName) throws TagDatabaseException {
        try {
            baseDirectory_ = srcDir;
            dbName_ = dbName;
            tags_ = new Tree();
            raf_ = new RandomAccessFile(dbName, "rw");
            writeHeader();
        } catch (IOException e) {
            throw new TagDatabaseException(e);
        }
    }

    @Override
    public void add(String newTag) {
        TagTreeObject tto = (TagTreeObject) tags_.get(newTag);
        if (tto == null) {
            tto = new TagTreeObject(newTag);
            tags_.put(tto);
            tto.addFile(anInt_);
        } else {
            tto.addFile(anInt_);
        }
    }

    @Override
    public void setFile(String fn) throws TagDatabaseException {
        try {
            anInt_ = (int) raf_.getFilePointer();
            raf_.writeUTF(fn.substring(baseDirectory_.length()));
        } catch (IOException e) {
            throw new TagDatabaseException(e);
        }
    }

    /**
     * Write the header.
     * @throws IOException if one occurs.
     */
    private void writeHeader() throws IOException {
        raf_.writeInt(FILE_TYPE_ID);	// File Type ID
        raf_.writeInt(FILE_VERSION);	// File version
        raf_.writeUTF(baseDirectory_);
        raf_.writeInt(listsOfFilesOffset_);
        raf_.writeInt(databaseOffset_);
    }

    @Override
    public void commit() throws TagDatabaseException {
        try {
            listsOfFilesOffset_ = (int) raf_.getFilePointer();

            databaseOffset_ = (int) raf_.getFilePointer();

            tags_.dump1(raf_);

            raf_.seek(databaseOffset_);

            tags_.dump2(raf_);

            raf_.seek(0);
            writeHeader();

            raf_.close();
        } catch (IOException e) {
            throw new TagDatabaseException(e);
        }
    }

    @Override
    public void printStats(PrintWriter pw) {
    }
}
