/*
 * -----------------------------------------------------------------------
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU Library General Public License as published
 *    by the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 * 
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 * 
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ----------------------------------------------------------------------
 */
package uk.co.ruelgnoj.orion.tags;

import java.io.PrintWriter;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

/**
 *
 * @author Steve Millington
 * @version $Id: JDBCTagDatabase.java 285 2011-07-19 12:37:42Z steve $
 */
public class JDBCTagDatabase implements TagDatabase {

    private static final String PROJECTS_TABLE = "PROJECTS";
    private final String url_;
    private final String sourceDirectory_;
    private String currentFile_;
    private final Map<String, TagEntry> tags_;
    private final Connection conn_;
    private final PreparedStatement afps_;
    private final PreparedStatement arps_;
    private boolean firstGo = true;
    private int fileIndex;
    private final String projectName_;
    private String tagTable_;
    private String fileTable_;

    public JDBCTagDatabase(String sourceDirectory, String url, boolean force) throws TagDatabaseException {
        sourceDirectory_ = sourceDirectory;
        // The project name is the last part of the URL, so we need to split things
        // up
        final int slashIndex = url.lastIndexOf('/');
        if (slashIndex < 0) {
            throw new TagDatabaseException("URL Format error");
        }
        final int semiIndex = url.indexOf(';');
        if (semiIndex >= slashIndex) {
            projectName_ = url.substring(slashIndex + 1, semiIndex);
            url = url.substring(0, slashIndex) + url.substring(semiIndex);
         } else {
            projectName_ = url.substring(slashIndex);
            url = url.substring(0, slashIndex);
        }
        try {
            conn_ = DriverManager.getConnection(url);
            conn_.setAutoCommit(false);

            ensureProjectsTableExists();

            createProjectEntry(force);

            createTables();

            afps_ = conn_.prepareStatement("INSERT INTO " + fileTable_ + " VALUES(?,?)");
            arps_ = conn_.prepareStatement("INSERT INTO " + tagTable_ + " VALUES(?,?,?,?)");

            url_ = url;
            currentFile_ = null;
            tags_ = new HashMap<>();
        } catch (SQLException e) {
            throw new TagDatabaseException(e);
        }
    }

    private static boolean isEmpty(ResultSet rs) throws SQLException {
        return !rs.next();
    }

    private String quote(String s) {
        final StringBuilder sb = new StringBuilder(s.length() + 2);
        sb.append('\'').append(s).append('\'');
        return sb.toString();
    }

    private void createProjectEntry(boolean force) throws SQLException {
        final String s = Long.toHexString(System.currentTimeMillis());
        tagTable_ = projectName_ + "TT" + s;
        fileTable_ = projectName_ + "FT" + s;
        ResultSet rs = null;
        Statement stmt = null;
        if (force) {
            try {
                final String text = "DELETE FROM " + PROJECTS_TABLE + " WHERE NAME = " + quote(projectName_);
                stmt = conn_.createStatement();
                stmt.executeUpdate(text);
                conn_.commit();
            } finally {
                if (rs != null) {
                    rs.close();
                }
                if (stmt != null) {
                    stmt.close();
                }
            }
        }
        try {
            final String text = "INSERT INTO " + PROJECTS_TABLE + " VALUES (" + quote(projectName_) + "," + quote(tagTable_) + "," + quote(fileTable_) + "," + quote(sourceDirectory_) + ")";
            stmt = conn_.createStatement();
            stmt.executeUpdate(text);
            conn_.commit();
        } finally {
            if (rs != null) {
                rs.close();
            }
            if (stmt != null) {
                stmt.close();
            }
        }
    }

    private void ensureProjectsTableExists() throws SQLException {
        ResultSet rs = null;
        Statement stmt = null;
        try {
            final DatabaseMetaData md = conn_.getMetaData();
            rs = md.getTables(null, null, "PROJECTS", null);
            if (isEmpty(rs)) {
                stmt = conn_.createStatement();
                stmt.execute("CREATE TABLE PROJECTS (NAME VARCHAR(100) NOT NULL PRIMARY KEY, TAGTABLE VARCHAR(100) NOT NULL, FILETABLE VARCHAR(100) NOT NULL, BASEDIR VARCHAR(500) NOT NULL)");
                conn_.commit();
            }
        } finally {
            if (rs != null) {
                rs.close();
            }
            if (stmt != null) {
                stmt.close();
            }
        }
    }

    private void createTables() throws SQLException {
        ResultSet rs = null;
        Statement stmt = null;
        try {
            stmt = conn_.createStatement();
            stmt.executeUpdate("CREATE TABLE " + tagTable_ + " (TAG VARCHAR(100) NOT NULL, TAGU VARCHAR(100) NOT NULL, FILEINDEX NUMERIC(5) NOT NULL, HITS NUMERIC(5) NOT NULL)");
            stmt.executeUpdate("CREATE INDEX " + tagTable_ + "TAGIDX ON " + tagTable_ + "(TAG)");
            stmt.executeUpdate("CREATE INDEX " + tagTable_ + "TAGUIDX ON " + tagTable_ + "(TAGU)");
            stmt.executeUpdate("CREATE TABLE " + fileTable_ + " (FILE VARCHAR(1000) NOT NULL PRIMARY KEY, INDEX NUMERIC(5) NOT NULL UNIQUE)");
            stmt.executeUpdate("CREATE INDEX " + fileTable_ + "INDEXIDX ON " + fileTable_ + "(INDEX)");
            conn_.commit();
        } finally {
            if (rs != null) {
                rs.close();
            }
            if (stmt != null) {
                stmt.close();
            }
        }
    }

    private Connection getConnection() {
        return conn_;
    }

    private PreparedStatement getAddRecordStatement() {
        return arps_;
    }

    private PreparedStatement getAddFileStatement() {
        return afps_;
    }

    private void emptyTables() throws SQLException {
        Statement stmt;
        stmt = conn_.createStatement();
        stmt.executeUpdate("delete from " + fileTable_);
        stmt.executeUpdate("delete from " + tagTable_);
        conn_.commit();
    }

    private void writeRecords() throws TagDatabaseException {
        if (currentFile_ != null && !tags_.isEmpty()) {
            try {
                if (firstGo) {
                    emptyTables();
                    firstGo = false;
                }

                final PreparedStatement afps = getAddFileStatement();
                afps.setString(1, currentFile_);
                afps.setInt(2, fileIndex);
                final int filesAdded = afps.executeUpdate();
                final PreparedStatement aps = getAddRecordStatement();
                for (Entry<String, TagEntry> entry : tags_.entrySet()) {
                    final TagEntry tag = entry.getValue();
                    aps.setString(1, tag.getName());
                    aps.setString(2, tag.getName().toUpperCase());
                    aps.setInt(3, fileIndex);
                    aps.setInt(4, tag.getHits());
                    final int rowsAdded = aps.executeUpdate();
                }
                tags_.clear();
            } catch (SQLException e) {
                throw new TagDatabaseException(e);
            }

        }
    }

    @Override
    public void setFile(String fn) throws TagDatabaseException {
        writeRecords();
        fileIndex++;
        if (fn.startsWith(sourceDirectory_)) {
            fn = fn.substring(sourceDirectory_.length());
        }
        currentFile_ = fn;
    }

    @Override
    public void add(String newTag) throws TagDatabaseException {
        TagEntry te = tags_.get(newTag);
        if (te == null) {
            te = new TagEntry(newTag);
            tags_.put(te.getName(), te);
        }
        te.hit();
    }

    @Override
    public void commit() throws TagDatabaseException {
        Connection conn = null;
        try {
            try {
                writeRecords();
                conn = getConnection();
                conn.commit();
                firstGo = true;
            } finally {
                if (conn != null) {
                    conn.close();
                }
                if (afps_ != null) {
                    afps_.close();
                }
                if (arps_ != null) {
                    arps_.close();
                }

            }
        } catch (SQLException e) {
            throw new TagDatabaseException(e);
        }
    }

    @Override
    public void printStats(PrintWriter pw) {
    }

    private static class TagEntry {

        private final String name_;
        private int hits_;

        public TagEntry(String name) {
            name_ = name;
            hits_ = 0;
        }

        public String getName() {
            return name_;
        }

        public int getHits() {
            return hits_;
        }

        public void hit() {
            hits_++;
        }
    }
}
