////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: Main.java 252 2008-07-11 09:02:10Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tags;

import java.io.File;
import java.io.FileFilter;
import java.io.PrintStream;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.StringTokenizer;

import java.util.regex.Pattern;
import uk.co.ruelgnoj.orion.tokenizers.impl.AbstractLanguageTokenizer;

import org.apache.commons.cli.Options;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.MissingOptionException;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionGroup;

/**
 * Command line start of orion tags.
 * @author Steve Millington
 * @version $Id: Main.java 252 2008-07-11 09:02:10Z steve $
 */
public final class Main {

    /**
     * Hide Constructor.
     */
    private Main() {
    }

    /**
     * Print a usage message.
     * @param ps The destination
     */
    private static void usage(PrintStream ps) {
        ps.println("Usage: java -jar orion-tags.jar -d <srcdir> [-f <outdbname> | -u <outdburl> [-x <dir,dir...>] [-X <pat,pat...>] [-q]");
    }

    /**
     * Entry point.
     * @param args The arguments
     */
    public static void main(String[] args) {
        //
        // Tell the tokenizers NOT to create new tokens each time, but to
        // reuse a single one. This improves performance
        //
        AbstractLanguageTokenizer.useStaticTokens();

        final Options options = new Options();
        final OptionGroup og = new OptionGroup();
        og.setRequired(true);
        og.addOption(new Option("f", true, "output file"));
        og.addOption(new Option("u", true, "database URL"));

        options.addOptionGroup(og);
        options.addOption("d", true, "source directory");
        options.addOption("x", true, "exclude <dir,dir,...>");
        options.addOption("X", true, "exclude <pat,pat,...>");
        options.addOption("q", false, "quiet");
        options.addOption("F", false, "force");

        final CommandLineParser parser = new DefaultParser();
        try {
            final CommandLine cmd = parser.parse(options, args);
            doStuff(cmd);
        } catch (MissingOptionException e) {
            System.err.println(e.getMessage());
            final HelpFormatter formatter = new HelpFormatter();
            formatter.printHelp("orion-tags", options);
        } catch (ParseException e) {
            e.printStackTrace();
            final HelpFormatter formatter = new HelpFormatter();
            formatter.printHelp("orion-tags", options);
        }

    }

    private static void doStuff(CommandLine cmd) {

        final String srcDir = cmd.getOptionValue("d");
        final String dbName = cmd.getOptionValue("f");
        final String url = cmd.getOptionValue("u");
        final boolean force = cmd.hasOption("F");

        //
        // Ensure that the parameters are OK
        //
        File tmpFile;

        tmpFile = new File(srcDir);
        if (!tmpFile.exists()) {
            System.err.println(srcDir + " does not exist.");
            usage(System.err);
            System.exit(1);
        }
        if (!tmpFile.isDirectory()) {
            System.err.println(srcDir + " is not a directory.");
            usage(System.err);
            System.exit(1);
        }

        if (dbName != null) {
            tmpFile = new File(dbName);
            if (tmpFile.exists()) {
                if (force) {
                    tmpFile.delete();
                } else {
                    System.err.println(dbName + " already exists.");
                    usage(System.err);
                    System.exit(1);
                }
            }
        }

        final OrionTags ot = new OrionTags();
        final Iterator<File> fit;

        FileFilter ff = null;

        final String tmpString = cmd.getOptionValue("x");
        if (tmpString != null) {
            final Set<File> exclusions = new HashSet<>();
            final StringTokenizer st = new StringTokenizer(tmpString, ",");
            while (st.hasMoreTokens()) {
                final File f = new File(srcDir, st.nextToken());
                exclusions.add(f);
            }
            ff = new ExclusionsFilter(exclusions);
        }
        final String tmpString1 = cmd.getOptionValue("X");
        if (tmpString1 != null) {
            final Set<Pattern> exclusions = new HashSet<>();
            final StringTokenizer st = new StringTokenizer(tmpString1, ",");
            while (st.hasMoreTokens()) {
                exclusions.add(Pattern.compile(st.nextToken()));
            }
            if (ff == null) {
                ff = new PatternExclusionsFilter(exclusions);
            } else {
                ff = new MuxFilter(ff, new PatternExclusionsFilter(exclusions));
            }
        }

        if (ff == null) {
            fit = new FileIterator(new File(srcDir));
        } else {
            fit = new FileIterator(new File(srcDir), ff);
        }


        try {
            ot.setSourceDirectoryName(srcDir);
            if (dbName != null) {
                ot.setDBFileName(dbName);
            }
            if (url != null) {
                ot.setDBURL(url);
            }
            ot.setVerbose(!cmd.hasOption("q"));
            ot.setForce(force);

            ot.generate(fit);

        } catch (TagDatabaseException e) {
            System.err.println(e);
            System.exit(1);
        }
    }
}
