/*
 * -----------------------------------------------------------------------
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU Library General Public License as published
 *    by the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 * 
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 * 
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ----------------------------------------------------------------------
 */
package uk.co.ruelgnoj.orion.tags;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;

import uk.co.ruelgnoj.orion.tokenizers.LanguageTokenizer;
import uk.co.ruelgnoj.orion.tokenizers.TokenizerFactory;

/**
 * 
 * @author steve
 * @version $Id: OrionTags.java 208 2008-01-15 00:32:50Z steve $
 */
public class OrionTags {

    private TagDatabase db;
    private Map<LanguageTokenizer, MutableInteger> stats;
    private Parser parser;
    private final TokenizerFactory tokenizerFactory;
    private String sourceDirectory;
    private String dbFileName;
    private String dbURL;
    private boolean verbose;
    private boolean silent;
    private int fileCount;
    private char[] buf = null;
    private boolean force;

    public OrionTags() {
        verbose = false;
        silent = false;
        force = false;
        fileCount = 0;
        tokenizerFactory = new TokenizerFactory();
    }

    public void setSourceDirectoryName(String sourceDirectory) {
        if (!sourceDirectory.endsWith(File.separator)) {
            sourceDirectory += File.separator;
        }
        this.sourceDirectory = sourceDirectory;
    }

    public void setSourceDirectory(File sourceDirectoryFile) {
        sourceDirectory = sourceDirectoryFile.getAbsolutePath();
        if (!sourceDirectory.endsWith(File.separator)) {
            sourceDirectory += File.separator;
        }
    }

    public void setDBFileName(String fileName) {
        this.dbFileName = fileName;
    }

    public void setDBFile(File file) {
        this.dbFileName = file.getAbsolutePath();
    }

    public void setDBURL(String url) {
        this.dbURL = url;
    }

    public void setVerbose(boolean b) {
        verbose = b;
    }

    public void setSilent(boolean b) {
        silent = b;
    }
    
    public void setForce(boolean b) {
        force = b;
    }

    public void generate(Iterator<File> it) throws TagDatabaseException {
        //
        // Collect statistics if verbose
        //
        if (!silent) {
            stats = new HashMap<>();
        }

        if (dbFileName != null) {
            db = new FileTagDatabase(sourceDirectory, dbFileName);
        } else {
            db = new JDBCTagDatabase(sourceDirectory, dbURL, force);
        }
        parser = new Parser(db);

        while (it.hasNext()) {
            doFile(it.next());
        }

        if (verbose) {
            System.out.print("Creating tag database ...");
            System.out.flush();
        }

        db.commit();

        if (verbose) {
            System.out.println("Done.");
            System.out.println();
        }

        if (!silent) {
            // Find out the longest description!
            int maxLength = 0;
            for (LanguageTokenizer t : stats.keySet()) {
                String s = t.getDescription();
                maxLength = Math.max(s.length(), maxLength);
            }

            System.out.println("------------------------------------------");
            System.out.println("Orion Database Statistics:");
            for (LanguageTokenizer t : stats.keySet()) {
                final MutableInteger c = stats.get(t);
                final int n = maxLength - t.getDescription().length();
                for (int i = 0; i < n; i++) {
                    System.out.print(' ');
                }
                System.out.println(t.getDescription() + " files: " + c.getValue());
            }
            db.printStats(new PrintWriter(System.out));
            System.out.println("------------------------------------------");
        }
    }

    private void doFile(File f) throws TagDatabaseException {
        final String s = f.getPath();
        final LanguageTokenizer currentTokenizer = tokenizerFactory.getTokenizer(s);
        if (currentTokenizer.canContainDBEntries()) {
            try {
                if (!silent) {
                    MutableInteger count = stats.get(currentTokenizer);
                    if (count == null) {
                        count = new MutableInteger();
                        stats.put(currentTokenizer, count);
                    }
                    count.increment();
                }

                if (verbose) {
                    System.out.println("Parsing [" + (++fileCount) + "] " + s);
                }

                final int l = (int) (f.length());
                try (BufferedReader r = new BufferedReader(new FileReader(f))) {
                    if ((buf == null) || (buf.length < l)) {
                        buf = new char[l];
                    }
                    r.read(buf);
                }
                db.setFile(s);
                parser.setTokenizer(currentTokenizer);
                parser.parse(buf, l);
            } catch (IOException e) {
                throw new TagDatabaseException(e);
            }
        }
    }
}
