////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2008  Steve Millington
//
// $Id: PatternExclusionsFilter.java 66 2006-01-30 20:47:58Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tags;

import java.io.File;
import java.io.FileFilter;

import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A <code>FileFilter</code> that rejects files in an exclusions list.
 *
 * @author Steve Millington
 * @version $Id: PatternExclusionsFilter.java 66 2006-01-30 20:47:58Z steve $
 */
public final class PatternExclusionsFilter implements FileFilter {
    private final Set<Pattern> exclusions;

    /**
     * Constructor.
     * <p>Take care that modification of the exclusions set outside of this
     * object will affect its operation. This does not occur in orion, and
     * is not a problem. However, to make this a product-worthy class,
     * the supplied exclusions should be copied.
     *
     * @param exclusions The files to be excluded from accepting.
     */
    public PatternExclusionsFilter(Set<Pattern> exclusions) {
	if (exclusions == null) {
	    throw new NullPointerException("exclusions");
	}
	this.exclusions = exclusions;
    }

    /**
     * Accept the file except when it is in the exclusions.
     * @param f The file
     * @return true if the file is not in the exclusions
     */
    @Override
    public boolean accept(File f) {
        for (Pattern p: exclusions) {
            final Matcher m = p.matcher(f.getAbsolutePath());
            if (m.matches()) {
                return false;
            }
        }
        return true;
    }
}
