////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: TagTreeObject.java 285 2011-07-19 12:37:42Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tags;

import java.io.IOException;
import java.io.RandomAccessFile;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeSet;

/**
 * This is a specific implementation of a TreeObject for TAGS.
 * @author Steve Millington
 * @version $Id: TagTreeObject.java 285 2011-07-19 12:37:42Z steve $
 */
public class TagTreeObject implements TreeObject {
    private final String tag;
    private final String tagUC;
    private final Collection<Integer> files;
    private final Map<Integer, MutableInteger> hitCount;

    public TagTreeObject(String tag) {
    	this.tag = tag;
	this.tagUC = tag.toUpperCase();
	//
	// Use a tree set so that the files are placed in the set in order.
	// If we were not to do this, then we would need to sort any output
	// results. It is better to do the sorting at the time the database
	// is constructed rather than when it is accessed.
	//
	files = new TreeSet<>();
	hitCount = new HashMap<>();
    }

    @Override
    public String getName() {
    	return tag;
    }

    @Override
    public String getUCName() {
    	return tagUC;
    }

    public void addFile(Integer fileHandle) {
	files.add(fileHandle);
	MutableInteger hc = (MutableInteger)hitCount.get(fileHandle);
	if (hc == null) {
	    hc = new MutableInteger();
	    hitCount.put(fileHandle, hc);
	}
	hc.increment();
    }

    /**
     * Compare the current TreeObject with the given one.
     * This implementation is for a case-insensitive tree.
     * This is VERY inefficient at this time, but it is not that important
     * since this is only used in the building of the database, and not
     * at run-time.
     * @param o The object to compare to
     * {@inheritDoc}
     * @return 
     */
    @Override
    public int compareTo(TreeObject o) {
    	int rc;
	rc = getUCName().compareTo(o.getUCName());
	if (rc == 0) {
	    rc = getName().compareTo(o.getName());
	}
	return rc;
    }

    @Override
    public int compareTo(String s, String ucs) {
    	int rc;
	rc = getUCName().compareTo(ucs);
	if (rc == 0) {
	    rc = getName().compareTo(s);
	}
	return rc;
    }

    @Override
    public boolean equals(Object o) {
	if (!(o instanceof TreeObject)) {
	    return false;
	}
	final TreeObject that = (TreeObject)o;
    	return getName().equals(that.getName());
    }

    @Override
    public int hashCode() {
	return getName().hashCode();
    }

    @Override
    public void dump1(RandomAccessFile raf) throws IOException {
	final byte len = (byte)tag.length();

	//
	// Write the token out. This is NOT using a UTF. The reason is so that
	// the reader does NOT have to create a String object. This is important
	// for the performance in Orion itself.
	//
	raf.writeByte(len);		// NOTE: Restricted to 128 byte token
	raf.write(tag.getBytes());

	//
	// Write out the indices for the file names
	//
	for (Integer val : files) {
	    raf.writeInt(val);
	    raf.writeInt(hitCount.get(val).getValue());
	}
	raf.writeInt(-1);
    }

    @Override
    public void dump2(RandomAccessFile raf) throws IOException {
	final byte len = (byte)tag.length();

	//
	// Write the token out. This is NOT using a UTF. The reason is so that
	// the reader does NOT have to create a String object. This is important
	// for the performance in Orion itself.
	//
	raf.writeByte(len);		// NOTE: Restricted to 128 byte token
	raf.write(tag.getBytes());

	//
	// Write out the indices for the file names
	//
	for (Integer val : files) {
	    raf.writeInt(val);
	    raf.writeInt(hitCount.get(val).getValue());
	}
	raf.writeInt(-1);
    }

    /*
    public void dumpFiles(RandomAccessFile raf) throws IOException {
	for ( int i=0;i<files.size();i++ ) {
	    raf.writeInt(((Integer)files.elementAt(i)).intValue());
	}
	raf.writeInt(-1);
    }
    */
}
