////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: Tree.java 66 2006-01-30 20:47:58Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tags;

import java.io.IOException;
import java.io.RandomAccessFile;

public class Tree {
    private static final boolean RED = false;
    private static final boolean BLACK = true;

    private Entry root;
    private int elementCount;

    public Tree() {
    	root = null;
	elementCount = 0;
    }

    public void put(TreeObject to) {
    	final Entry newEntry = new Entry(to);
	if (root == null) {
	    root = newEntry;
	    elementCount++;
	    return;
	}

	Entry t = root;

	while (true) {
	    final int cmp = t.obj.compareTo(newEntry.obj);
	    if (cmp == 0) {
	    	System.out.println("NOT ADDING!");
		return;
	    } else if (cmp > 0) {
	    	if (t.left == null) {
		    elementCount++;
		    t.left = newEntry;
		    newEntry.parent = t;
		    fixAfterInsertion(t.left);
		    return;
		} else {
		    t = t.left;
		}
	    } else {
	    	if (t.right == null) {
		    elementCount++;
		    t.right = newEntry;
		    newEntry.parent = t;
		    fixAfterInsertion(t.right);
		    return;
		} else {
		    t = t.right;
		}
	    }
	}

    }

    public TreeObject get(String key) {
    	final Entry te = getEntry(key);
	if (te == null) {
	    return null;
	} else {
	    return te.getTreeObject();
	}
    }

    private Entry getEntry(String key) {
	Entry p = root;
	final String ucKey = key.toUpperCase();
	while (p != null) {
	    final int cmp = p.obj.compareTo(key, ucKey);
	    if (cmp == 0) {
	    	return p;
	    } else if (cmp > 0) {
	    	p = p.left;
	    } else {
	    	p = p.right;
	    }
	}
	return null;
    }

    /*
    public Enumeration keys() {
    	return new TreeEnumeration(root);
    }
    */

    public int getSize() {
    	return elementCount;
    }

    public void dump1(RandomAccessFile raf) throws IOException {
    	root.dump1(raf);
    }

    public void dump2(RandomAccessFile raf) throws IOException {
    	root.dump2(raf);
    }




    private class Entry {
	private TreeObject obj;
	Entry left;
	Entry right;
	Entry parent;
	boolean color;
	private long fileOffset;

    	public Entry(TreeObject obj) {
	    this.obj	= obj;
	    left	= null;
	    right	= null;
	    parent	= null;
	    color	= BLACK;
	}

	public TreeObject getTreeObject() {
	    return obj;
	}

	public long getOff() {
	    return fileOffset;
	}

	public void dump1(RandomAccessFile raf) throws IOException {
	    fileOffset = raf.getFilePointer();
	    if (left != null) {
		raf.writeInt((int)left.getOff());
	    } else {
		raf.writeInt(-1);
	    }
	    if (right != null) {
		raf.writeInt((int)right.getOff());
	    } else {
		raf.writeInt(-1);
	    }
	    obj.dump1(raf);

	    if (left != null) {
	    	left.dump1(raf);
	    }
	    if (right != null) {
	    	right.dump1(raf);
	    }
	}

	public void dump2(RandomAccessFile raf) throws IOException {
	    dump1(raf);
	}

	public Entry get(String srch, String ucsrch) {
	    int cmp = obj.compareTo(srch, ucsrch);
	    if (cmp == 0) {
	    	return this;
	    } else if (cmp > 0) {
	    	if (left != null) {
		    return left.get(srch, ucsrch);
		} else {
		    return null;
		}
	    } else {
	    	if (right != null) {
		    return right.get(srch, ucsrch);
		} else {
		    return null;
		}
	    }
	}

	public Entry zoomLess() {
	    if (left != null) {
	    	return left.zoomLess();
	    } else {
	    	return this;
	    }
	}
    }

    /*
    private class TreeEnumeration implements Enumeration {
	private Entry root;
	private Entry current;

    	public TreeEnumeration(Entry root) {
	    this.root = root;
	    current = root;
	    if ( root != null ) {
	    	current = root.zoomLess();
	    }
	}

	public Object nextElement() {
	    Entry te = current;
	    if ( current.right != null ) {
	    	current = current.right.zoomLess();
	    } else {
		Entry old = current;
		current = current.parent;
		while ( current!= null && current.right == old ) {
		    old = current;
		    current = current.parent;
		}
	    }
	    return te.obj;
	}

	public boolean hasMoreElements() {
	    return (current!=null);
	}
    }
    */

    /**
     * Balancing Operations.
     */

    private static boolean colorOf(Entry p) {
    	return (p == null ? BLACK : p.color);
    }

    private static Entry parentOf(Entry p) {
    	return (p == null ? null : p.parent);
    }

    private static void setColor(Entry p, boolean c) {
    	if (p != null) {
	    p.color = c;
	}
    }

    private static Entry leftOf(Entry p) {
    	return (p == null) ? null : p.left;
    }

    private static Entry rightOf(Entry p) {
    	return (p == null) ? null : p.right;
    }
    /** From CLR **/
    private void rotateLeft(Entry p) {
        Entry r = p.right;
        p.right = r.left;
        if (r.left != null) {
            r.left.parent = p;
	}
        r.parent = p.parent;
        if (p.parent == null) {
            root = r;
	} else if (p.parent.left == p) {
            p.parent.left = r;
	} else {
            p.parent.right = r;
	}
        r.left = p;
        p.parent = r;
    }

    /** From CLR **/
    private void rotateRight(Entry p) {
        Entry l = p.left;
        p.left = l.right;
        if (l.right != null) {
	    l.right.parent = p;
	}
        l.parent = p.parent;
        if (p.parent == null) {
            root = l;
	} else if (p.parent.right == p) {
            p.parent.right = l;
	} else {
	    p.parent.left = l;
	}
        l.right = p;
        p.parent = l;
    }
    /** From CLR **/
    private void fixAfterInsertion(Entry x) {
        x.color = RED;

        while (x != null && x != root && x.parent.color == RED) {
            if (parentOf(x) == leftOf(parentOf(parentOf(x)))) {
                Entry y = rightOf(parentOf(parentOf(x)));
                if (colorOf(y) == RED) {
                    setColor(parentOf(x), BLACK);
                    setColor(y, BLACK);
                    setColor(parentOf(parentOf(x)), RED);
                    x = parentOf(parentOf(x));
                } else {
                    if (x == rightOf(parentOf(x))) {
                        x = parentOf(x);
                        rotateLeft(x);
                    }
                    setColor(parentOf(x), BLACK);
                    setColor(parentOf(parentOf(x)), RED);
                    if (parentOf(parentOf(x)) != null) {
                        rotateRight(parentOf(parentOf(x)));
		    }
                }
            } else {
                Entry y = leftOf(parentOf(parentOf(x)));
                if (colorOf(y) == RED) {
                    setColor(parentOf(x), BLACK);
                    setColor(y, BLACK);
                    setColor(parentOf(parentOf(x)), RED);
                    x = parentOf(parentOf(x));
                } else {
                    if (x == leftOf(parentOf(x))) {
                        x = parentOf(x);
                        rotateRight(x);
                    }
                    setColor(parentOf(x), BLACK);
                    setColor(parentOf(parentOf(x)), RED);
                    if (parentOf(parentOf(x)) != null) {
                        rotateLeft(parentOf(parentOf(x)));
		    }
                }
            }
        }
        root.color = BLACK;
    }
}
