////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: LanguageToken.java 278 2009-05-17 15:31:01Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tokenizers;

/**
 * A single parsed token.
 * <p>Note that this object has no knowledge of the source text and cannot be
 * used to get at that text.
 * @author Steve Millington
 * @version $Id: LanguageToken.java 278 2009-05-17 15:31:01Z steve $
 */
public class LanguageToken implements Cloneable {

    private int so;
    private int len;
    private TokenType type;

    /**
     * This null constructor is used in rare cases.
     */
    protected LanguageToken() {
    }

    /**
     * Constructor.
     * @param so The start offset of the token.
     * @param len The length of the token.
     * @param type The type of token
     * @see uk.co.ruelgnoj.orion.tokenizers.TokenTypes
     */
    public LanguageToken(int so, int len, TokenType type) {
        this.so = so;
        this.len = len;
        this.type = type;
    }

    @Override
    public String toString() {
        return type.getName();
    }

    @Override
    public int hashCode() {
        return type.hashCode() + so + len;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (!(o instanceof LanguageToken)) {
            return false;
        }
        final LanguageToken that = (LanguageToken) o;
        return (this.so == that.so) && (this.len == that.len) && this.type.equals(that.type);
    }

    /**
     * Determine the type of token.
     * @return The type
     */
    public TokenType getType() {
        return type;
    }

    /**
     * Set the type of token.
     * @param type The new type
     */
    public void setType(TokenType type) {
        if (type == null) {
            throw new NullPointerException("type");
        }
        this.type = type;
    }

    /**
     * Get the buffer offset of the start of the token.
     * @return The offset
     */
    public int getStartOffset() {
        return so;
    }

    /**
     * Get the length of the token.
     * @return The length
     */
    public int getLength() {
        return len;
    }

    /**
     * Get the buffer offset of the start of the token.
     * @return The offset
     */
    public int getEndOffset() {
        return so + len - 1;
    }

    /**
     * Obtain a string that is this tokens portion of the supplied buffer.
     * @param buffer The buffer containing the character data
     * @return The string
     */
    public String asString(CharSequence buffer) {
        return buffer.subSequence(so, so + len).toString();
    }

    @Override
    public LanguageToken clone() {
        try {
            return (LanguageToken) super.clone();
        } catch (CloneNotSupportedException e) {
            throw new Error(e);
        }
    }

    /**
     * Set the token start offset.
     * <p>Package privacy so that only known people can change a token.
     * This is used by the static token implementation.
     * @param newOffset The new start offset
     */
    protected final void setStartOffset(int newOffset) {
        so = newOffset;
    }

    /**
     * Set the token length.
     * <p>Package privacy so that only known people can change a token.
     * This is used by the static token implementation.
     * @param newLength The new length
     */
    protected final void setLength(int newLength) {
        len = newLength;
    }
}
