////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: LanguageTokenizer.java 88 2006-12-01 22:39:47Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tokenizers;

import java.util.Iterator;
import java.util.Set;

/**
 * The definition of a tokenizer.
 * <p>A tokenizer is responsible for splitting a buffer into a number of tokens.
 * These tokens are used for database generation, and the display of the file.
 *
 * @author Steve Millington
 * @version $Id: LanguageTokenizer.java 88 2006-12-01 22:39:47Z steve $
 */
public interface LanguageTokenizer extends Iterator<LanguageToken>, Iterable<LanguageToken> {
    /**
     * Get the description of the parser.
     * This should be a short description of the try of files that it can
     * parse.
     * <p>For example, a C parser might return
     * <pre>
     * C and H
     * </pre>
     * @return The description.
     */
    String getDescription();

    /**
     * Get the language of the tokenizer.
     * This is typically a short string that is normally used for hash table
     * looup.
     * @return The language.
     */
    String getLanguage();

    /**
     * Set the buffer to be tokenized.
     * <p>This buffer supercedes any previous buffer setting and resets the
     * enumeration of this class.
     * @param buffer The characters to be parsed.
     */
    void setBuffer(CharSequence buffer);

    /**
     * Get the buffer being tokenized.
     * @return The buffer
     */
    CharSequence getBuffer();

    /**
     * Get the next token from the current buffer.
     * @return The token.
     */
//    public LanguageToken nextToken();

    /**
     * Determine if there are any more tokens.
     * @return true if there are more tokens.
     */
//    public boolean hasMoreTokens();

    /**
     * Determine if the tokenizer is capable of producing output that would
     * result in entries in the the tags database.
     * <p>This method can be used by a token generation function to determine
     * if it is worth parsing a particular file.
     * @return true if database entries could be produced.
     */
    boolean canContainDBEntries();

    /**
     * Get a set containing all of the possible toke types for this tokenizer.
     * @return A set whose elements are <code>TokenType</code> objects.
     */
    Set<TokenType> getPossibleTokenTypes();
}
