////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: TokenizerFactory.java 276 2009-04-22 07:55:37Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tokenizers;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * A factory for tokenizers.
 * <p>The tokenizer factory contains a two level mapping from file suffices
 * through tokenizer names to tokenizer instances.
 * @author Steve Millington
 * @version $Id: TokenizerFactory.java 276 2009-04-22 07:55:37Z steve $
 */
public class TokenizerFactory {

    private static TokenizerFactory sharedInstance = null;
    /** The location of the 'default' tokenizers. */
    private static final String DEFAULT_PACKAGE =
            "uk.co.ruelgnoj.orion.tokenizers.impl.";
    private Map<String, String> suffixToTokenizerName;
    private Map<String, Wrapper> tokenizerNameToTokenizer;

    /**
     * Create a tokenizer factory populated with default mappings.
     * <p>This should be rarley used. It is more normal to use the shared
     * instance.
     */
    public TokenizerFactory() {
        this(true);
    }

    /**
     * Create a tokenizer factory.
     * <p>This should be rarley used. It is more normal to use the shared
     * instance.
     * <p>If default mappings are not used, then the factory will contain
     * a single tokenizer named <code>default</code>.
     * @param useStandardMappings to have the standard set of mappings
     */
    public TokenizerFactory(boolean useStandardMappings) {
        tokenizerNameToTokenizer = new HashMap<>();
        suffixToTokenizerName = new HashMap<>();

        //
        // Always put the default tokenizer in the map
        //
        registerTokenizer("default",
                DEFAULT_PACKAGE + "UnknownLanguageTokenizer");

        if (useStandardMappings) {
            registerStandardTokenizers();
            registerStandardSuffixes();
        }
    }

    private void registerStandardTokenizers() {
        registerTokenizer("java",
                DEFAULT_PACKAGE + "JavaLanguageTokenizer");
        registerTokenizer("javascript",
                DEFAULT_PACKAGE + "JavascriptLanguageTokenizer");
        registerTokenizer("csharp",
                DEFAULT_PACKAGE + "CSharpLanguageTokenizer");
        registerTokenizer("c",
                DEFAULT_PACKAGE + "CLanguageTokenizer");
        registerTokenizer("cicsc",
                DEFAULT_PACKAGE + "CICSCLanguageTokenizer");
        registerTokenizer("lex",
                DEFAULT_PACKAGE + "LexLanguageTokenizer");
        registerTokenizer("yacc",
                DEFAULT_PACKAGE + "YaccLanguageTokenizer");
        registerTokenizer("cpp",
                DEFAULT_PACKAGE + "CppLanguageTokenizer");
        registerTokenizer("idl",
                DEFAULT_PACKAGE + "IDLLanguageTokenizer");
        registerTokenizer("cobol",
                DEFAULT_PACKAGE + "CobolLanguageTokenizer");
        registerTokenizer("cicscobol",
                DEFAULT_PACKAGE + "CICSCobolLanguageTokenizer");
        registerTokenizer("pli",
                DEFAULT_PACKAGE + "PLILanguageTokenizer");
        registerTokenizer("cicspli",
                DEFAULT_PACKAGE + "CICSPLILanguageTokenizer");
        registerTokenizer("installshield",
                DEFAULT_PACKAGE + "InstallshieldLanguageTokenizer");
        registerTokenizer("unixshell",
                DEFAULT_PACKAGE + "UnixShellLanguageTokenizer");
        registerTokenizer("properties",
                DEFAULT_PACKAGE + "PropertiesLanguageTokenizer");

    }

    private void registerStandardSuffixes() {
        registerSuffix("java", "java");
        registerSuffix("h", "c");
        registerSuffix("H", "c");
        registerSuffix("c", "c");
        registerSuffix("C", "c");
        registerSuffix("cs", "csharp");
        registerSuffix("ccs", "cicsc");
        registerSuffix("pc", "cicsc");
        registerSuffix("cp", "cicsc");
        registerSuffix("sqc", "cicsc");
        registerSuffix("l", "lex");
        registerSuffix("y", "yacc");
        registerSuffix("cpp", "cpp");
        registerSuffix("c++", "cpp");
        registerSuffix("hpp", "cpp");
        registerSuffix("h++", "cpp");
        registerSuffix("idl", "idl");
        registerSuffix("cbl", "cobol");
        registerSuffix("clt", "cicscobol");
        registerSuffix("cl2", "cicscobol");
        registerSuffix("pco", "cicscobol");
        registerSuffix("sqb", "cicscobol");
        registerSuffix("cop", "cicscobol");
        registerSuffix("cpy", "cobol");
        registerSuffix("CPY", "cobol");
        registerSuffix("COB", "cobol");
        registerSuffix("CBL", "cobol");
        registerSuffix("pli", "pli");
        registerSuffix("PLI", "pli");
        registerSuffix("inc", "pli");
        registerSuffix("INC", "pli");
        registerSuffix("plt", "cicspli");
        registerSuffix("ppl", "cicspli");
        registerSuffix("rul", "installshield");
        registerSuffix("sh", "unixshell");
        registerSuffix("properties", "properties");
        registerSuffix("js", "javascript");
    }

    /**
     * Obtain a shared instance of the factory.
     * @return The singleton instance
     */
    public static TokenizerFactory getSharedInstance() {
        if (sharedInstance == null) {
            synchronized (TokenizerFactory.class) {
                if (sharedInstance == null) {
                    final TokenizerFactory tf = new TokenizerFactory();
                    sharedInstance = tf;
                }
            }
        }
        return sharedInstance;
    }

    /**
     * Register a tokenizer.
     * <p>It is the callers responsibility to ensure that the class specified
     * can be loaded and instantiated correctly. The tokenizer must be an
     * implementation of the <code>LanguageTokenizer</code> interface and
     * define a public no-parameter constructor.
     * @param name The name of the tokenizer
     * @param tokenizerClassName The name of the tokenizer class.
     * @throws java.lang.NullPointerException if either parameter is null.
     */
    public final void registerTokenizer(String name, String tokenizerClassName) {
        if (name == null) {
            throw new NullPointerException("name");
        }
        if (tokenizerClassName == null) {
            throw new NullPointerException("tokenizerClassName");
        }
        tokenizerNameToTokenizer.put(name, new Wrapper(tokenizerClassName));
    }

    /**
     * Deregister a tokenizer.
     * <p>It a tokenizer by that name is not registered, then no error occurs.
     * @param name The name of the tokenizer to deregister.
     * @throws java.lang.NullPointerException if the name is null.
     */
    public void deregisterTokenizer(String name) {
        if (name == null) {
            throw new NullPointerException("name");
        }
        tokenizerNameToTokenizer.remove(name);
    }

    /**
     * Register a suffix to tokenizer name mapping entry.
     * @param suffix The file name suffix
     * @param tokenizerName The name of the tokenizer.
     * @throws java.lang.NullPointerException if either parameter is null.
     */
    public void registerSuffix(String suffix, String tokenizerName) {
        if (suffix == null) {
            throw new NullPointerException("suffix");
        }
        if (tokenizerName == null) {
            throw new NullPointerException("tokenizerName");
        }
        suffixToTokenizerName.put(suffix, tokenizerName);
    }

    private static String getSuffix(String fileName) {
        final int n = fileName.lastIndexOf('.');
        if (n >= 0) {
            return fileName.substring(n + 1);
        }
        return "";
    }

    /**
     * Get the tokenizer for a specified file name.
     * This retuens a tokenizer depending on the suffix of the name.
     * @param fileName The file name
     * @return a tokenizer
     */
    public LanguageTokenizer getTokenizer(String fileName) {
        final String suffix = getSuffix(fileName);
        String tokenizerName = suffixToTokenizerName.get(suffix);
        if (tokenizerName == null) {
            tokenizerName = "default";
        }
        return getTokenizerByName(tokenizerName);
    }

    /**
     * Get a tokenizer by its name.
     * A <code>null</code> name is valid, and returns the 'default'
     * tokenizer.
     * <p>A null return can be produced, but only when there is no
     * default tokenizer in place.
     * @param tokenizerName The name of the tokenizer to be obtained.
     * @return The tokenizer
     */
    public LanguageTokenizer getTokenizerByName(String tokenizerName) {
        Wrapper wrapper = tokenizerNameToTokenizer.get(tokenizerName);
        if (wrapper == null) {
            tokenizerName = "default";
            wrapper = tokenizerNameToTokenizer.get(tokenizerName);
        }
        if (wrapper != null) {
            return wrapper.getTokenizer();
        }
        return null;
    }

    /**
     * Get the names of all of the tokenizers known to this factory.
     * @return The names
     */
    public Set<String> getRegisteredTokenizerNames() {
        return tokenizerNameToTokenizer.keySet();
    }

    private static class Wrapper {

        private final String tokenizerName;
        private LanguageTokenizer tokenizer;

        public Wrapper(String tokenizerName) {
            this.tokenizerName = tokenizerName;
            this.tokenizer = null;
        }

        public LanguageTokenizer getTokenizer() {
            if (tokenizer == null) {
                try {
                    final Class c = Class.forName(tokenizerName);
                    tokenizer = (LanguageTokenizer) c.newInstance();
                } catch (ClassNotFoundException e) {
                    e.printStackTrace();
                } catch (InstantiationException e) {
                    e.printStackTrace();
                } catch (IllegalAccessException e) {
                    e.printStackTrace();
                }
            }
            return tokenizer;
        }
    }
}
