////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: AbstractLanguageTokenizer.java 226 2008-05-11 15:10:59Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tokenizers.impl;

import java.util.Iterator;
import uk.co.ruelgnoj.orion.tokenizers.LanguageToken;
import uk.co.ruelgnoj.orion.tokenizers.LanguageTokenizer;
import uk.co.ruelgnoj.orion.tokenizers.TokenType;

/**
 * This abstract class is the basis of all tokenizers in this package.
 */
public abstract class AbstractLanguageTokenizer implements LanguageTokenizer {
    protected CharSequence buffer;
    protected int pos;

    private boolean cached;
    private boolean canContainTokens;

    /**
     * Constructor.
     */
    public AbstractLanguageTokenizer() {
	cached = false;
    }

    /**
     * This static token is only used when the useStaticToken() method
     * had been called. It saves on the creation of loads of objects in the
     * tags file generation case.
     */
    private static StaticLanguageToken staticToken = null;

    /** @see uk.co.ruelgnoj.orion.tokenizers.LanguageTokenizer */
    @Override
    public void setBuffer(CharSequence buffer) {
	this.buffer = buffer;
	this.pos = 0;
    }

    /** @see uk.co.ruelgnoj.orion.tokenizers.LanguageTokenizer */
    @Override
    public CharSequence getBuffer() {
	return buffer;
    }

    /**
     * Order this class to reuse the tokens produced.
     * <p>When parsing a file for display, it is a bad idea to reuse tokens
     * since this will mess up any previously extracted tokens. In this case it
     * is best to let the garbage collector deal with this instances.
     * <p>However, for the token database generation, tokens are never used
     * after the next one is created. This method allow the caller to give this
     * hint to the tokenizer, and thus reduce the number of token objects
     * created.
     */
    public static void useStaticTokens() {
    	staticToken = new StaticLanguageToken();
    }

    /**
     * Create a token.
     * <p>If using a static token, then the same object is always returned,
     * but its instance variables are modified.
     * <p>If not, then a new token is created.
     * @param so The start offset of the token.
     * @param len The length of the token.
     * @param type The type of the token. This must be one of the
     * {@link uk.co.ruelgnoj.orion.tokenizers.TokenTypes TokenTypes}.
     * @return The token
     */
    protected LanguageToken createToken(int so, int len, TokenType type) {
    	if (staticToken == null) {
	    return new LanguageToken(so, len, type);
	} else {
	    staticToken.reInitialize(so, len, type);
	    return staticToken;
	}
    }

    /**
     * By default, all tokenizers can produce database entries.
     * @see uk.co.ruelgnoj.orion.tokenizers.LanguageTokenizer
     */
    @Override
    public boolean canContainDBEntries() {
	if (!cached) {
	    cached = true;
	    canContainTokens = false;
	    for (TokenType tt: getPossibleTokenTypes()) {
		if (tt.isIndexable()) {
		    canContainTokens = true;
		    break;
		}
	    }
	}
	return canContainTokens;
    }

    @Override
    public boolean hasNext() {
    	return pos < buffer.length();
    }

    /**
     * A class defining a static token.
     * <p>This token is designed to be mutable, and is mutated using the
     * <code>reinitialize()</code> method.
     */
    private static class StaticLanguageToken extends LanguageToken {
	public StaticLanguageToken() {
	}
	/**
	 * Mutate this token to have a new set of values.
	 * @param so The token start offset
	 * @param len The token length
	 * @param type The token type
	 */
	public void reInitialize(int so, int len, TokenType type) {
	    setStartOffset(so);
	    setLength(len);
	    setType(type);
	}
    }

    /**
     * Removing is not supported.
     */
    @Override
    public void remove() {
	throw new UnsupportedOperationException();
    }

//    public final LanguageToken nextToken() {
//	return next();
//    }

//    public final boolean hasMoreTokens() {
//	return hasNext();
//    }

    @Override
    public Iterator<LanguageToken> iterator() {
	return this;
    }
}
