////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: CICSMixIn.java 232 2008-06-04 15:31:24Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tokenizers.impl;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Queue;
import java.util.ResourceBundle;
import java.util.Set;
import uk.co.ruelgnoj.orion.tokenizers.LanguageToken;
import uk.co.ruelgnoj.orion.tokenizers.LanguageTokenizer;
import uk.co.ruelgnoj.orion.tokenizers.TokenType;
import uk.co.ruelgnoj.orion.tokenizers.TokenTypes;

/**
 * A mixing for tokenizing EXEC CICS Statements.
 * @author Steve Millington
 * @version $Id: CICSMixIn.java 232 2008-06-04 15:31:24Z steve $
 */
public class CICSMixIn implements LanguageTokenizer {

    public enum Flavour {

        C,
        COBOL,
        PLI
    }
    private static final String RBNAME =
            "uk.co.ruelgnoj.orion.tokenizers.resources";
    private LanguageTokenizer delegate;
    private Flavour flavour;
    private boolean inExecCICS;
    private boolean inBrace;
    private Queue<LanguageToken> queue;

    /**
     * Constructor.
     * @param delegate The tokenizer from which to obtain base functionality
     * @param flavour The flavour of mixin, either C or COBOL.
     * @throws java.lang.IllegalArgumentException if the flavor is not one
     * of the valid static constants defined by this class.
     * @throws java.lang.NullPointerException if the delegate is null
     */
    public CICSMixIn(LanguageTokenizer delegate,
            Flavour flavour) {
        if (flavour == null) {
            throw new IllegalArgumentException("flavour");
        }

        this.delegate = delegate;
        this.flavour = flavour;
        this.queue = new LinkedList<>();
        inExecCICS = false;
        inBrace = false;
    }

    @Override
    public void setBuffer(CharSequence buffer) {
        delegate.setBuffer(buffer);
    }

    @Override
    public CharSequence getBuffer() {
        return delegate.getBuffer();
    }

    @Override
    public boolean hasNext() {
        return (!queue.isEmpty()) || delegate.hasNext();
    }

    /** {@inheritDoc} */
    @Override
    public boolean canContainDBEntries() {
        return delegate.canContainDBEntries();
    }

    @Override
    public LanguageToken next() {
        if (!queue.isEmpty()) {
            final LanguageToken tok = queue.poll();
            if (inExecCICS && (tok.getType() == TokenTypes.TT_WORD)) {
                tok.setType(TokenTypes.TT_CICS_API);
            }
            return tok;
        }
        LanguageToken tok = delegate.next();
        if (tok == null) {
            return tok;
        }
        //
        // Attempt to highlight EXEC CICS commands.
        //
        if (inExecCICS) {
            if (inBrace) {
                if (tok.getType() == TokenTypes.TT_PUNC) {
                    final char startChar =
                            delegate.getBuffer().charAt(tok.getStartOffset());
                    if (startChar == ')') {
                        inBrace = false;
                    }
                    if (flavour == Flavour.C ||
                            flavour == Flavour.PLI) {
                        final char endChar =
                                delegate.getBuffer().charAt(tok.getEndOffset());
                        if (endChar == ';') {
                            inExecCICS = false;
                        }
                    }
                }
            } else if (tok.getType() == TokenTypes.TT_PUNC) {
                if (delegate.getBuffer().charAt(tok.getStartOffset()) == '(') {
                    inBrace = true;
                } else if (delegate.getBuffer().charAt(tok.getEndOffset()) == ';') {
                    if (flavour == Flavour.C ||
                            flavour == Flavour.PLI) {
                        inExecCICS = false;
                    }
                }
            } else if (tok.getType() == TokenTypes.TT_KEYWORD) {
                tok.setType(TokenTypes.TT_CICS_API);
            } else if (tok.getType() == TokenTypes.TT_WORD) {
                final String tokStr = tok.asString(delegate.getBuffer());
                if ((flavour == Flavour.COBOL) && "END-EXEC".equalsIgnoreCase(tokStr)) {
                    tok.setType(TokenTypes.TT_CICS_API);
                    inExecCICS = false;
                } else {
                    tok.setType(TokenTypes.TT_CICS_API);
                }
            }
        } else {
            if (tok.getType() == TokenTypes.TT_WORD) {
                String tokStr = tok.asString(delegate.getBuffer());
                if ("EXEC".equalsIgnoreCase(tokStr)) {
                    queue.offer(tok.clone());
                    tok = delegate.next();
                    while (true) {
                        queue.offer(tok.clone());
                        if (tok.getType() == TokenTypes.TT_WORD) {
                            tokStr = tok.asString(delegate.getBuffer());
                            tok = queue.poll();
                            if ("CICS".equalsIgnoreCase(tokStr)) {
                                tok.setType(TokenTypes.TT_CICS_API);
                                inExecCICS = true;
                            }
                            break;
                        } else {
                            tok = delegate.next();
                        }
                    }
                //		    tok.setType(TokenTypes.TT_CICS_API);
                //		    inExecCICS = true;
                }
            }
        }
        return tok;
    }

    /** {@inheritDoc} */
    @Override
    public String getDescription() {
        return ResourceBundle.getBundle(RBNAME).getString("lang.cicsc.desc");
    }

    /** {@inheritDoc} */
    @Override
    public String getLanguage() {
        return "cicscobol";
    }

    @Override
    public Set<TokenType> getPossibleTokenTypes() {
        final Set<TokenType> ret = new HashSet<>();
        ret.addAll(delegate.getPossibleTokenTypes());
        ret.add(TokenTypes.TT_CICS_API);
        return ret;
    }

    /**
     * Removing is not supported.
     */
    @Override
    public void remove() {
        throw new UnsupportedOperationException();
    }

    @Override
    public Iterator<LanguageToken> iterator() {
        return this;
    }
}
