////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: CLanguageTokenizer.java 226 2008-05-11 15:10:59Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tokenizers.impl;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Queue;
import uk.co.ruelgnoj.orion.tokenizers.LanguageToken;
import uk.co.ruelgnoj.orion.tokenizers.TokenTypes;

/**
 * A tokenizer for 'C' and 'H' files.
 *
 * @author Steve Millington
 * @version $Id: CLanguageTokenizer.java 226 2008-05-11 15:10:59Z steve $
 */
public class CLanguageTokenizer extends ClikeLanguageTokenizer {
    //
    // The reserved words is a static table, that is initialized by the static
    // initializer at the bottom of this file. It is returned to the superclass
    // when it asks for them.
    //
    private static final Collection<String> RESERVED_WORDS;
    private static final Collection<String> PP_RESERVED_WORDS;

    /**
     * Static initialization of the reserved words.
     */
    private static final String[] WORDS = {
        "asm",
        "auto",
        "break",
        "case",
        "char",
        "const",
        "continue",
        "default",
        "do",
        "double",
        "else",
        "enum",
        "extern",
        "float",
        "for",
        "goto",
        "if",
        "int",
        "long",
        "register",
        "return",
        "short",
        "signed",
        "sizeof",
        "static",
        "struct",
        "switch",
        "typedef",
        "union",
        "unsigned",
        "void",
        "volatile",
        "while",
        "defined" // Hack for #if defined(FISH)
    };

    private static final String[] PPWORDS = {
        "if",
        "line",
        "include",
        "define",
        "elif",
        "else",
        "ifdef",
        "ifndef",
        "error",
        "pragma",
        "undef",
        "endif"
    };

    static {
	RESERVED_WORDS = new HashSet<>();
        RESERVED_WORDS.addAll(Arrays.asList(WORDS));

	PP_RESERVED_WORDS = new HashSet<>();
        PP_RESERVED_WORDS.addAll(Arrays.asList(PPWORDS));
    }

    private Queue<LanguageToken> queue;

    /**
     * Constructor.
     *
     * @see uk.co.ruelgnoj.orion.tokenizers.impl.ClikeLanguageTokenizer
     */
    public CLanguageTokenizer() {
    }

    @Override
    public void setBuffer(CharSequence buffer) {
        super.setBuffer(buffer);
        if (queue == null) {
            queue = new LinkedList<>();
        } else {
            queue.clear();
        }
    }

    @Override
    protected Collection<String> getReservedWords() {
        return RESERVED_WORDS;
    }

    /**
     * Get the pre-processor reserved words.
     *
     * @return the words
     */
    protected Collection<String> getPreprocessorReservedWords() {
        return PP_RESERVED_WORDS;
    }

    /**
     * Is the supplied word a reserved one?
     *
     * @param word The word to test
     * @return true if it is a reserved word.
     */
    private boolean isPPReservedWord(String word) {
        return getPreprocessorReservedWords().contains(word);
    }

    //
    // This method copes with preprocessor statements
    // It has to queue elements looking for stuff after the white space in the
    // set of tokens.
    // In order to deal with the 'reuse token' functionality, used by the
    // OrionTags utility, it has to store away 'new' copies of the 
    // token for its own use.
    //
    @Override
    public LanguageToken next() {

        if (!queue.isEmpty()) {
            return queue.poll();
        }

        final LanguageToken t = super.next();

        if ((t != null)
                && (t.getType() == TokenTypes.TT_PUNC)
                && (buffer.charAt(t.getStartOffset()) == '#')) {
            return processPP(t);
        }
        return t;
    }

    /**
     * Process a pre-processor command. This acts by putting a bunch of tokens
     * onto a queue so that they are read first by any future readers.
     *
     * @param t The token that caused us to realise a PP line exists.
     * @return The first token
     */
    private LanguageToken processPP(LanguageToken t) {
        final LanguageToken hashElement = t.clone();
        queue.offer(hashElement);
        t = super.next();
        if (t.getType() == TokenTypes.TT_WHITESPACE) {
            queue.offer(t.clone());
            t = super.next();
        }
        final String tokStr = t.asString(buffer);
        if (tokStr.equals("include")) {
            queue.clear();
            queue.offer(new LanguageToken(
                    hashElement.getStartOffset(),
                    t.getStartOffset()
                    - hashElement.getStartOffset()
                    + t.getLength(),
                    TokenTypes.TT_KEYWORD));
            boolean done = false;
            int so = 0;
            while (!done) {
                t = super.next();
                if (t.getType() == TokenTypes.TT_WHITESPACE) {
                    queue.offer(t.clone());
                } else if (t.getType() == TokenTypes.TT_STRING) {
                    queue.offer(t.clone());
                    done = true;
                } else if (t.getType() == TokenTypes.TT_PUNC) {
                    if (buffer.charAt(t.getStartOffset()) == '<') {
                        so = t.getStartOffset();
                    } else if (buffer.charAt(t.getStartOffset()) == '>') {
                        queue.offer(new LanguageToken(
                                so,
                                (t.getStartOffset() - so)
                                + t.getLength(),
                                TokenTypes.TT_STRING));
                        done = true;
                    }
                }
            }
            return queue.poll();
        } else if (isPPReservedWord(tokStr)) {
            queue.clear();
            return createToken(hashElement.getStartOffset(),
                    t.getStartOffset()
                    - hashElement.getStartOffset()
                    + t.getLength(),
                    TokenTypes.TT_KEYWORD);
        } else {
            queue.offer(new LanguageToken(t.getStartOffset(),
                    t.getLength(),
                    t.getType()));
            return queue.poll();
        }
    }

    @Override
    public boolean hasNext() {
        return (!queue.isEmpty()) || super.hasNext();
    }

    @Override
    public String getDescription() {
        return "C and H";
    }

    @Override
    public String getLanguage() {
        return "c";
    }
}
