////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: ClikeLanguageTokenizer.java 226 2008-05-11 15:10:59Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tokenizers.impl;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import uk.co.ruelgnoj.orion.tokenizers.CharUtils;
import uk.co.ruelgnoj.orion.tokenizers.LanguageToken;
import uk.co.ruelgnoj.orion.tokenizers.TokenType;
import uk.co.ruelgnoj.orion.tokenizers.TokenTypes;

/**
 * A tokenizer for languages that are essentially like 'C' in their
 * construction. This class is designed to be subclasses to provide the
 * necessary information regarding reserved words, etc.
 *
 * @author Steve Millington
 * @version $Id: ClikeLanguageTokenizer.java 226 2008-05-11 15:10:59Z steve $
 */
public abstract class ClikeLanguageTokenizer extends AbstractLanguageTokenizer {

    private static final CLLTState INDBLQ = new CLLTStateInDoubleQuotes();
    private static final CLLTState INSGLQ = new CLLTStateInSingleQuotes();
    private static final CLLTState BASE = new CLLTStateBase();
    private static final CLLTState INCOMMENT = new CLLTStateInComment();
    private static final CLLTState INSCOMMENT = new CLLTStateInSingleComment();
    private static final CLLTState INPUNC = new CLLTStateInPunctuation();
    private static final CLLTState INNUM = new CLLTStateInNumber();
    private static final CLLTState INWHITE = new CLLTStateInWhitespace();
    private static final CLLTState INWORD = new CLLTStateInWord();
    private static final CLLTState INHASH = new CLLTStateInHash();

    private CLLTState state;
    private final Collection<String> resWords;

    /**
     * A tokenizer state.
     */
    private static interface CLLTState {

        /**
         * Get the next token.
         *
         * @param tokenizer The tokenizer.
         * @param thisChar The current character
         * @param sp The buffer index.
         * @return The next token
         */
        LanguageToken nextToken(ClikeLanguageTokenizer tokenizer,
                char thisChar,
                int sp);

        /**
         * Special case version of next token, for when the end of the buffer
         * has been reached.
         *
         * @param tokenizer The tokenizer.
         * @param sp The buffer index.
         * @return The next token
         */
        LanguageToken nextTokenLast(ClikeLanguageTokenizer tokenizer,
                int sp);
    }

    /**
     * State In Hash.
     */
    private static final class CLLTStateInHash implements CLLTState {

        /**
         * Constructor.
         */
        public CLLTStateInHash() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextToken(ClikeLanguageTokenizer tokenizer,
                char thisChar,
                int sp) {
            LanguageToken ret = null;
            if (!CharUtils.isHash(thisChar)) {
                tokenizer.pos--;
                ret = tokenizer.createToken(sp,
                        tokenizer.pos - sp + 1,
                        TokenTypes.TT_PUNC);
                tokenizer.state = BASE;
            }
            return ret;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextTokenLast(ClikeLanguageTokenizer tokenizer,
                int sp) {
            return null;
        }
    }

    /**
     * State In Word.
     */
    private static final class CLLTStateInWord implements CLLTState {

        /**
         * Constructor.
         */
        public CLLTStateInWord() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextToken(ClikeLanguageTokenizer tokenizer,
                char thisChar,
                int sp) {
            LanguageToken ret = null;
            if (!CharUtils.isAlNumUnderscore(thisChar)) {
                tokenizer.pos--;
                if (tokenizer.isReservedWord(
                        tokenizer.buffer.subSequence(sp,
                                tokenizer.pos + 1).toString())) {
                    ret = tokenizer.createToken(sp,
                            tokenizer.pos - sp + 1,
                            TokenTypes.TT_KEYWORD);
                } else {
                    ret = tokenizer.createToken(sp,
                            tokenizer.pos - sp + 1,
                            TokenTypes.TT_WORD);
                }
                tokenizer.state = BASE;
            }
            return ret;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextTokenLast(ClikeLanguageTokenizer tokenizer,
                int sp) {
            if (tokenizer.isReservedWord(
                    tokenizer.buffer.subSequence(sp, tokenizer.pos).toString())) {
                return tokenizer.createToken(sp,
                        tokenizer.pos - sp,
                        TokenTypes.TT_KEYWORD);
            } else {
                return tokenizer.createToken(sp,
                        tokenizer.pos - sp, TokenTypes.TT_WORD);
            }
        }
    }

    /**
     * State In Whitespace.
     */
    private static final class CLLTStateInWhitespace implements CLLTState {

        /**
         * Constructor.
         */
        public CLLTStateInWhitespace() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextToken(ClikeLanguageTokenizer tokenizer,
                char thisChar,
                int sp) {
            LanguageToken ret = null;
            if (thisChar != ' '
                    && (thisChar != '\t')) {
                tokenizer.pos--;
                ret = tokenizer.createToken(sp,
                        tokenizer.pos - sp + 1,
                        TokenTypes.TT_WHITESPACE);
                tokenizer.state = BASE;
            }
            return ret;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextTokenLast(ClikeLanguageTokenizer tokenizer,
                int sp) {
            return tokenizer.createToken(sp, tokenizer.pos - sp, TokenTypes.TT_WHITESPACE);
        }
    }

    /**
     * State In a number.
     */
    private static final class CLLTStateInNumber implements CLLTState {

        /**
         * Constructor.
         */
        public CLLTStateInNumber() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextToken(ClikeLanguageTokenizer tokenizer,
                char thisChar,
                int sp) {
            LanguageToken ret = null;
            if (!CharUtils.isAlNum(thisChar)
                    && (thisChar != '.')) {
                tokenizer.pos--;
                ret = tokenizer.createToken(sp,
                        tokenizer.pos - sp + 1,
                        TokenTypes.TT_NUMBER);
                tokenizer.state = BASE;
            }
            return ret;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextTokenLast(ClikeLanguageTokenizer tokenizer,
                int sp) {
            return tokenizer.createToken(sp, tokenizer.pos - sp, TokenTypes.TT_NUMBER);
        }
    }

    /**
     * State In punctuation.
     */
    private static final class CLLTStateInPunctuation implements CLLTState {

        /**
         * Constructor.
         */
        public CLLTStateInPunctuation() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextToken(ClikeLanguageTokenizer tokenizer,
                char thisChar,
                int sp) {
            LanguageToken ret = null;
            if (tokenizer.isWordStartChar(thisChar)
                    || CharUtils.isDigit(thisChar)
                    || thisChar == ' '
                    || thisChar == '/'
                    || thisChar == '\''
                    || thisChar == '"'
                    || thisChar == '\n'
                    || thisChar == '\t') {
                tokenizer.pos--;
                ret = tokenizer.createToken(sp,
                        tokenizer.pos - sp + 1,
                        TokenTypes.TT_PUNC);
                tokenizer.state = BASE;
            }
            return ret;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextTokenLast(ClikeLanguageTokenizer tokenizer,
                int sp) {
            return tokenizer.createToken(sp, tokenizer.pos - sp, TokenTypes.TT_PUNC);
        }
    }

    /**
     * State in a sigle line comment.
     */
    private static final class CLLTStateInSingleComment implements CLLTState {

        /**
         * Constructor.
         */
        public CLLTStateInSingleComment() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextToken(ClikeLanguageTokenizer tokenizer,
                char thisChar,
                int sp) {
            LanguageToken ret = null;
            if (thisChar == '\n') {
                tokenizer.pos--;
                ret = tokenizer.createToken(sp,
                        tokenizer.pos - sp + 1,
                        TokenTypes.TT_COMMENT);
                tokenizer.state = BASE;
            }
            return ret;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextTokenLast(ClikeLanguageTokenizer tokenizer,
                int sp) {
            return tokenizer.createToken(sp, tokenizer.pos - sp, TokenTypes.TT_COMMENT);
        }
    }

    /**
     * State In comment.
     */
    private static final class CLLTStateInComment implements CLLTState {

        /**
         * Constructor.
         */
        public CLLTStateInComment() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextToken(ClikeLanguageTokenizer tokenizer,
                char thisChar,
                int sp) {
            LanguageToken ret = null;
            switch (thisChar) {
                case '/':
                    if (tokenizer.buffer.charAt(tokenizer.pos - 1) == '*') {
                        tokenizer.state = BASE;
                        ret = tokenizer.createToken(sp,
                                tokenizer.pos - sp + 1,
                                TokenTypes.TT_COMMENT);
                    }
                    break;
                case '\n':
                    if (sp == tokenizer.pos) {
                        ret = tokenizer.createToken(sp,
                                1,
                                TokenTypes.TT_NEWLINE);
                    } else {
                        tokenizer.pos--;
                        ret = tokenizer.createToken(sp,
                                tokenizer.pos - sp + 1,
                                TokenTypes.TT_COMMENT);
                    }
                    break;
                default:
                    break;
            }
            return ret;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextTokenLast(ClikeLanguageTokenizer tokenizer,
                int sp) {
            return tokenizer.createToken(sp, tokenizer.pos - sp, TokenTypes.TT_COMMENT);
        }
    }

    /**
     * State base.
     */
    private static final class CLLTStateBase implements CLLTState {

        /**
         * Constructor.
         */
        public CLLTStateBase() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextToken(ClikeLanguageTokenizer tokenizer,
                char thisChar,
                int sp) {
            LanguageToken ret = null;
            switch (thisChar) {
                case '\'':
                    tokenizer.state = INSGLQ;
                    break;
                case '"':
                    tokenizer.state = INDBLQ;
                    break;
                case '/':
                    if (tokenizer.buffer.charAt(tokenizer.pos + 1) == '/') {
                        tokenizer.state = INSCOMMENT;
                    } else if (tokenizer.buffer.charAt(tokenizer.pos + 1) == '*') {
                        tokenizer.state = INCOMMENT;
                    } else {
                        tokenizer.state = INPUNC;
                    }
                    break;
                case ' ':
                case '\t':
                    tokenizer.state = INWHITE;
                    break;
                case '\n':
                    ret = tokenizer.createToken(sp, 1, TokenTypes.TT_NEWLINE);
                    break;
                default:
                    if (tokenizer.isWordStartChar(thisChar)) {
                        tokenizer.state = INWORD;
                    } else if (CharUtils.isDigit(thisChar)) {
                        tokenizer.state = INNUM;
                    } else if (CharUtils.isHash(thisChar)) {
                        tokenizer.state = INHASH;
                    } else {
                        tokenizer.state = INPUNC;
                    }
                    break;
            }
            return ret;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextTokenLast(ClikeLanguageTokenizer tokenizer,
                int sp) {
            return tokenizer.createToken(sp, 1, TokenTypes.TT_NEWLINE);
        }
    }

    /**
     * State in single quotes.
     */
    private static final class CLLTStateInSingleQuotes implements CLLTState {

        /**
         * Constructor.
         */
        public CLLTStateInSingleQuotes() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextToken(ClikeLanguageTokenizer tokenizer,
                char thisChar,
                int sp) {
            LanguageToken ret = null;
            switch (thisChar) {
                case '\\':
                    tokenizer.pos++;
                    break;
                case '\'':
                    tokenizer.state = BASE;
                    ret = tokenizer.createToken(sp,
                            tokenizer.pos - sp + 1,
                            TokenTypes.TT_CHAR);
                    break;
                case '\n':	// INVALID TOKEN
                    tokenizer.state = BASE;
                    tokenizer.pos--;
                    ret = tokenizer.createToken(sp,
                            tokenizer.pos - sp + 1,
                            TokenTypes.TT_CHAR);
                    break;
                default:
                    // Do nothing and wait for the token end.
                    break;
            }
            return ret;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextTokenLast(ClikeLanguageTokenizer tokenizer,
                int sp) {
            return tokenizer.createToken(sp, tokenizer.pos - sp, TokenTypes.TT_CHAR);
        }
    }

    /**
     * State in double quotes.
     */
    private static final class CLLTStateInDoubleQuotes implements CLLTState {

        /**
         * Constructor.
         */
        public CLLTStateInDoubleQuotes() {
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextToken(ClikeLanguageTokenizer tokenizer,
                char thisChar,
                int sp) {
            LanguageToken ret = null;
            switch (thisChar) {
                case '\\':
                    tokenizer.pos++;
                    break;
                case '"':
                    tokenizer.state = BASE;
                    ret = tokenizer.createToken(sp,
                            tokenizer.pos - sp + 1,
                            TokenTypes.TT_STRING);
                    break;
                case '\n':	// INVALID TOKEN
                    tokenizer.state = BASE;
                    tokenizer.pos--;
                    ret = tokenizer.createToken(sp,
                            tokenizer.pos - sp + 1,
                            TokenTypes.TT_STRING);
                    break;
                default:
                    // Do nothing and wait for the token end.
                    break;
            }
            return ret;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public LanguageToken nextTokenLast(ClikeLanguageTokenizer tokenizer, int sp) {
            return tokenizer.createToken(sp, tokenizer.pos - sp, TokenTypes.TT_STRING);
        }
    }

    /**
     * Default constructor.
     */
    public ClikeLanguageTokenizer() {
        resWords = getReservedWords();
    }

    @Override
    public void setBuffer(CharSequence buffer) {
        super.setBuffer(buffer);
        this.state = BASE;
    }

    /**
     * Get a collection of the reserved words for this language.
     *
     * @return The reserved words.
     */
    protected abstract Collection<String> getReservedWords();

    /**
     * Is the given word a reserved one?
     *
     * @param s The word.
     * @return True if it is in the reserved words.
     */
    protected boolean isReservedWord(String s) {
        return resWords.contains(s);
    }

    /**
     * Is the given character the start of a word?
     *
     * @param c The character.
     * @return True if it is.
     */
    protected boolean isWordStartChar(char c) {
        return CharUtils.isAlphaUnderscore(c);
    }

    @Override
    public LanguageToken next() {
        final int sp = pos;

        final int bLen = buffer.length();
        while (pos < bLen) {
            final char thisChar = buffer.charAt(pos);
//	    if (thisChar == '\r') {
//		buffer.charAt(pos) = ' ';
//	    }

            final LanguageToken token = state.nextToken(this, thisChar, sp);
            pos++;
            if (token != null) {
                return token;
            }
        }
        if (pos != sp) {
            return state.nextTokenLast(this, sp);
        }
        return null; // ** SHOULD NEVER GET HERE!!!!!!!
    }

    @Override
    public Set<TokenType> getPossibleTokenTypes() {
        final Set<TokenType> ret = new HashSet<>();
        ret.add(TokenTypes.TT_NEWLINE);
        ret.add(TokenTypes.TT_CHAR);
        ret.add(TokenTypes.TT_STRING);
        ret.add(TokenTypes.TT_COMMENT);
        ret.add(TokenTypes.TT_NEWLINE);
        ret.add(TokenTypes.TT_PUNC);
        ret.add(TokenTypes.TT_NUMBER);
        ret.add(TokenTypes.TT_WHITESPACE);
        ret.add(TokenTypes.TT_KEYWORD);
        ret.add(TokenTypes.TT_WORD);
        return ret;
    }
}
