////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: CobolLanguageTokenizer.java 280 2009-08-07 08:26:45Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tokenizers.impl;

import java.util.Collection;
import java.util.HashSet;
import java.util.ResourceBundle;
import java.util.Set;
import uk.co.ruelgnoj.orion.tokenizers.CharUtils;
import uk.co.ruelgnoj.orion.tokenizers.LanguageToken;
import uk.co.ruelgnoj.orion.tokenizers.TokenType;
import uk.co.ruelgnoj.orion.tokenizers.TokenTypes;

/**
 * A tokenizer for the COBOL langauge.
 *
 * @author Steve Millington
 * @version $Id: CobolLanguageTokenizer.java 280 2009-08-07 08:26:45Z steve $
 */
public class CobolLanguageTokenizer extends AbstractLanguageTokenizer {

    private static final Collection<String> RES_WORDS;
    //
    // Static initialization of the reserved WORDS
    //
    private static final String[] WORDS = {
        "accept",
        "access",
        "add",
        "address",
        "advancing",
        "after",
        "aligned",
        "all",
        "alphabet",
        "alphabetic",
        "also",
        "alternate",
        "and",
        "any",
        "are",
        "ascending",
        "assign",
        "at",
        "author",
        "automatic",
        "before",
        "beginning",
        "binary",
        "blank",
        "block",
        "bottom",
        "by",
        "call",
        "character",
        "characters",
        "close",
        "code-set",
        "collating",
        "contains",
        "content",
        "comp",
        "comp-1",
        "comp-2",
        "comp-3",
        "comp-4",
        "comp-5",
        "comp-6",
        "comp-n",
        "comp-x",
        "compute",
        "configuration",
        "continue",
        "converting",
        "copy",
        "corresponding",
        "data",
        "date-compiled",
        "date-written",
        "debugging",
        "declaratives",
        "delete",
        "delimited",
        "depending",
        "display",
        "display-1",
        "divide",
        "division",
        "do",
        "dynamic",
        "ebcdic",
        "eject",
        "else",
        "end",
        "ending",
        "end-evaluate",
        "end-if",
        "end-invoke",
        "end-perform",
        "end-search",
        "end-start",
        "end-unstring",
        "end-xml",
        "entry",
        "environment",
        "equal",
        "error",
        "evaluate",
        "examine",
        "exception",
        "exit",
        "extend",
        "external",
        "false",
        "fd",
        "file",
        "file-control",
        "filler",
        "footing",
        "for",
        "from",
        "function-pointer",
        "giving",
        "global",
        "go",
        "goback",
        "greater",
        "high-value",
        "high-values",
        "i-o",
        "id",
        "identification",
        "if",
        "in",
        "indexed",
        "input",
        "input-output",
        "inspect",
        "into",
        "invoke",
        "is",
        "key",
        "label",
        "length",
        "less",
        "linage",
        "lines",
        "linkage",
        "local-storage",
        "lock",
        "low-value",
        "low-values",
        "manual",
        "merge",
        "mode",
        "move",
        "multiply",
        "national",
        "native",
        "negative",
        "new",
        "next",
        "not",
        "numeric",
        "null",
        "object-computer",
        "occurs",
        "of",
        "omitted",
        "on",
        "open",
        "or",
        "organization",
        "other",
        "output",
        "overflow",
        "padding",
        "parse",
        "password",
        "perform",
        "pic",
        "plus",
        "pointer",
        "procedure",
        "procedure-pointer",
        "procedures",
        "processing",
        "program",
        "program-id",
        "quote",
        "quotes",
        "reel",
        "random",
        "rd",
        "read",
        "record",
        "recording",
        "records",
        "redefines",
        "reference",
        "relative",
        "release",
        "remainder",
        "remarks",
        "renames",
        "replacing",
        "reports",
        "reserve",
        "returning",
        "rewrite",
        "run",
        "sd",
        "search",
        "section",
        "select",
        "sentence",
        "sequence",
        "sequential",
        "set",
        "size",
        "sort",
        "source-computer",
        "space",
        "spaces",
        "special-names",
        "standard",
        "standard-1",
        "standard-2",
        "start",
        "status",
        "stop",
        "string",
        "subtract",
        "sync",
        "test",
        "then", // CL2 at least!
        "than",
        "through",
        "thru",
        "times",
        "to",
        "top",
        "transform",
        "true",
        "unit",
        "unstring",
        "until",
        "usage",
        "use",
        "using",
        "value",
        "varying",
        "when",
        "while",
        "with",
        "working-storage",
        "write",
        "xml",
        "zero",
        "zeroes",
        "zeros"
    };

    static {
        RES_WORDS = new HashSet<>();
        for (String word : WORDS) {
            RES_WORDS.add(word);
            RES_WORDS.add(word.toUpperCase());
        }
    }
    private static final String RBNAME = "uk.co.ruelgnoj.orion.tokenizers.resources";
    private static final char EOF = (char) 0x00;
    private static final char EOS = (char) -1;
    private static final int COLUMNS = 72;
    private int column_;

    /**
     * Read the next character, but return EOS if we have reached column_ 72.
     * This is used by code for tokens within the main text area.
     *
     * @return The character, or EOS we are outside the middle section.
     */
    private char readBounded() {
        if (column_ >= COLUMNS) {
            return EOS;
        } else {
            return read();
        }
    }

    private char peek() {
        if (pos >= buffer.length()) {
            return EOF;
        } else {
            final char ret = buffer.charAt(pos);
            return ret;
        }

    }

    private char read() {
        if (pos >= buffer.length()) {
            return EOF;
        } else {
            final char ret = buffer.charAt(pos);
            pos++;
            column_++;
            return ret;
        }
    }

    private void unread() {
        column_--;
        pos--;
    }

    /**
     * Constructor.
     *
     * @see uk.co.ruelgnoj.orion.tokenizers.impl.AbstractLanguageTokenizer
     */
    public CobolLanguageTokenizer() {
    }

    @Override
    public void setBuffer(CharSequence buffer) {
        super.setBuffer(buffer);
        this.column_ = 0;
    }

    private boolean isReservedWord(String s) {
        return RES_WORDS.contains(s);
    }

    private static boolean isAlNumMinus(char c) {
        return CharUtils.isAlNum(c) || (c == '-');
    }

    private LanguageToken doLineNumbers() {
        final int sp = pos;		// save the start position
        while (column_ < 6) {
            final char c = read();
            if (c == EOF) {
                return createToken(sp, pos - sp, TokenTypes.TT_LINENUM);
            } else if (c == '\n') {
                unread();
                return createToken(sp, pos - sp, TokenTypes.TT_LINENUM);
            }
            if (c == '\t') {
                column_ = 6;
                return createToken(sp, pos - sp, TokenTypes.TT_LINENUM);
            }
        }
        return createToken(sp, pos - sp, TokenTypes.TT_LINENUM);
    }

    private LanguageToken doComment() {
        final int sp = pos;
        char c;
        while ((c = readBounded()) != EOF) {
            switch (c) {
                case EOS:
                    return createToken(sp, pos - sp, TokenTypes.TT_COMMENT);
                case '\n':
                    unread();
                    return createToken(sp, pos - sp, TokenTypes.TT_COMMENT);
            }
        }
        return createToken(sp, pos - sp, TokenTypes.TT_COMMENT);
    }

    private LanguageToken doSingleQuotedString() {
        final int sp = pos;
        read(); // Skip the opening quote
        char c;
        while ((c = readBounded()) != EOF) {
            switch (c) {
                case '\'':
                    return createToken(sp, pos - sp, TokenTypes.TT_STRING);
                case EOS:
                    return createToken(sp, pos - sp, TokenTypes.TT_STRING);
                default:
                    break;
            }
        }
        return createToken(sp, pos - sp, TokenTypes.TT_STRING);
    }

    private LanguageToken doDoubleQuotedString() {
        final int sp = pos;
        read(); // Skip the opening quote
        char c;
        while ((c = readBounded()) != EOF) {
            switch (c) {
                case '"':
                    return createToken(sp, pos - sp, TokenTypes.TT_STRING);
                case EOS:
                    return createToken(sp, pos - sp, TokenTypes.TT_STRING);
                default:
                    break;
            }
        }
        return createToken(sp, pos - sp, TokenTypes.TT_STRING);
    }

    private LanguageToken doWord() {
        final int sp = pos;
        read(); // Consume the first character of the word.
        LanguageToken tok = null;
        char c;
        while ((c = readBounded()) != EOF) {
            if (c == EOS) {
                tok = createToken(sp, pos - sp, TokenTypes.TT_WORD);
                break;
            } else if (!isAlNumMinus(c)) {
                unread();
                tok = createToken(sp, pos - sp, TokenTypes.TT_WORD);
                break;
            }
        }
        if (tok == null) {
            tok = createToken(sp, pos - sp, TokenTypes.TT_WORD);
        }

        //
        // Now modify the token if it is a reserved word!
        //
        if (isReservedWord(tok.asString(buffer))) {
            tok.setType(TokenTypes.TT_KEYWORD);
        }
        return tok;
    }

    private LanguageToken doNumber() {
        final int sp = pos;
        read(); // Consume the first character of the number.

        char c;
        while ((c = readBounded()) != EOF) {
            if (c == EOS) {
                unread();
                while ((c = read()) == '.') {
                    unread();
                    unread();
                }
                return createToken(sp, pos - sp, TokenTypes.TT_NUMBER);
            } else if (isAlNumMinus(c) || (c == '.')) {
                // Do Nothing
            } else {
                unread();
                unread();
                while ((c = readBounded()) == '.') {
                    unread();
                    unread();
                }
                return createToken(sp, pos - sp, TokenTypes.TT_NUMBER);
            }
        }
        return createToken(sp, pos - sp, TokenTypes.TT_NUMBER);
    }

    private LanguageToken doWhite() {
        final int sp = pos;
        read(); // Consume the first white character.
        char c;
        while ((c = read()) != EOF) {
            if (!CharUtils.isWhite(c)) {
                unread();
                return createToken(sp, pos - sp, TokenTypes.TT_WHITESPACE);
            }
        }
        return createToken(sp, pos - sp, TokenTypes.TT_WHITESPACE);
    }

    private LanguageToken doTrailingLineNumbers() {
        final int sp = pos;
        char c;
        while ((c = read()) != EOF) {
            if (c == '\n') {
                unread();
                break;
            }
        }
        return createToken(sp, pos - sp, TokenTypes.TT_LINENUM);
    }

    private LanguageToken doPunc() {
        final int sp = pos;
        read(); // Consume the first punctuation char.
        char c;
        while ((c = read()) != EOF) {
            if (isAlNumMinus(c) || c == ' ' || c == '"' || c == '\'' || c == '\n' || c == '\t') {
                unread();
                return createToken(sp, pos - sp, TokenTypes.TT_PUNC);
            }
        }
        return createToken(sp, pos - sp, TokenTypes.TT_PUNC);
    }

    private LanguageToken doNormalBit() {
        final LanguageToken token;

        final char thisChar = read();
        unread();
        switch (thisChar) {
            case '\n':
                token = createToken(pos, 1, TokenTypes.TT_NEWLINE);
                read();
                column_ = 0;
                break;
            case '\'':
                token = doSingleQuotedString();
                break;
            case '"':
                token = doDoubleQuotedString();
                break;
            default:
                if (CharUtils.isAlpha(thisChar)) {
                    token = doWord();
                } else if (CharUtils.isDigit(thisChar)) {
                    token = doNumber();
                } else if (CharUtils.isWhite(thisChar)) {
                    token = doWhite();
                } else {
                    token = doPunc();
                }
                break;
        }
        return token;
    }

    @Override
    public LanguageToken next() {
        final LanguageToken token;
        final char thisChar = read();

        if (thisChar == '\n') {
            token = createToken(pos, 1, TokenTypes.TT_NEWLINE);
            column_ = 0;
        } else {
            unread();
            switch (column_) {
                case 0:
                    token = doLineNumbers();
                    break;
                case 6:
                    // Cope with a tab in the LINENUM stuff
                    if ((pos > 0) && (buffer.charAt(pos - 1) == '\t')) {
                        token = doNormalBit();
                    } else if (buffer.charAt(pos) == '*') {
                        token = doComment();
                    } else if (buffer.charAt(pos) == '-') {
                        token = doPunc();   // This is a continuation char
                    } else if (!CharUtils.isWhite(buffer.charAt(pos))) {
                        token = doPunc();   // This is a hack - it should be an error!
                    } else {
                        token = doWhite();
                    }
                    break;
                case COLUMNS:
                    token = doTrailingLineNumbers();
                    break;
                default:
                    token = doNormalBit();
                    break;
            }
        }
        return token;
    }

    @Override
    public String getDescription() {
        return ResourceBundle.getBundle(RBNAME).getString("lang.cobol.desc");
    }

    @Override
    public String getLanguage() {
        return "cobol";
    }

    @Override
    public Set<TokenType> getPossibleTokenTypes() {
        Set<TokenType> ret = new HashSet<>();
        ret.add(TokenTypes.TT_NEWLINE);
        ret.add(TokenTypes.TT_CHAR);
        ret.add(TokenTypes.TT_STRING);
        ret.add(TokenTypes.TT_COMMENT);
        ret.add(TokenTypes.TT_PUNC);
        ret.add(TokenTypes.TT_NUMBER);
        ret.add(TokenTypes.TT_WHITESPACE);
        ret.add(TokenTypes.TT_KEYWORD);
        ret.add(TokenTypes.TT_WORD);
        ret.add(TokenTypes.TT_LINENUM);
        return ret;
    }
}
