////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: PLILanguageTokenizer.java 285 2011-07-19 12:37:42Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tokenizers.impl;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import uk.co.ruelgnoj.orion.tokenizers.LanguageToken;
import uk.co.ruelgnoj.orion.tokenizers.TokenType;
import uk.co.ruelgnoj.orion.tokenizers.TokenTypes;

/**
 * A tokenizer for PL/I files.
 *
 * @author Steve Millington
 * @version $Id: PLILanguageTokenizer.java 285 2011-07-19 12:37:42Z steve $
 */
public class PLILanguageTokenizer extends ClikeLanguageTokenizer {
    //
    // The reserved words is a static table, that is initialized by the static
    // initializer at the bottom of this file. It is returned to the superclass
    // when it asks for them.
    //
    private static final Collection<String> RES_WORDS;
    private static final String[] WORDS = {
        "ALIGNED",
        "ALLOCATE",
        "ALLOC",
        "AREA",
        "ASCII",
        "BASED",
        "BEGIN",
        "BIN",
        "BINARY",
        "BACKWARD",
        "BIT",
        "BKWD",
        "BUFFERED",
        "BUF",
        "BUILTIN",
        "BY",
        "BYNAME",
        "CALL",
        "CHAR",
        "CHARACTER",
        "CLOSE",
        "CONTROLLLED",
        "DATA",
        "DCL",
        "DEC",
        "DECLARE",
        "DECIMAL",
        "DEFINED",
        "DO",
        "ELSE",
        "EDIT",
        "END",
        "ENTRY",
        "ERROR",
        "EXTERNAL",
        "EXT",
        "FILE",
        "FINISH",
        "FIXED",
        "FLOAT",
        "FORTRAN",
        "FOREVER",
        "GENERIC",
        "GET",
        "GO",
        "GOTO",
        "GRAPHIC",
        "IF",
        "INIT",
        "INITIAL",
        "INCLUDE",
        "ITERATE",
        "LEAVE",
        "LIST",
        "MAIN",
        "NAME",
        "NULL",
        "ON",
        "OPTIONS",
        "OTHERWISE",
        "PIC",
        "PICTURE",
        "PROC",
        "PROCEDURE",
        "PUT",
        "RECURSIVE",
        "REORDER",
        "REPLACE",
        "RETURN",
        "RETURNS",
        "SELECT",
        "SIGNAL",
        "SKIP",
        "SPACE",
        "STATIC",
        "STOP",
        "STORAGE",
        "THEN",
        "TO",
        "UNTIL",
        "UNION",
        "VARIABLE",
        "VAR",
        "VARYING",
        "WHEN",
        "WHILE"
    };

    static {
        RES_WORDS = new HashSet<>(Arrays.asList(WORDS));
    }

    /**
     * Constructor.
     * @see uk.co.ruelgnoj.orion.tokenizers.impl.ClikeLanguageTokenizer
     */
    public PLILanguageTokenizer() {
        super();
    }

    @Override
    protected Collection<String> getReservedWords() {
        return RES_WORDS;
    }

    @Override
    public LanguageToken next() {
        //
        // Convert 'CHAR' tokens to "STRING" tokens!
        //
        final LanguageToken tok = super.next();
        if (tok.getType() == TokenTypes.TT_CHAR) {
            tok.setType(TokenTypes.TT_STRING);
        }
        return tok;
    }

    @Override
    public String getDescription() {
        return "PL/I";
    }

    @Override
    public String getLanguage() {
        return "pli";
    }

    @Override
    protected boolean isReservedWord(String s) {
        //
        // PL/I is case insensative. Therefore we need to ensure that keyword
        // checking is case-insensitive.
        //
        return super.isReservedWord(s.toUpperCase());
    }

    @Override
    public Set<TokenType> getPossibleTokenTypes() {
        final Set<TokenType> s = super.getPossibleTokenTypes();
        final Set<TokenType> ret = new HashSet<>(s);
        ret.remove(TokenTypes.TT_CHAR);
        return ret;
    }
}
