////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: PropertiesLanguageTokenizer.java 231 2008-06-04 15:09:57Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tokenizers.impl;

import java.util.HashSet;
import java.util.Set;
import uk.co.ruelgnoj.orion.tokenizers.LanguageToken;
import uk.co.ruelgnoj.orion.tokenizers.TokenType;
import uk.co.ruelgnoj.orion.tokenizers.TokenTypes;

/**
 * A tokenizer of a Java Properties file.
 * <p>This is a poor first-pass at this tokenizer.
 * <p>It does not cope with:
 * <ul>
 *   <li>Continuation lines</li>
 *   <li>'Choice' type variables</li>
 * </ul>
 *
 * @author Steve Millington
 * @version $Id: PropertiesLanguageTokenizer.java 231 2008-06-04 15:09:57Z steve $
 */
public class PropertiesLanguageTokenizer extends AbstractLanguageTokenizer {
    private static final int ST_BASE = 0;
    private static final int ST_INSCOMMENT = 1;
    private static final int ST_INWHITE = 2;
    private static final int ST_INPNAME = 3;
    private static final int ST_INPVALUE = 4;
    private static final int ST_INEQUALS = 5;
    private static final int ST_ININSERT = 6;

    //
    // These are a fudge.
    //
    private static final TokenType TT_PROPNAME = TokenTypes.TT_KEYWORD;
    private static final TokenType TT_PROPVALUE = TokenTypes.TT_STRING;
    private static final TokenType TT_PROPINSERT = TokenTypes.TT_CHAR;

    private int state;

    /**
     * Constructor.
     * @see uk.co.ruelgnoj.orion.tokenizers.impl.AbstractLanguageTokenizer
     */
    public PropertiesLanguageTokenizer() {
    }

    @Override
    public void setBuffer(CharSequence buffer) {
    	super.setBuffer(buffer);
	this.state = ST_BASE;
    }

    @Override
    public LanguageToken next() {
	final int sp = pos;
	LanguageToken token = null;

	final int bLen = buffer.length();
    	while (pos < bLen) {
	    switch (state) {
	    	case ST_BASE:
		    switch (buffer.charAt(pos)) {
		    	case '#':
			    state = ST_INSCOMMENT;
			    break;
			case ' ':
			case '\t':
			    state = ST_INWHITE;
			    break;
			case '\n':
			    token = createToken(sp, 1, TokenTypes.TT_NEWLINE);
			    break;
			default:
			    state = ST_INPNAME;
			    break;
		    }
		    break;
		case ST_INSCOMMENT:
		    if (buffer.charAt(pos) == '\n') {
		    	pos--;
			token = createToken(sp, pos - sp + 1,
				TokenTypes.TT_COMMENT);
			state = ST_BASE;
		    }
		    break;
		case ST_INPNAME:
		    switch (buffer.charAt(pos)) {
		    	case '=':
			    pos--;
			    token = createToken(sp, pos - sp + 1,
				    TokenTypes.TT_KEYWORD);
			    state = ST_INEQUALS;
			    break;
			default:
			    break;
		    }
		    break;
		case ST_ININSERT:
		    if (buffer.charAt(pos) == '}') {
			token = createToken(sp, pos - sp + 1, TT_PROPINSERT);
			state = ST_INPVALUE;
		    }
		    break;
		case ST_INPVALUE:
		    if (buffer.charAt(pos) == '{') {
		    	pos--;
			token = createToken(sp, pos - sp + 1, TT_PROPVALUE);
			state = ST_ININSERT;
		    } else if (buffer.charAt(pos) == '\n') {
			if (pos != sp) {
			    pos--;
			    token = createToken(sp, pos - sp + 1, TT_PROPVALUE);
			} else {
			    pos--;
			}
			state = ST_BASE;
		    }
		    break;
		case ST_INEQUALS:
		    if (buffer.charAt(pos) == '=') {
			token = createToken(sp, pos - sp + 1,
				TokenTypes.TT_PUNC);
			state = ST_INPVALUE;
		    }
		    break;
		default:
		    break;
	    }
	    pos++;
	    if (token != null) {
	    	return token;
	    }
	}
	if (pos != sp) {
	    switch (state) {
	    	case ST_INPVALUE:
		    token = createToken(sp, pos - sp, TT_PROPVALUE);
		    break;
		default:
		    break;
	    }
	    return token;
	}
	// SHOULD NEVER GET HERE!!!!!!!
	return null;
    }

    @Override
    public Set<TokenType> getPossibleTokenTypes() {
	final Set<TokenType> ret = new HashSet<>();
	ret.add(TokenTypes.TT_NEWLINE);
	ret.add(TokenTypes.TT_COMMENT);
	ret.add(TokenTypes.TT_NEWLINE);
	ret.add(TokenTypes.TT_PUNC);
	ret.add(TokenTypes.TT_WHITESPACE);
	ret.add(TT_PROPNAME);
	ret.add(TT_PROPVALUE);
	ret.add(TT_PROPINSERT);
	return ret;
    }

    @Override
    public String getDescription() {
    	return "Java Properties";
    }

    @Override
    public String getLanguage() {
    	return "properties";
    }
}
