////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: SQLMixIn.java 232 2008-06-04 15:31:24Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.tokenizers.impl;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Set;
import uk.co.ruelgnoj.orion.tokenizers.LanguageToken;
import uk.co.ruelgnoj.orion.tokenizers.LanguageTokenizer;
import uk.co.ruelgnoj.orion.tokenizers.TokenType;
import uk.co.ruelgnoj.orion.tokenizers.TokenTypes;

/**
 * A mixin for tokenizing EXEC SQL Statements.
 *
 * @author Steve Millington
 * @version $Id: SQLMixIn.java 232 2008-06-04 15:31:24Z steve $
 */
public class SQLMixIn implements LanguageTokenizer {
    public enum Flavour {

        C,
        COBOL,
        PLI
    }

    private static final Set<String> RESERVED_WORDS;

    static {
	final Set<String> words = new HashSet<>();
	words.add("BEGIN");
	words.add("BY");
	words.add("CLOSE");
        words.add("COMMIT");
        words.add("CONNECT");
        words.add("CURRENT");
	words.add("CURSOR");
	words.add("DECLARE");
	words.add("DELETE");
        words.add("DISCONNECT");
	words.add("END");
	words.add("FETCH");
	words.add("FOR");
	words.add("FOUND");
	words.add("FROM");
	words.add("GOTO");
        words.add("IDENTIFIED");
        words.add("IN");
	words.add("INCLUDE");
	words.add("INSERT");
	words.add("INTO");
        words.add("MODE");
	words.add("NOT");
	words.add("OPEN");
	words.add("ORDER");
        words.add("RELEASE");
        words.add("RESET");
        words.add("ROLLBACK");
        words.add("SAVEPOINT");
	words.add("SECTION");
	words.add("SELECT");
	words.add("SET");
        words.add("SHARE");
	words.add("SQL");
	words.add("SQLERROR");
        words.add("SQLWARNING");
	words.add("UPDATE");
        words.add("USING");
	words.add("VALUES");
	words.add("WHENEVER");
	words.add("WHERE");
        words.add("WORK");
	RESERVED_WORDS = words;
    }

    private LanguageTokenizer delegate;
    private final Flavour flavour;
    private boolean inExecSQL;
    private Queue<LanguageToken> queue;

    /**
     * Constructor.
     * @param delegate The tokenizer from which to obtain base functionality
     * @param flavour The flavour of mixin, either C or COBOL.
     * @throws java.lang.IllegalArgumentException if the flavor is not one
     * of the valid static constants defined by this class.
     * @throws java.lang.NullPointerException if the delegate is null
     */
    public SQLMixIn(LanguageTokenizer delegate, Flavour flavour) {
	if (delegate == null) {
	    throw new NullPointerException("delegate");
	}
        
	if (flavour == null) {
		throw new IllegalArgumentException("flavour");
	}

	this.delegate = delegate;
	this.flavour = flavour;
	this.queue = new LinkedList<>();
	inExecSQL = false;
    }

    @Override
    public void setBuffer(CharSequence buffer) {
	delegate.setBuffer(buffer);
    }

    @Override
    public CharSequence getBuffer() {
	return delegate.getBuffer();
    }

    @Override
    public boolean hasNext() {
	return (!queue.isEmpty()) || delegate.hasNext();
    }

    @Override
    public boolean canContainDBEntries() {
	return delegate.canContainDBEntries();
    }

    @Override
    public LanguageToken next() {
	if (!queue.isEmpty()) {
	    final LanguageToken tok = queue.poll();
	    if (inExecSQL) {
		if ((tok.getType() == TokenTypes.TT_WORD)
			|| (tok.getType() == TokenTypes.TT_KEYWORD)) {
		    final String tokStr = tok.asString(delegate.getBuffer());
		    if (RESERVED_WORDS.contains(tokStr.toUpperCase())) {
			tok.setType(TokenTypes.TT_SQL_API);
		    } else {
			tok.setType(TokenTypes.TT_WORD);
		    }
		}
	    }
	    return tok;
	}
    	LanguageToken tok = delegate.next();
	if (tok == null) {
	    return tok;
	}
	//
	// Attempt to highlight EXEC SQL commands.
	//
	if (inExecSQL) {
	    if (tok.getType() == TokenTypes.TT_PUNC) {
		if (delegate.getBuffer().charAt(tok.getEndOffset()) == ';') {
		    if (flavour == Flavour.C ||
                            flavour == Flavour.PLI) {
			inExecSQL = false;
		    }
		}
	    } else if ((tok.getType() == TokenTypes.TT_WORD)
		    || (tok.getType() == TokenTypes.TT_KEYWORD)) {
		String tokStr = tok.asString(delegate.getBuffer());
		if (RESERVED_WORDS.contains(tokStr.toUpperCase())) {
		    tok.setType(TokenTypes.TT_SQL_API);
		} else {
		    tok.setType(TokenTypes.TT_WORD);
		}
		if ((flavour == Flavour.COBOL)
			&& "END-EXEC".equalsIgnoreCase(tokStr)) {
		    tok.setType(TokenTypes.TT_SQL_API);
		    inExecSQL = false;
		}
	    }
	} else {
	    if (tok.getType() == TokenTypes.TT_WORD) {
		String tokStr = tok.asString(delegate.getBuffer());
		if ("EXEC".equalsIgnoreCase(tokStr)) {
		    queue.offer(tok.clone());
		    tok = delegate.next();
		    while (true) {
			queue.offer(tok.clone());
			if (tok.getType() == TokenTypes.TT_WORD) {
			    tokStr = tok.asString(delegate.getBuffer());
			    tok = queue.poll();
			    if ("SQL".equalsIgnoreCase(tokStr)) {
				tok.setType(TokenTypes.TT_SQL_API);
				inExecSQL = true;
			    }
			    break;
			} else {
			    tok = delegate.next();
			}
		    }
		}
	    }
	}
	return tok;
    }

    /** {@inheritDoc} */
    @Override
    public String getDescription() {
	return ""; // Dummy
    }

    /** {@inheritDoc} */
    @Override
    public String getLanguage() {
    	return ""; // Dummy
    }

    @Override
    public Set<TokenType> getPossibleTokenTypes() {
	final Set<TokenType> ret = new HashSet<>();
	ret.addAll(delegate.getPossibleTokenTypes());
	ret.add(TokenTypes.TT_SQL_API);
	return ret;
    }

    /**
     * Removing is not supported.
     */
    @Override
    public void remove() {
	throw new UnsupportedOperationException();
    }

//    public LanguageToken next() {
//	return nextToken();
//    }

//    public boolean hasNext() {
//	return hasMoreTokens();
//    }

    @Override
    public Iterator<LanguageToken> iterator() {
	return this;
    }

}
