package uk.co.ruelgnoj.orion.tokenizers;

import java.util.Set;
import java.util.Iterator;
import org.junit.Test;
import static org.junit.Assert.*;

public class TokenizerFactoryTest {

    @Test(expected = NullPointerException.class)
    public void registerTokenizerNull1() {
        final TokenizerFactory factory = new TokenizerFactory();
        factory.registerTokenizer(null, "foo");
    }

    @Test(expected = NullPointerException.class)
    public void registerTokenizerNull2() {
        final TokenizerFactory factory = new TokenizerFactory();
        factory.registerTokenizer("foo", null);
    }

    @Test(expected = NullPointerException.class)
    public void registerSuffixNull1() {
        final TokenizerFactory factory = new TokenizerFactory();
        factory.registerSuffix(null, "foo");
    }

    @Test(expected = NullPointerException.class)
    public void registerSuffixNull2() {
        final TokenizerFactory factory = new TokenizerFactory();
        factory.registerSuffix("foo", null);
    }

    @Test
    public void testGetTokenizerByName() {
        final TokenizerFactory factory = new TokenizerFactory();
        //
        // Should get the default tokenizer
        //
        LanguageTokenizer lt1 = factory.getTokenizerByName(null);
        LanguageTokenizer lt2 = factory.getTokenizerByName("default");
        assertSame(lt1, lt2);

        //
        // Check that an empty name gives (by default) the default one
        //
        LanguageTokenizer lt3 = factory.getTokenizerByName("");
        assertSame(lt2, lt3);

    //
    // Check that an arbitrary name produces the default tokenizer
    //
    }

    @Test(expected = NullPointerException.class)
    public void getTokenizerNull() {
        final TokenizerFactory factory = new TokenizerFactory();
        final LanguageTokenizer lt = factory.getTokenizer(null);
    }

    @Test
    public void getTokenizerDefault() {
        final TokenizerFactory factory = new TokenizerFactory();
        final LanguageTokenizer dlt = factory.getTokenizer("default");
        final LanguageTokenizer lt1 = factory.getTokenizer("foo.hoo");
        assertSame(dlt, lt1);

        final LanguageTokenizer lt2 = factory.getTokenizer("hoo");
        assertSame(dlt, lt2);
    }

    @Test
    public void getTokenizerValues() {
        //
        // Check that .java and .c produce different things.
        //
        final TokenizerFactory factory = new TokenizerFactory();
        final LanguageTokenizer dlt = factory.getTokenizer("default");
        final LanguageTokenizer lt1 = factory.getTokenizer("foo.java");
        final LanguageTokenizer lt2 = factory.getTokenizer("foo.c");
        assertNotSame(lt1, lt2);
        assertNotSame(dlt, lt1);
        assertNotSame(dlt, lt2);
    }

    @Test
    public void getRegisteredTokenizerNames() {
        final TokenizerFactory factory = new TokenizerFactory();
        final Set toks = factory.getRegisteredTokenizerNames();
        assertTrue(!toks.isEmpty());

        final Iterator it = toks.iterator();
        while (it.hasNext()) {
            final LanguageTokenizer lt = factory.getTokenizer((String) it.next());
        }
    }

    @Test
    public void deregisterTokenizer() {
        final TokenizerFactory factory = new TokenizerFactory();

        Set toks = factory.getRegisteredTokenizerNames();
        final int numBefore = toks.size();
        factory.deregisterTokenizer("foo");
        toks = factory.getRegisteredTokenizerNames();
        int numAfter = toks.size();
        assertEquals(numBefore, numAfter);

        factory.deregisterTokenizer("java");
        toks = factory.getRegisteredTokenizerNames();
        numAfter = toks.size();
        assertEquals(numBefore, numAfter + 1);
        //
        // Ensure that a java thing now maps correctly to the default one.
        //
        final LanguageTokenizer dlt = factory.getTokenizerByName("default");
        LanguageTokenizer lt1 = factory.getTokenizer("foo.java");
        assertSame(dlt, lt1);
        //
        // Try deleting the default tokenizer, and ensure we get a null
        //
        factory.deregisterTokenizer("default");
        lt1 = factory.getTokenizer("foo.java");
        assertNull(lt1);
    }
}
