////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: DBManager.java 199 2008-01-14 17:04:44Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.application;

import uk.co.ruelgnoj.orion.project.Project;
import uk.co.ruelgnoj.orion.project.ProjectFactory;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.prefs.Preferences;
import uk.co.ruelgnoj.orion.project.Category;

/**
 * A manager for the Orion databases.
 *
 * @author Steve Millington
 * @version $Id: DBManager.java 199 2008-01-14 17:04:44Z steve $
 */
public class BasicProjectManager implements ProjectManager {

    private static final int DEFAULT_MAX_HISTORY_SIZE = 9;
    private int _maxHistorySize;
    private Preferences _prefs;

    /**
     * Constructor.
     */
    public BasicProjectManager() {
        this(null);
    }

    /**
     * Constructor that takes a preferences item. This is package privacy to
     * allow testing.
     *
     * @param _prefs the preferences.
     */
    BasicProjectManager(Preferences prefs) {
        _maxHistorySize = DEFAULT_MAX_HISTORY_SIZE;
        if (prefs == null) {
            prefs = Preferences.userNodeForPackage(getClass());
        }
        _prefs = prefs;
    }

    /**
     * Get the initial database to use.
     *
     * @return The database.
     */
    @Override
    public List<Project> getInitialProjects() {
        final String dbURL = _prefs.get("initial.db", null);
        if (dbURL == null) {
            return null;
        }
        final List<Category> cats = ProjectFactory.getCategories(dbURL);
        final List<Project> ret = new ArrayList<>();
        for (Category cat : cats) {
            for (Project p : cat.getProjects()) {
                ret.add(p);
            }
        }
        return ret;
    }

    /**
     * Set the initial database to use.
     *
     * @param db The database.
     */
    @Override
    public void setInitialProject(Project db) {
        _prefs.put("initial.db", db.getURL());
    }

    /**
     *
     * Get a list of the history of the files accessed.
     *
     * @return the list
     */
    @Override
    public List<Project> getHistory() {
        final List<Project> projects = new ArrayList<>();
        for (int i = 0; i < _maxHistorySize; i++) {
            final String name = _prefs.get("db.history." + i, null);
            if (name != null) {
                final List<Category> cats = ProjectFactory.getCategories(name);
                for (Category cat : cats) {
                    for (Project p : cat.getProjects()) {
                        projects.add(p);
                    }
                }
            }
        }
        return Collections.unmodifiableList(projects);
    }

    /**
     * Register a database as having been used. This will store the name in the
     * preferences, and ensure that it is in the list next time this application
     * is run.
     *
     * @param odb The database
     */
    @Override
    public void use(Project odb) {
        final String newURL = odb.getURL();
        String safe = newURL;
        for (int i = 0; i < _maxHistorySize; i++) {
            final String oldURL = _prefs.get("db.history." + i, null);

            _prefs.put("db.history." + i, safe);
            if ((oldURL == null) || oldURL.equals(newURL)) {
                return;
            }
            safe = oldURL;
        }
    }

    @Override
    public boolean allowsArbitraryProjects() {
        return true;
    }

    @Override
    public List<Category> getPredefined() {
        return Collections.emptyList();
    }
}
