////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: LookAndFeelPI.java 143 2007-07-09 15:25:34Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.application;

import javax.swing.JPanel;
import javax.swing.UIManager;
import java.awt.Frame;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.util.Locale;
import java.util.Set;
import java.util.prefs.Preferences;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.Icon;

import java.util.ResourceBundle;
import javax.swing.DefaultListModel;
import javax.swing.JList;
import javax.swing.JScrollPane;
import javax.swing.ListCellRenderer;
import javax.swing.SwingUtilities;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import uk.co.ruelgnoj.prefs.AbstractPreferenceItem;
import uk.co.ruelgnoj.prefs.PreferenceCommitMechanism;
import uk.co.ruelgnoj.prefs.treepanel.util.VerticalFlowLayout;

public class LookAndFeelPI extends AbstractPreferenceItem {

    private static final String RBNAME = "uk.co.ruelgnoj.orion.application.resources";
    private String safeName;
    private LAFPIComponent lafPIComponent;
    private final Icon icon;

    public LookAndFeelPI() {
        final Preferences prefs = Preferences.userNodeForPackage(getClass());
        final String name = prefs.get("look.and.feel", null);
        if (name != null) {
            useLAF(name);
        }
        safeName = UIManager.getLookAndFeel().getClass().getName();
        icon = new LAFIcon();
    }

    @Override
    public Icon getIcon() {
        return icon;
    }

    @Override
    protected JComponent createComponent() {
        lafPIComponent = new LAFPIComponent();
        return lafPIComponent;
    }

    private static class LAFPIComponent extends JPanel {

        private final JList<UIManager.LookAndFeelInfo> lafsList;
        private DefaultListModel<UIManager.LookAndFeelInfo> lafsListModel;

        public LAFPIComponent() {
            super(new VerticalFlowLayout());
            final JPanel panel = new JPanel(new BorderLayout());
            lafsListModel = new DefaultListModel<>();
            lafsList = new JList<>(lafsListModel);
            updateLafsList();
            lafsList.setCellRenderer(new LAFRenderer(lafsList.getCellRenderer()));

            lafsList.addListSelectionListener(new ListSelectionListener() {
                @SuppressWarnings("unchecked")
                @Override
                public void valueChanged(ListSelectionEvent e) {
                    if (e.getValueIsAdjusting()) {
                        return;
                    }
                    final JList<UIManager.LookAndFeelInfo> list = (JList<UIManager.LookAndFeelInfo>) e.getSource();
                    final UIManager.LookAndFeelInfo lafInfo = list.getSelectedValue();
                    if (lafInfo != null) {
                        SwingUtilities.invokeLater(() -> useLAF(lafInfo));
                    }
                }
            });
            final JScrollPane jsp = new JScrollPane(lafsList);
            panel.add(BorderLayout.CENTER, jsp);

            add(panel);
        }

        private void updateLafsList() {
            final UIManager.LookAndFeelInfo[] lafs = UIManager.getInstalledLookAndFeels();
            lafsListModel.removeAllElements();
            for (UIManager.LookAndFeelInfo laf : lafs) {
                lafsListModel.addElement(laf);
            }

            final String currentLafClassName = UIManager.getLookAndFeel().getClass().getName();
            for (UIManager.LookAndFeelInfo laf : lafs) {
                if (laf.getClassName().equals(currentLafClassName)) {
                    lafsList.setSelectedValue(laf, true);
                }
            }
        }

        public UIManager.LookAndFeelInfo getLookAndFeelInfo() {
            return lafsList.getSelectedValue();
        }

    }

    private static void useLAF(UIManager.LookAndFeelInfo lafInfo) {
        try {
            UIManager.setLookAndFeel(lafInfo.getClassName());
            updateFrames();
        } catch (UnsupportedLookAndFeelException | ClassNotFoundException | InstantiationException | IllegalAccessException e) {
            e.printStackTrace();
        }
    }

    private static void updateFrames() {
        for (Frame frame : Frame.getFrames()) {
            if (frame instanceof JFrame) {
                SwingUtilities.updateComponentTreeUI((JFrame) frame);
            }
        }
    }

    private static void useLAF(String className) {
        try {
            UIManager.setLookAndFeel(className);
            updateFrames();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Override
    public String getDisplayName(Locale locale) {
        return getBundleString("displayname", locale);
    }

    @Override
    protected void rollbackChangesImpl() {
        SwingUtilities.invokeLater(() -> useLAF(safeName));
    }

    @Override
    protected void commitChangesImpl(Set<PreferenceCommitMechanism> mechanisms) {
        final Preferences prefs = Preferences.userNodeForPackage(getClass());
        final String name = lafPIComponent.getLookAndFeelInfo().getClassName();
        prefs.put("look.and.feel", name);
        safeName = name;
    }

    private static String getBundleString(String name, Locale locale) {
        return ResourceBundle.getBundle(RBNAME, locale).getString("lafpi." + name);
    }

    private static class LAFRenderer implements ListCellRenderer<UIManager.LookAndFeelInfo> {

        private final ListCellRenderer other;

        public LAFRenderer(ListCellRenderer other) {
            this.other = other;
        }

        @Override
        public Component getListCellRendererComponent(JList<? extends UIManager.LookAndFeelInfo> list,
                UIManager.LookAndFeelInfo lafi,
                int index,
                boolean isSelected,
                boolean cellHasFocus) {
            return other.getListCellRendererComponent(list,
                    lafi.getName(),
                    index,
                    isSelected,
                    cellHasFocus);
        }

    }

    private static class LAFIcon implements Icon {

        public LAFIcon() {
        }

        @Override
        public int getIconWidth() {
            return 15;
        }

        @Override
        public int getIconHeight() {
            return 15;
        }

        @Override
        public void paintIcon(Component comp, Graphics g1d, int x, int y) {
            final Graphics2D g = (Graphics2D) g1d;
            final Paint p = g.getPaint();
            g.setPaint(Color.YELLOW);
            g.fillRect(x + 2, y + 2, getIconWidth() - 2 - 2, getIconHeight() - 2 - 2);
            g.setPaint(Color.RED);
            g.fillRect(x + 4, y + 4, getIconWidth() - 4 - 4, getIconHeight() - 4 - 4);
            g.setPaint(p);
        }
    }
}
