////////////////////////////////////////////////////////////////////////////////
// orion: Provides a set of source code browsing utilities.
// Copyright (C) 2003  Steve Millington
//
// $Id: TabbedPaneOCManager.java 199 2008-01-14 17:04:44Z steve $
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
package uk.co.ruelgnoj.orion.application;

import uk.co.ruelgnoj.orion.project.Project;
import javax.swing.JComponent;
import javax.swing.JTabbedPane;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ContainerListener;
import java.awt.event.ContainerEvent;
import java.awt.event.ComponentEvent;
import java.awt.event.KeyEvent;
import java.util.Map;
import java.util.Set;
import javax.swing.AbstractAction;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.KeyStroke;

import uk.co.ruelgnoj.orion.component.ExceptionView;
import uk.co.ruelgnoj.orion.component.OrionComponent;
import uk.co.ruelgnoj.prefs.PreferenceManager;
import uk.co.ruelgnoj.orion.component.FileViewSettings;
import uk.co.ruelgnoj.orion.dm.OrionDataModelException;

public class TabbedPaneOCManager implements OCManager {

    private final JTabbedPane jtp;
    private final Map<String, String> props;
    private boolean first = true;

    public TabbedPaneOCManager(Map<String, String> props) {
        this.props = props;
        jtp = new JTabbedPane(JTabbedPane.BOTTOM);
        initializeTabbedPane();
    }

    private void initializeTabbedPane() {
        // Create a curses component that is the same size as the
        // OrionComponents that will be added to this pane.
        final OrionComponent orionComponent = new OrionComponent(props);
        final int rows = orionComponent.getRows();
        final int columns = orionComponent.getColumns();
        final WelcomePanel wp = new WelcomePanel(columns, rows);
        wp.setFont(orionComponent.getFont());
        final InputMap iMap = wp.getInputMap();
        iMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), "close");
        iMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_F3, 0), "close");

        final ActionMap aMap = wp.getActionMap();
        aMap.put("close", new CloseAction());

        jtp.addTab("Welcome", wp);

        jtp.setSelectedComponent(wp);
        wp.requestFocus();
    }

    /**
     * Action to pop a view from the stack.
     */
    private class CloseAction extends AbstractAction {

        /**
         * Constructor.
         */
        public CloseAction() {
            super("Close");
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void actionPerformed(ActionEvent e) {
            if (isEnabled()) {
                jtp.removeTabAt(0);
            }
        }
    }

    @Override
    public JComponent getRootComponent() {
        return jtp;
    }

    @Override
    public void add(Project dbDef) {
        final OrionComponent orionComponent = new OrionComponent(props);
        if (first) {
            if (jtp.getTabCount() == 1) {
                jtp.removeTabAt(0);
            }
            new ViewManager();

            final Set<FileViewSettings> set = orionComponent.getSettingsSet();
            PreferenceManager.getInstance().add(new FileViewColorsPI(set));
            first = false;
        }

        try {
            orionComponent.setModel(dbDef.createDataModel());
        } catch (OrionDataModelException e) {
            orionComponent.setView(new ExceptionView(orionComponent, e));
        }

        orionComponent.addComponentListener(new ComponentAdapter() {
            @Override
            public void componentResized(ComponentEvent e) {
                /*
                 //
                 // Make ourself a proper amount of pixels wide and
                 // high.
                 //
                 Dimension sz = orionComponent.getSize();
                 Dimension psz = orionComponent.getPreferredSize();
                 int dw = psz.width - sz.width;
                 int dh = psz.height - sz.height;
                 if ((dw != 0) || (dh != 0)) {
                 Dimension bb = getSize();
                 //
                 // I don't know if this invalidation is required,
                 // but it works!
                 //
                 orionComponent.invalidate();
                 Component cl[] = orionComponent.getComponents();
                 for (int i = 0; i < cl.length; i++) {
                 cl[i].invalidate();
                 }
                         
                 pack();
                 }
                 */
            }
        });
        jtp.addTab(dbDef.getName(), dbDef.getIcon(), orionComponent, dbDef.getURL());
        //------------------------------------------------------------------
        // Test code
        //
        //orionComponent.setBorder(
        //	BorderFactory.createMatteBorder(10,20,30,40,Color.red));
        //------------------------------------------------------------------
        jtp.setSelectedComponent(orionComponent);
        orionComponent.requestFocus();
    }

    /**
     * A class to manage the jtabbed pane. It uses that 'current view' as the
     * trigger form performing add/remove operations and even calling
     * <code>System.exit()</code> when the last view is closed.
     */
    private class ViewManager implements PropertyChangeListener, ContainerListener {

        public ViewManager() {
            jtp.addContainerListener(this);
        }

        @Override
        public void propertyChange(PropertyChangeEvent e) {
            // "view"
            if (e.getNewValue() == null) {
                //
                // When a view changes to null, then we need to remove the
                // tab from the pane.
                //
                jtp.remove((JComponent) e.getSource());
            }
        }

        @Override
        public void componentAdded(ContainerEvent e) {
            //
            // When a tab is added, we want to know about any view that it
            // has so that we can remove the tab when it is done with.
            //
            final Component comp = e.getChild();
            comp.addPropertyChangeListener("view", this);
        }

        @Override
        public void componentRemoved(ContainerEvent e) {
            final Component comp = e.getChild();
            comp.removePropertyChangeListener(this);

            //
            // When the last tab is removed, end the application
            //
            if (jtp.getTabCount() == 0) {
                System.exit(0);
            }
        }
    }
}
