/**
 * *******************************************************************
 */
/*                                                                    */
/* Copyright (c) 2004 by Sun Microsystems, Inc.                       */
/* All rights reserved.                                               */
/*                                                                    */
/**
 * *******************************************************************
 */
package uk.co.ruelgnoj.prefs;

import java.awt.Frame;

import javax.swing.JFrame;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.Action;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;

/**
 * A manager of preferences.
 * <p>
 * There is essentially a single preference manager within an application
 * (SunMAT). Normal use of this class is to get hold of the singleton instance
 * via the <code>getInstance()</code> method.
 * <p>
 * The preference manager manages a set of preferences items. These items
 * essentially form a tree structure, with each item potentially having a set of
 * children items.
 */
public final class PreferenceManager {

    private static final String BUNDLE_NAME = "uk.co.ruelgnoj.prefs.resources";
    //-----------------------------------------------------------------------
    // Singleton support.
    //
    private static PreferenceManager INSTANCE;

    private static synchronized PreferenceManager createInstance() {
        return new PreferenceManager();
    }

    public static PreferenceManager getInstance() {
        if (INSTANCE == null) {
            INSTANCE = createInstance();
        }
        return INSTANCE;
    }
    //-----------------------------------------------------------------------

    private final List<PreferenceItem> items;
    private PreferenceDialog dialog;
    private final Locale locale;

    /**
     * Constructor.
     */
    public PreferenceManager() {
        locale = Locale.getDefault();
        items = new ArrayList<>();
        dialog = null;
        UIManager.addPropertyChangeListener(new PropertyChangeListener() {
            @Override
            public void propertyChange(PropertyChangeEvent e) {
                for (PreferenceItem pi : items) {
                    try {
                        pi.updateUI();
                    } catch (Throwable t) {
                        t.printStackTrace();
                    }
                }
                if (dialog != null) {
                    SwingUtilities.updateComponentTreeUI(dialog);
                }
            }
        });
    }

    /**
     * Add an item to the preference tree.
     *
     * @param pi The new preference item.
     * @throws java.lang.NullPointerException if <code>pi</code> is
     * <code>null</code>.
     */
    public void add(PreferenceItem pi) {
        items.add(pi);
    }

    /**
     * Display the preference management dialog.
     *
     * @param frame The parent frame (may be <code>null</code>)
     */
    public void displayDialog(JFrame frame) {
        final PreferenceDialog dlg = getDialog(frame);
        dlg.setVisible(true);
    }

    /**
     * Display the preference management dialog using the first opened Frame in
     * the application as the parent.
     */
    public void displayDialog() {
        JFrame jf = null;
        final Frame[] frames = Frame.getFrames();
        for (Frame frame : frames) {
            if (frame instanceof JFrame) {
                jf = (JFrame) frame;
                break;
            }
        }
        displayDialog(jf);
    }

    private PreferenceDialog getDialog(JFrame frame) {
        if (dialog == null) {
            dialog = new PreferenceDialog(frame, Collections.unmodifiableList(items));
            dialog.setTitle(ResourceBundle.getBundle(BUNDLE_NAME, locale).getString("dialog.title"));
            UIUpdateHandler.getInstance().registerComponent(dialog);
        }
        return dialog;
    }

    public Action getShowPreferencesAction() {
        return new ShowPreferencesAction(this);
    }
}
