/**********************************************************************/
/*                                                                    */
/* Copyright (c) 2004 by Sun Microsystems, Inc.                       */
/* All rights reserved.                                               */
/*                                                                    */
/**********************************************************************/
package uk.co.ruelgnoj.prefs.treepanel;

import java.awt.Color;
import java.awt.Dimension;

import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JTree;
import javax.swing.UIManager;
import javax.swing.SwingConstants;
import javax.swing.plaf.ColorUIResource;

import javax.swing.tree.TreePath;

/**
 * The area to display the category name.
 * <p>This class should <b>not</b> be treated as a label, but should be left
 * untouched once created. It manages its own color and updating its own
 * value dependant on the supplied tree.
 * <p>The text of the label will depend upon the selection in the tree.
 * If the selection is a <code>TreePanelItem</code> then the text will
 * contain the value obtained from <code>getDisplayName()</code>. Otherwise
 * the text is blank.
 * <p>Not also that the implementation of the <code>TreeSelectionListener</code>
 * interface is an implementation specific and should not be relied upon.
 * @see uk.co.ruelgnoj.prefs.treepanel.TreePanelItem
 */
public class CategoryLabel extends JLabel implements TreeSelectionListener {
    private static final String DEFAULT_TEXT = "";
    private static final int MIN_PREFERRED_WIDTH = 450;

    /**
     * Constructor, using a tree as the text source.
     * @param tree The tree to monitor.
     * @throws java.lang.NullPointerException if tree is null.
     */
    public CategoryLabel(JTree tree) {
	setOpaque(true);
	tree.addTreeSelectionListener(this);

	setValue(tree);
    }

    @Override
    public void valueChanged(TreeSelectionEvent e) {
	JTree tree = (JTree)e.getSource();
	setValue(tree);
    }

    /**
     * Set the current text value dependant on the selection in the supplied
     * tree.
     * @param tree The data source.
     * @throws java.lang.NullPointerException if tree is null.
     */
    private void setValue(JTree tree) {
	Object o = null;
	TreePath selectionPath = tree.getSelectionPath();
	if (selectionPath != null) {
	    o = selectionPath.getLastPathComponent();
	}
	if (o instanceof TreePanelItem) {
	    TreePanelItem tpi = (TreePanelItem)o;
	    setText(tpi.getDisplayName(tree.getLocale()));
	    setIcon(tpi.getIcon());
	} else {
	    setText(DEFAULT_TEXT);
	    setIcon(null);
	}
    }

    @Override
    public Dimension getPreferredSize() {
	Dimension d = super.getPreferredSize();
	if (d.width < MIN_PREFERRED_WIDTH) {
	    d.width = MIN_PREFERRED_WIDTH;
	}
	return d;
    }

    @Override
    public void updateUI() {
	super.updateUI();
	//
	// Try to set our colors and border so that it is pleasing for all LAFs.
	// This is trick, and we have a horrid heuristic here. Basically, we
	// try to use the same colors as an active JInternalFrame title bar.
	// Our border is drawn in the foreground color to give the label
	// some emphasis.
	//
	// Since I am unsure if these properties are ALWAYS set, we heed some
	// protected for if they are not. We then do a hard-coded black/white
	// scenario!
	//
	Color bg = UIManager.getColor("InternalFrame.activeTitleBackground");
	if (bg == null) {
	    bg = new ColorUIResource(0, 0, 0);
	}
	Color fg = UIManager.getColor("InternalFrame.activeTitleForeground");
	if (fg == null) {
	    fg = new ColorUIResource(255, 255, 255);
	}

	setHorizontalTextPosition(SwingConstants.LEADING);
	setBackground(bg);
	setForeground(fg);
	setBorder(BorderFactory.createCompoundBorder(
		    BorderFactory.createLineBorder(fg),
		    BorderFactory.createMatteBorder(0, 5, 0, 5, bg)));
    }
}
