/**
 * *******************************************************************
 */
/*                                                                    */
/* Copyright (c) 2004 by Sun Microsystems, Inc.                       */
/* All rights reserved.                                               */
/*                                                                    */
/**
 * *******************************************************************
 */
package uk.co.ruelgnoj.prefs.treepanel;

import uk.co.ruelgnoj.prefs.treepanel.util.StatusBar;
import uk.co.ruelgnoj.prefs.treepanel.util.SingleLayout;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.SwingConstants;

import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeWillExpandListener;
import javax.swing.event.TreeExpansionEvent;

import javax.swing.tree.ExpandVetoException;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import java.util.ArrayList;
import java.util.List;
import java.util.ResourceBundle;

/**
 * A panel displaying a tree of items, and corresponding components.
 * <p>
 * The panel operates on a list of <code>TreePanelItem</code>s.
 */
public class TreePanel extends JPanel {

    private static final String BUNDLE_NAME = "uk.co.ruelgnoj.prefs.treepanel.resources";

    private final StatusBar statusBar;
    private final List<? extends TreePanelItem> items;
    private final JTree tree;
    private boolean allowTreeContraction;
    private boolean statusBarShowing;

    public TreePanel(List<? extends TreePanelItem> items) {
        super(new TreePanelLayout());
        this.items = items;
        statusBarShowing = true;

        //
        // Implement contraction stopping if required.
        //
        allowTreeContraction = false;
        TreeWillExpandListener expandListener = new TreeWillExpandListener() {
            @Override
            public void treeWillExpand(TreeExpansionEvent e) {
                // Always allow expansion
            }

            @Override
            public void treeWillCollapse(TreeExpansionEvent e) throws ExpandVetoException {
                if (!allowTreeContraction) {
                    throw new ExpandVetoException(e);
                }
            }
        };

        ResourceBundle bundle = ResourceBundle.getBundle(BUNDLE_NAME, getLocale());
        //
        // Create the Tree for the LHS of the display
        //
//	tree = new TreePanelTree(items);
        tree = createJTree();
        tree.addTreeWillExpandListener(expandListener);

        statusBar = new StatusBar(StatusBar.PrefixPolicy.NONE);
        final JPanel ppp = new JPanel(new BorderLayout());
        final JPanel contentPanel = new ContentPanel(tree);
        ppp.add(contentPanel, BorderLayout.CENTER);
        ppp.add(statusBar, BorderLayout.SOUTH);

        final JScrollPane jsp = new JScrollPane(tree);
        final JLabel categoryTitle = new JLabel(
                bundle.getString("category.title"),
                SwingConstants.CENTER);
        categoryTitle.setBorder(BorderFactory.createEtchedBorder());
        final JLabel categoryLabel = new CategoryLabel(tree);

        tree.setSelectionRow(0);

        add(categoryTitle);
        add(categoryLabel);
        add(jsp);
        add(ppp);
    }

    private JTree createJTree() {
        final JTree t = new JTree(new TreePanelTreeModel(items));
        t.setRootVisible(false);
        t.setShowsRootHandles(true);
        t.getSelectionModel().setSelectionMode(
                TreeSelectionModel.SINGLE_TREE_SELECTION);
        t.setCellRenderer(new TreePanelTreeRenderer());

        //
        // Expand all the nodes
        //
        for (int i = 0; i < t.getRowCount(); i++) {
            t.expandRow(i);
        }
        return t;
    }

    @Override
    public void updateUI() {
        super.updateUI();
        if (tree != null) {
            tree.setCellRenderer(new TreePanelTreeRenderer());
        }
    }

    public void setSelection(TreePanelItem item) {
        final List<TreePanelItem> l = new ArrayList<>();
        final List<TreePanelItem> l1 = new ArrayList<>();
        boolean found = getTreePath(items, item, l1);
        if (!l1.isEmpty()) {
            l.addAll(l1);
            final TreePath tp = new TreePath(l.toArray());
            tree.setSelectionPath(tp);
        }
    }

    private boolean getTreePath(TreePanelItem<?> items, TreePanelItem item, List<TreePanelItem> l) {
        if (items.equals(item)) {
            l.add(0, item);
            return true;
        }
        for (TreePanelItem tpi : items.getChildren()) {
            if (getTreePath(tpi, item, l)) {
                l.add(0, items);
                return true;
            }
        }
        return false;
    }

    private boolean getTreePath(List<? extends TreePanelItem> items, TreePanelItem item, List<TreePanelItem> l) {
        for (TreePanelItem tpi : items) {
            if (getTreePath(tpi, item, l)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Determine whether to show the status bar or not.
     *
     * @return True if the status bar is shown
     */
    public boolean isStatusBarShowing() {
        return statusBarShowing;
    }

    /**
     * Set whether to show the status bar or not.
     *
     * @param b True to show the status bar
     */
    public void setStatusBarShowing(boolean b) {
        if (b != statusBarShowing) {
            statusBarShowing = b;
            statusBar.setVisible(statusBarShowing);
            firePropertyChange("statusBarShowing",
                    !statusBarShowing,
                    statusBarShowing);
        }
    }

    /**
     * Set whether the tree allows contaction.
     *
     * @param b True to allow contraction.
     */
    public void setAllowTreeContraction(boolean b) {
        boolean oldAllowTreeContraction = allowTreeContraction;
        allowTreeContraction = b;
        firePropertyChange("allowTreeContraction",
                oldAllowTreeContraction,
                allowTreeContraction);
    }

    /**
     * Determine whether the tree allows contaction.
     *
     * @return True if contraction is allowed.
     */
    public boolean setAllowTreeContraction() {
        return allowTreeContraction;
    }

    /**
     * The main panel displaying tree item content.
     */
    private class ContentPanel extends JPanel {

        private final ValidityListener validityListener;
        private final TreeSelectionListener tsl;
        private final JTree tree;

        /**
         * Constructor.
         *
         * @param tree The tree that governs which child panel this content area
         * should display.
         */
        public ContentPanel(JTree tree) {
            super(new SingleLayout());
            this.tree = tree;
            //
            // Get all of the components
            //
            final TreeModel model = tree.getModel();
            final Object root = model.getRoot();
            addComponents(model, root);

            tsl = new TreeSelectionListener() {
                @Override
                public void valueChanged(TreeSelectionEvent e) {
                    JTree theTree = (JTree) e.getSource();
                    Object o = null;
                    if (theTree.getSelectionPath() != null) {
                        o = theTree.getSelectionPath().getLastPathComponent();
                    }
                    setContent(o);
                }
            };
            validityListener = new ValidityListener();
        }

        @Override
        public void addNotify() {
            super.addNotify();
            tree.addTreeSelectionListener(tsl);
            final Object o;
            if (tree.getSelectionPath() != null) {
                o = tree.getSelectionPath().getLastPathComponent();
            } else {
                o = null;
            }
            setContent(o);

            //
            // Ensure that the tree and buttons are correctly enabled etc
            // for the current state.
            //
            validityListener.refresh();
        }

        @Override
        public void removeNotify() {
            tree.removeTreeSelectionListener(tsl);
            removeValidityListener(this);
            super.removeNotify();
        }

        private void addComponents(TreeModel model, Object o) {
            if (!model.isLeaf(o)) {
                final int n = model.getChildCount(o);
                for (int i = 0; i < n; i++) {
                    addComponents(model, model.getChild(o, i));
                }
            }
            if (o instanceof TreePanelItem) {
                final TreePanelItem tpi = (TreePanelItem) o;
                final Component comp = tpi.getComponent();
                add(comp);
                comp.setVisible(false);
            }
        }

        private void addValidityListener(Container cont) {
            for (int i = 0; i < cont.getComponentCount(); i++) {
                final Component comp = cont.getComponent(i);
                if (comp instanceof Validity) {
                    comp.addPropertyChangeListener("valid", validityListener);
                }
                if (comp instanceof Container) {
                    addValidityListener((Container) comp);
                }
            }
        }

        private void removeValidityListener(Container cont) {
            for (int i = 0; i < cont.getComponentCount(); i++) {
                final Component comp = cont.getComponent(i);
                if (comp instanceof Validity) {
                    comp.removePropertyChangeListener("valid", validityListener);
                }
                if (comp instanceof Container) {
                    removeValidityListener((Container) comp);
                }
            }
        }

        private boolean isValid(Container cont) {
            for (int i = 0; i < cont.getComponentCount(); i++) {
                final Component comp = cont.getComponent(i);
                if (comp instanceof Validity) {
                    Validity val = (Validity) comp;
                    if (!val.isValid()) {
                        return false;
                    }
                }
                if (comp instanceof Container) {
                    final boolean ret = isValid((Container) comp);
                    if (!ret) {
                        return ret;
                    }
                }
            }
            return true;
        }

        private String getValidityMessage(Container cont) {
            for (int i = 0; i < cont.getComponentCount(); i++) {
                final Component comp = cont.getComponent(i);
                if (comp instanceof Validity) {
                    final Validity val = (Validity) comp;
                    if (!val.isValid()) {
                        return val.getValidityMessage();
                    }
                }
                if (comp instanceof Container) {
                    final String ret = getValidityMessage((Container) comp);
                    if (ret != null) {
                        return ret;
                    }
                }
            }
            return null;
        }

        private void setContent(Object o) {
            removeValidityListener(this);
            Component comp = null;
            if (o instanceof TreePanelItem) {
                final TreePanelItem tpi = (TreePanelItem) o;
                comp = tpi.getComponent();
            } else if (o instanceof Component) {
                comp = (Component) o;
            }
            if (comp != null) {
                if (comp.getParent() != this) {
                    add(comp);
                    doLayout();
                }
                comp.setVisible(true);
                invalidate();
                comp.invalidate();
                doLayout();
                repaint();
            }
            for (int i = 0; i < getComponentCount(); i++) {
                if (getComponent(i) != comp) {
                    getComponent(i).setVisible(false);
                }
            }
            addValidityListener(this);
            repaint();
        }

        private class ValidityListener implements PropertyChangeListener {

            private boolean valid;

            public ValidityListener() {
                valid = true;
            }

            @Override
            public void propertyChange(PropertyChangeEvent e) {
                refresh();
            }

            public void refresh() {
                final String message = getValidityMessage(ContentPanel.this);

                final boolean oldValid = valid;
                valid = (message == null);

                if (message != null) {
                    statusBar.error(message);
                } else {
                    statusBar.clear();
                }
                tree.setEnabled(valid);
                final Action okAction = getRootPane().getActionMap().get("ok");
                if (okAction != null) {
                    okAction.setEnabled(valid);
                }

                TreePanel.this.firePropertyChange("valid", oldValid, valid);
            }
        }
    }
}
